#define _GNU_SOURCE
#define _LARGEFILE_SOURCE
#define _LARGEFILE64_SOURCE

#include "dictionary.h"
#include "usb.h"

#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <ctype.h>
#include <stdarg.h>
#include <stdlib.h>
#include <unistd.h>

struct keyinfo {
	struct keyinfo* next;
	size_t valueLength;
	char* value;
	size_t keyLength;
	char key[0];
};

static struct keyinfo* keys = NULL;
static const char* fileName = NULL;

static void
addDictionaryKey(const char* key,
	         const char* value,
		 size_t valueLength)
{
	size_t klen;
	struct keyinfo* k;

	klen = strlen(key) + 1;
	k = malloc(sizeof(*k) + klen + valueLength + 1);
	if (k) {
		memcpy(k->key, key, klen);
		k->keyLength = klen - 1;
		k->value = k->key + klen;
		memcpy(k->value, value, valueLength);
		k->value[valueLength] = 0;
		k->valueLength = valueLength;
		k->next = keys;
		keys = k;
	}
}

int 
readFile(const char* filename)
{
	FILE* f;
	char buf[4096];

	if (!filename) {
		return -EINVAL;
	}	
	f = fopen(filename, "rt");
	if (!f) {
		return -errno;
	}
	fileName = filename;
	while (fgets(buf, sizeof(buf), f)) {
		char* c = buf;
		char* kwd;
		char* kwdend;
		char* val;
		char* valend;

		while (*c && isspace(*c)) {
			c++;
		}
		if (*c == '#' || *c == 0) {
			continue;
		}
		kwd = c;
		while (*c && !isspace(*c) && *c != '=') {
			c++;
		}
		kwdend = c;
		while (*c && isspace(*c)) {
			c++;
		}
		if (*c != '=') {
			continue;
		}
		*kwdend = 0;
		c++;
		while (*c && isspace(*c)) {
			c++;
		}
		if (*c == '\"') {
			c++;
			val = c;
			valend = c;
			while (*c && *c != '\"') {
				if (*c == '|' && isxdigit(c[1]) && isxdigit(c[2])) {
					c[0] = c[1];
					c[1] = c[2];
					c[2] = 0;
					*valend++ = strtoul(c, NULL, 16);
					c += 3;
				} else {
					*valend++ = *c++;
				}
			}
			if (*c != '\"') {
				continue;
			}	
		} else {
			val = c;
			while (*c && !isspace(*c)) {
				c++;
			}
			valend = c;
		}
		addDictionaryKey(kwd, val, valend - val);	
	}
	fclose(f);
	return 0;
}

static int 
updateFile(const char* key, size_t keylen, const char* value, size_t valueLen)
{
	FILE* f;
	FILE* w;
	char buf[4096];
	char wf[4096];
	
	if (!fileName) {
		return -EINVAL;
	}
	sprintf(wf, "%s.tmp.%u", fileName, getpid());
	w = fopen(wf, "wt");
	if (!w) {
		return -errno;
	}
	f = fopen(fileName, "rt");
	if (!f) {
		return -errno;
	}
	for (;fgets(buf, sizeof(buf), f); fprintf(w, "%s", buf)) {
		char* c = buf;
		const char* ptr;
		const char* ptrend;
		char* val;
		const char* valend;

		if (!key) {
			continue;
		}
		while (*c && isspace(*c)) {
			c++;
		}
		if (*c == '#' || *c == 0) {
			continue;
		}
		if (strncasecmp(c, key, keylen)) {
			continue;
		}
		c += keylen;
		while (*c && isspace(*c)) {
			c++;
		}
		if (*c != '=') {
			continue;
		}
		c++;
		while (*c && isspace(*c)) {
			c++;
		}
		val = c;
		if (*c == '\"') {
			c++;
			while (*c && *c != '\"') {
				if (*c == '|' && isxdigit(c[1]) && isxdigit(c[2])) {
					c += 3;
				} else {
					c++;
				}
			}
			if (*c != '\"') {
				valend = c;
			} else {
				valend = c + 1;
			}
		} else {
			while (*c && !isspace(*c)) {
				c++;
			}
			valend = c;
		}
		*val = 0;
		fprintf(w, "%s\"", buf);
		ptrend = value + valueLen;
		for (ptr = value; ptr < ptrend; ptr++) {
			if (*ptr < 0x20 || *ptr == '\"' || *ptr == '|') {
				fprintf(w, "|%02X", *ptr & 0xFF);
			} else {
				fwrite(ptr, 1, 1, w);
			}
		}
		fprintf(w, "\"%s", valend);
		buf[0] = 0;
		key = NULL;
	}
	fclose(f);
	if (key) {
		const char* ptr;
		const char* ptrend;

		fprintf(w, "%s = \"", key);
		ptrend = value + valueLen;
		for (ptr = value; ptr < ptrend; ptr++) {
			if (*ptr < 0x20 || *ptr == '\"' || *ptr == '|') {
				fprintf(w, "|%02X", *ptr & 0xFF);
			} else {
				fwrite(ptr, 1, 1, w);
			}
		}
		fprintf(w, "\"\n");
	}
	fclose(w);
	rename(wf, fileName);
	return 0;
}


static int 
updateFileSafe(const char* key, size_t keylen, const char* value, size_t valueLen)
{
	uid_t u;
	int r;

	u = becomeNormalUser();
	r = updateFile(key, keylen, value, valueLen);
	restoreNormalUser(u);
	return r;
}

static const char*
dictVGetString(const char* defaultValue, const char* fmt, va_list va)
{
	struct keyinfo* k;
	size_t nameLen;
	char name[1024];

	vsnprintf(name, sizeof(name), fmt, va);
	
	nameLen = strlen(name);
	for (k = keys; k; k = k->next) {
		if (k->keyLength == nameLen && !strcasecmp(k->key, name)) {
			return k->value;
		}
	}
	return defaultValue;
}

const char*
dictGetString(const char* defaultValue, const char* fmt, ...)
{
	va_list va;
	const char* ret;

	va_start(va, fmt);
	ret = dictVGetString(defaultValue, fmt, va);
	va_end(va);
	return ret;
}

unsigned int
dictGetInteger(unsigned int defaultValue, const char* fmt, ...)
{
	va_list va;
	const char* ret;
	
	va_start(va, fmt);
	ret = dictVGetString(NULL, fmt, va);
	va_end(va);
	if (ret) {
		char* e;
		unsigned int mv;
		
		mv = strtoul(ret, &e, 0);
		if (!e || !*e) {
			return mv;
		}
		if (!strcasecmp(ret, "TRUE")) {
			return 1;
		}
		if (!strcasecmp(ret, "FALSE")) {
			return 0;
		}
	}
	return defaultValue;
}

unsigned int
dictGetBoolean(unsigned int defaultValue, const char* fmt, ...)
{
	va_list va;
	const char* ret;
	
	va_start(va, fmt);
	ret = dictVGetString(NULL, fmt, va);
	va_end(va);
	if (ret) {
		char* e;
		unsigned int mv;
		
		mv = strtoul(ret, &e, 0);
		if ((mv == 0 || mv == 1) && (!e || !*e)) {
			return mv;
		}
		if (!strcasecmp(ret, "TRUE")) {
			return 1;
		}
		if (!strcasecmp(ret, "FALSE")) {
			return 0;
		}
	}
	return defaultValue;
}

static int
dictVSetString(const char* newValue, const char* fmt, va_list va)
{
	char name[1024];

	vsnprintf(name, sizeof(name), fmt, va);
	return updateFileSafe(name, strlen(name), newValue, strlen(newValue));
}

int
dictSetString(const char* newValue, const char* fmt, ...)
{
	va_list va;
	int ret;
	
	va_start(va, fmt);
	ret = dictVSetString(newValue, fmt, va);
	va_end(va);
	return ret;
}

int
dictSetInteger(unsigned int newValue, const char* fmt, ...)
{
	char tmp[40];
	va_list va;
	int ret;
	
	sprintf(tmp, "%u", newValue);
	va_start(va, fmt);
	ret = dictVSetString(tmp, fmt, va);
	va_end(va);
	return ret;
}

int
dictSetBoolean(int newValue, const char* fmt, ...)
{
	va_list va;
	int ret;
	
	va_start(va, fmt);
	ret = dictVSetString(newValue ? "TRUE" : "FALSE", fmt, va);
	va_end(va);
	return ret;
}
