#define _GNU_SOURCE
#define _LARGEFILE_SOURCE
#define _LARGEFILE64_SOURCE

#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
#include <stdarg.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <dlfcn.h>
#include <dirent.h>
#include <ctype.h>
#include <pthread.h>
#include <obstack.h>

#include <sys/ioctl.h>
#include <sys/socket.h>
#include <sys/types.h>

#include "usb.h"

#include "dictionary.h"

#include "usbcontrol.h"	/* For usbcontrolInit() */

static const char pbud[] = "/proc/bus/usb/";
static const int pbud_len = sizeof(pbud) - 1;
static uid_t realUserId;

#define USE_NEXT_FN	(-2)

#define BUSID		(4095)

static FILE* fo = NULL;
static int debug = 0;

void usb_log(const char* msg, ...) {
	if (debug) {
		if (fo) {
			va_list va;

			va_start(va, msg);
			vfprintf(fo, msg, va);
			va_end(va);
			fflush(fo);
		}
	}
}

void usb_warn(const char* msg, ...) {
	if (debug) {
		va_list va;

		va_start(va, msg);
		vfprintf(stderr, msg, va);
		if (fo) {
			vfprintf(fo, msg, va);
			fflush(fo);
		}
		va_end(va);
	}
}

void
restoreNormalUser(uid_t euid)
{
        if (euid == 0) {
                seteuid(euid);
        }
}

uid_t
becomeNormalUser(void)
{
        uid_t euid;

        euid = geteuid();
        if (euid == 0) {
                seteuid(realUserId);
        }
        return euid;
}

#define obstack_chunk_alloc malloc
#define obstack_chunk_free free

struct dirent64_21
  {
    __ino_t d_ino;
    __off64_t d_off;
    unsigned short int d_reclen;
    unsigned char d_type;
    char d_name[256];           /* We must not include limits.h! */
  };

static void dirent64_to_dirent64_21(struct dirent64_21* dst, const struct dirent64* src) {
	size_t namelen;
	
	namelen = src->d_reclen - offsetof(struct dirent64, d_name);
	dst->d_ino = src->d_ino;
	dst->d_off = src->d_off;
	dst->d_reclen = namelen + offsetof(struct dirent64_21, d_name);
	dst->d_type = src->d_type;
	memcpy(dst->d_name, src->d_name, namelen);
}

static void dirent64_to_dirent(struct dirent* dst, const struct dirent64* src) {
	size_t namelen;
	
	namelen = src->d_reclen - offsetof(struct dirent64, d_name);
	dst->d_ino = src->d_ino;
	dst->d_off = src->d_off;
	dst->d_reclen = namelen + offsetof(struct dirent, d_name);
	dst->d_type = src->d_type;
	memcpy(dst->d_name, src->d_name, namelen);
}

static int (*__next_open)(const char* filename, int flags, mode_t mode) = NULL;
static int (*__next_open64)(const char* filename, int flags, mode_t mode) = NULL;
static off_t (*__next_lseek)(int fildes, off_t offset, int whence) = NULL;
static off64_t (*__next_lseek64)(int fildes, off64_t offset, int whence) = NULL;
static int (*__next_read)(int fildes, void* buffer, size_t size) = NULL;
static int (*__next_write)(int fildes, const void* buffer, size_t size) = NULL;
static int (*__next_ioctl)(int fildes, unsigned long int command, void* param) = NULL;
static int (*__next_close)(int fildes) = NULL;
static FILE* (*__next_fopen)(const char* filename, const char* mode) = NULL;
static FILE* (*__next_fopen64)(const char* filename, const char* mode) = NULL;
static DIR* (*__next_opendir)(const char* filename) = NULL;
static struct dirent* (*__next_readdir)(DIR* dirp) = NULL;
static struct dirent64* (*__next_readdir64)(DIR* dirp) = NULL;
static struct dirent64_21* (*__next_readdir64_21)(DIR* dirp) = NULL;
static int (*__next_closedir)(DIR* dirp) = NULL;
static void (*__next_rewinddir)(DIR* dirp) = NULL;
static void (*__next_seekdir)(DIR* dirp, long int offset) = NULL;
static long int (*__next_telldir)(DIR* dirp) = NULL;

struct usbdev {
	struct usbdev* next;
	int devid;
	void* cookie;
	struct usbdev_fn* fns;
};

struct usbbus {
	struct usbbus* next;
	int busid;
	struct usbdev* devlist;
};

struct opninfo {
	struct usbdev* dev;
	int file;
	int pipe0;
	int pipe1;
	struct urbs* urb;
};

static void make_writeable(struct opninfo* opn) {
	static const int bigval = 131072;
	
	setsockopt(opn->pipe0, SOL_SOCKET, SO_SNDBUF, &bigval, sizeof(bigval));
}

static void make_full(struct opninfo* opn) {
	static const int zero = 0;
	
	setsockopt(opn->pipe0, SOL_SOCKET, SO_SNDBUF, &zero, sizeof(zero));
	__next_write(opn->pipe0, pbud, sizeof(pbud));
}

static void dbd(const char* msg, unsigned char* buffer, size_t len) {
	if (debug) {
		size_t i;

		usb_warn("%s", msg);
		for (i = 0; i < len; i++) {
			usb_warn("%02X ", buffer[i]);
		}
		usb_warn("\n");
	}
}

void usbdev_puturb(struct urbs* urbs) {
	struct opninfo* opn = urbs->opn;
	struct usbdevfs_urb* urb = urbs->urb;

	if (urb->endpoint & 0x80) {
		switch (urb->type) {
			case USBDEVFS_URB_TYPE_CONTROL:
				dbd("Control Data:   ", urb->buffer + 8, urb->actual_length);
				break;
			case USBDEVFS_URB_TYPE_BULK:
			case USBDEVFS_URB_TYPE_INTERRUPT:
				dbd("Bulk/Intr Data: ", urb->buffer, urb->actual_length);
				break;
		}
	}
	usb_log("Count: %u, Error: %d, Errors Count: %u\n",
		urb->actual_length,
		urb->status,
		urb->error_count);

	urbs->next = opn->urb;
	opn->urb = urbs;
	if (urbs->next == NULL) {
		make_writeable(opn);
	}
}

static struct usbdevfs_urb* fetchurb(struct opninfo* opn) {
	struct urbs* urbs;
	
	urbs = opn->urb;
	if (urbs) {
		struct usbdevfs_urb* urb;

		urb = urbs->urb;
		opn->urb = urbs->next;
		free(urbs);
		if (!opn->urb) {
			make_full(opn);
		}
		return urb;
	}
	return NULL;
}		
		
static void init_usb_data(struct usbdev* dev, int fd) {
	struct usb_device_descriptor udd;
	char buffer[4096];
	struct usbdevfs_ctrltransfer ct;
	int ln;
	int i;
	
	ct.bRequestType = USB_DIR_IN | USB_TYPE_STANDARD | USB_RECIP_DEVICE;
	ct.bRequest = USB_REQ_GET_DESCRIPTOR;
	ct.wValue = (USB_DT_DEVICE << 8) | 0x00;
	ct.wIndex = 0;
	ct.wLength = sizeof(udd);
	ct.data = &udd;
	ln = dev->fns->USBControl(dev->cookie, &ct);
	__next_write(fd, &udd, ln);
	for (i = 0; i < udd.bNumConfigurations; i++) {
		ct.wValue = (USB_DT_CONFIG << 8) | i;
		ct.wLength = sizeof(buffer);
		ct.data = buffer;
		ln = dev->fns->USBControl(dev->cookie, &ct);
		__next_write(fd, buffer, ln);
	}
	__next_lseek64(fd, 0, SEEK_SET);
}

static int usbdevioctl(void* cookie, unsigned long int command, void* data) {
	struct opninfo* opn = cookie;
	struct usbdev* dev = opn->dev;
	int ret = USE_NEXT_FN;

	switch (command) {
		case USBDEVFS_CONTROL:
			{
				struct usbdevfs_ctrltransfer* ct = data;
				usb_log("USBDEVFS_CONTROL rt=%02X rq=%02X vl=%04X idx=%04X len=%04X tmo=%u data=%p\n",
					ct->bRequestType, ct->bRequest, ct->wValue, ct->wIndex, ct->wLength, ct->timeout,
					ct->data);
				ret = dev->fns->USBControl(dev->cookie, ct);
				if (ret < 0) {
					errno = -ret;
					ret = -1;
				}
			}
			break;//           _IOWR('U', 0, struct usbdevfs_ctrltransfer)
		case USBDEVFS_BULK:
			usb_log("USBDEVFS_BULK\n");
			break;//unused              _IOWR('U', 2, struct usbdevfs_bulktransfer)
		case USBDEVFS_RESETEP:
			usb_log("USBDEVFS_RESETEP\n");
			break;//unused           _IOR('U', 3, unsigned int)
		case USBDEVFS_SETINTERFACE:
			{
				struct usbdevfs_setinterface* si = data;
				struct usbdevfs_ctrltransfer ct;

				usb_log("USBDEVFS_SETINTERFACE intf=%u, alt=%u\n", si->interface, si->altsetting);
				
				ct.bRequestType = 0x01;
				ct.bRequest = USB_REQ_SET_INTERFACE;
				ct.wValue = si->altsetting;
				ct.wIndex = si->interface;
				ct.wLength = 0;
				ct.data = NULL;
				
				ret = dev->fns->USBControl(dev->cookie, &ct);
				usb_log("... %d\n", ret);
				if (ret < 0) {
					errno = -ret;
					ret = -1;
				}
			}
			break;//      _IOR('U', 4, struct usbdevfs_setinterface)
		case USBDEVFS_SETCONFIGURATION:
			{
				unsigned int* sc = data;
				struct usbdevfs_ctrltransfer ct;
				
				usb_log("USBDEVFS_SETCONFIGURATION cfg=%u\n", *sc);
			
				ct.bRequestType = 0x00;
				ct.bRequest = USB_REQ_SET_CONFIGURATION;
				ct.wValue = *sc;
				ct.wIndex = 0;
				ct.wLength = 0;
				ct.data = NULL;	

				ret = dev->fns->USBControl(dev->cookie, &ct);
				usb_log("... %d\n", ret);
				if (ret < 0) {
					errno = -ret;
					ret = -1;
				}
			}
			break;//  _IOR('U', 5, unsigned int)
		case USBDEVFS_GETDRIVER:
			usb_log("USBDEVFS_GETDRIVER\n");
			break;//         _IOW('U', 8, struct usbdevfs_getdriver)
		case USBDEVFS_SUBMITURB:
			{
				struct usbdevfs_urb* urb = data;
				struct urbs* urbs;
				int i;
				
				usb_log("USBDEVFS_SUBMITURB type=%u, endpoint=%u, status=%u, flags=%u, buffer=%p, blen=%u, actlen=%u, sf=%u, num=%u, errcnt=%u, signr=%u, userctx=%p\n",
					urb->type, urb->endpoint, urb->status, urb->flags,
					urb->buffer, urb->buffer_length, urb->actual_length,
					urb->start_frame, urb->number_of_packets, urb->error_count,
					urb->signr, urb->usercontext);
				for (i = 0; i < urb->number_of_packets; i++) {
					usb_log("... #%u: len=%u, actlen=%u, status=%u\n",
						i, urb->iso_frame_desc[i].length,
						urb->iso_frame_desc[i].actual_length,
						urb->iso_frame_desc[i].status);
				}
				switch (urb->type) {
					case USBDEVFS_URB_TYPE_CONTROL:
						usb_log("Control");
						break;
					case USBDEVFS_URB_TYPE_BULK:
						usb_log("Bulk");
						break;
					case USBDEVFS_URB_TYPE_ISO:
						usb_log("ISO");
						break;
					case USBDEVFS_URB_TYPE_INTERRUPT:
						usb_log("Interrupt");
						break;
					default:
						errno = EINVAL;
						return -1;
				}
				if ((urb->endpoint & 0x80) == 0) {
					dbd("Data: ", urb->buffer, urb->buffer_length);
				} else if (urb->type == USBDEVFS_URB_TYPE_CONTROL) {
					dbd("Data: ", urb->buffer, 8);
				} else {
					usb_log("\n");
				}
				
				urbs = malloc(sizeof(*urbs));
				if (!urbs) {
					usb_log("SUBMITURB run out of memory\n");
					errno = ENOMEM;
					return -1;
				}
				urbs->urb = urb;
				urbs->opn = opn;
				urbs->next = NULL;
				switch (urb->type) {
					case USBDEVFS_URB_TYPE_CONTROL: {
						struct usbdevfs_ctrltransfer ct;
						struct usb_ctrlrequest* cr = urb->buffer;
						
						ct.bRequestType = cr->bRequestType;
						ct.bRequest = cr->bRequest;
						ct.wValue = cr->wValue;
						ct.wIndex = cr->wIndex;
						ct.wLength = cr->wLength;
						ct.data = urb->buffer + 8;
						ret = dev->fns->USBControl(dev->cookie, &ct);
						usb_log("USB Control xfer: %d\n", ret);
						if (ret < 0) {
							free(urbs);
							errno = -ret;
							return -1;
						}
						urb->actual_length = ret;
						urb->status = 0;
						urb->error_count = 0;
						usbdev_puturb(urbs);
						ret = 0;
						break;
					}
					case USBDEVFS_URB_TYPE_BULK: 
					case USBDEVFS_URB_TYPE_INTERRUPT: {
						ret = dev->fns->USBXfer(dev->cookie, urbs, (urb->type == USBDEVFS_URB_TYPE_BULK) ? USB_ENDPOINT_XFER_BULK : USB_ENDPOINT_XFER_INT);
						usb_log("USB Bulk xfer: %d\n", ret);
						if (ret < 0) {
							free(urbs);
							errno = -ret;
							return -1;
						}
						ret = 0;
						break;
					}
					default:
						free(urbs);
						break;
				}
						
			}
			break;//         _IOR('U', 10, struct usbdevfs_urb)
		case USBDEVFS_DISCARDURB:
			{
				struct usbdevfs_urb* urb = data;
				
				usb_log("USBDEVFS_DISCARDURB, urb=%p\n", data);
				ret = dev->fns->USBDiscardUrb(dev->cookie, urb);
				usb_log("USBDEVFS_DISCARDURB returns %d\n", ret);
				if (ret < 0) {
					errno = -ret;
					ret = -1;
				}
			}
			break;//        _IO('U', 11)
		case USBDEVFS_REAPURB:
			usb_log("USBDEVFS_REAPURB\n");
			break;//           _IOW('U', 12, void *)
		case USBDEVFS_REAPURBNDELAY:
			{
					void** ptr = data;
					struct usbdevfs_urb* urb;

					urb = fetchurb(opn);
					usb_log("USBDEVFS_REAPURBNDELAY, found %p\n", urb);
					if (!urb) {
						errno = EAGAIN;
						return -1;
					}
					*ptr = urb;
					return 0;
			} //     _IOW('U', 13, void *)
		case USBDEVFS_DISCSIGNAL:
			usb_log("USBDEVFS_DISCSIGNAL\n");
			break;//unused        _IOR('U', 14, struct usbdevfs_disconnectsignal)
		case USBDEVFS_CLAIMINTERFACE:
			{
				unsigned int* interface = data;

				usb_log("USBDEVFS_CLAIMINTERFACE num=%u\n", *interface);
				ret = dev->fns->USBClaimInterface(dev->cookie, *interface);
				if (ret < 0) {
					errno = -ret;
					ret = -1;
				}
			}
			break;//    _IOR('U', 15, unsigned int)
		case USBDEVFS_RELEASEINTERFACE:
			{
				unsigned int* interface = data;

				usb_log("USBDEVFS_RELEASEINTERFACE num=%u\n", *interface);
				ret = dev->fns->USBReleaseInterface(dev->cookie, *interface);
				if (ret < 0) {
					errno = -ret;
					ret = -1;
				}
			}
			break;//  _IOR('U', 16, unsigned int)
		case USBDEVFS_CONNECTINFO:
			usb_log("USBDEVFS_CONNECTINFO\n");
			break;//unused       _IOW('U', 17, struct usbdevfs_connectinfo)
		case USBDEVFS_IOCTL:
			usb_log("USBDEVFS_IOCTL\n");
			break;//unused             _IOWR('U', 18, struct usbdevfs_ioctl)
		case USBDEVFS_HUB_PORTINFO:
			usb_log("USBDEVFS_HUB_PORTINFO\n");
			break;//unused      _IOR('U', 19, struct usbdevfs_hub_portinfo)
		case USBDEVFS_RESET:
			usb_log("USBDEVFS_RESET\n");
			break;//             _IO('U', 20)
		case USBDEVFS_CLEAR_HALT:
			{
				unsigned int* si = data;
				struct usbdevfs_ctrltransfer ct;
				
				usb_log("USBDEVFS_CLEAR_HALT ep=%02X\n", *si);
				
				ct.bRequestType = 0x02;
				ct.bRequest = USB_REQ_CLEAR_FEATURE;
				ct.wValue = 0;
				ct.wIndex = *si;
				ct.wLength = 0;
				ct.data = NULL;
				
				ret = dev->fns->USBControl(dev->cookie, &ct);
				if (ret < 0) {
					errno = -ret;
					ret = -1;
				}
			}
			break;//        _IOR('U', 21, unsigned int)
		case USBDEVFS_DISCONNECT:
			usb_log("USBDEVFS_DISCONNECT\n");
			break;//unused        _IO('U', 22)
		case USBDEVFS_CONNECT:
			usb_log("USBDEVFS_CONNECT\n");
			break;//unused           _IO('U', 23)
	}
	return ret;
}

struct usbfs_devices_info {
	int fd;
	int start_off;
	void* buff;
	int buff_len;
};

static char* usbdev_allstrings(size_t* len);

static struct usbfs_devices_info* usbfs_devices_init(int fd) {
	struct usbfs_devices_info* info;
	
	info = malloc(sizeof(*info));
	if (info) {
		info->fd = fd;
		info->start_off = 0;
		info->buff = usbdev_allstrings(&info->buff_len);
	}
	return info;
}

static int usbfs_devices_read(void* cookie, void* buffer, size_t size) {
	struct usbfs_devices_info* info = cookie;
	int r;
	
	if (info->start_off == 0) {
		r = __next_read(info->fd, buffer, size);
		if (r == -1 || (size_t)r >= size) {
			return r;
		}
		buffer += r;
		size -= r;
	} else {
		r = 0;
	}
	if (size > (size_t)info->buff_len - info->start_off) {
		size = info->buff_len - info->start_off;
	}
	memcpy(buffer, info->buff + info->start_off, size);
	info->start_off += size;
	return r + size;
}

static void usbfs_devices_cleanup(void* cookie) {
	struct usbfs_devices_info* info = cookie;

	free(info);
}

static off64_t usbfs_dev_lseek(void* cookie, off64_t offset, int whence) {
	struct opninfo* opn = cookie;
	
	return __next_lseek64(opn->file, offset, whence);
}

static int usbfs_dev_read(void* cookie, void* buffer, size_t size) {
	struct opninfo* opn = cookie;

	return __next_read(opn->file, buffer, size);
}

static void usbfs_dev_cleanup(void* cookie) {
	struct opninfo* opn = cookie;
	
	close(opn->file);
	close(opn->pipe1);
	free(opn);
}

static u_int32_t spfd[1024];

struct fdfns {
	off64_t (*lseek)(void* cookie, off64_t offset, int whence);
	int (*read)(void* cookie, void* buffer, size_t size);
	int (*write)(void* cookie, const void* buffer, size_t size);
	int (*ioctl)(void* cookie, unsigned long int command, void* data);
	void (*cleanup)(void* cookie);
};

static struct fdfns fd_devices = {
	.lseek = NULL,
	.read = usbfs_devices_read,
	.write = NULL,
	.ioctl = NULL,
	.cleanup = usbfs_devices_cleanup,
};

static struct fdfns fd_devnode = {
	.lseek = usbfs_dev_lseek,
	.read = usbfs_dev_read,
	.write = NULL,
	.ioctl = usbdevioctl,
	.cleanup = usbfs_dev_cleanup,
};

struct mystr {
	struct mystr* next;
	unsigned int fildes;
	int usecnt;
	void* cookie;
	const struct fdfns* fns;
};

static struct mystr* fds;
static pthread_mutex_t fdlock = PTHREAD_MUTEX_INITIALIZER;

static struct mystr* newfd(int fildes, void* cookie, const struct fdfns* fns) {
	struct mystr* tmp;
	
	tmp = malloc(sizeof(*tmp));
	tmp->next = NULL;
	tmp->fildes = fildes;
	tmp->usecnt = 1;
	tmp->cookie = cookie;
	tmp->fns = fns;
	return tmp;
}

static struct mystr* isspecialfd(unsigned int fildes) {
	struct mystr* cp;

	if (fildes < sizeof(spfd) * 8) {
		if ((spfd[fildes / 32] & (1 << (fildes & 31))) == 0) {
			return NULL;
		}
	}
	pthread_mutex_lock(&fdlock);
	for (cp = fds; cp; cp = cp->next) {
		if (cp->fildes == fildes) {
			cp->usecnt++;
			pthread_mutex_unlock(&fdlock);
			return cp;
		}
		if (cp->fildes > fildes) {
			break;
		}
	}
	pthread_mutex_unlock(&fdlock);
	return NULL;
}

static void insertspecialfd(struct mystr* p) {
	struct mystr** cp;
	struct mystr* c2;
	unsigned int fildes = p->fildes;
	
	pthread_mutex_lock(&fdlock);
	for (cp = &fds; (c2 = *cp) != NULL; cp = &c2->next) {
		if (c2->fildes >= fildes) {
			break;
		}
	}
	p->next = c2;
	*cp = p;
	if (fildes < sizeof(spfd) * 8) {
		spfd[fildes / 32] |= 1 << (fildes & 31);
	}
	pthread_mutex_unlock(&fdlock);
}

static struct mystr* removespecialfd(unsigned int fildes) {
	struct mystr** cp;
	struct mystr* c2;
	
	if (fildes < sizeof(spfd) * 8) {
		if ((spfd[fildes / 32] & (1 << (fildes & 31))) == 0) {
			return NULL;
		}
	}
	pthread_mutex_lock(&fdlock);
	for (cp = &fds; (c2 = *cp) != NULL; cp = &c2->next) {
		if (c2->fildes == fildes) {
			*cp = c2->next;
			c2->next = NULL;
			spfd[fildes / 32] &= ~(1 << (fildes & 31));
			pthread_mutex_unlock(&fdlock);
			return c2;
		}
		if (c2->fildes > fildes) {
			break;
		}
	}
	pthread_mutex_unlock(&fdlock);
	return NULL;
}

static void releasefd(struct mystr* cp) {
	if (--cp->usecnt == 0) {
		if (cp->fns->cleanup) {
			cp->fns->cleanup(cp->cookie);
		}
		free(cp);
	}
}

static struct usbbus* usbroot = NULL;
static pthread_mutex_t usbdevlock = PTHREAD_MUTEX_INITIALIZER;

size_t get_registered_usbdevs(u_int32_t* devlist, size_t maxlen) {
	struct usbbus* bus;
	struct usbdev* dev;
	u_int32_t* end;
	u_int32_t* ptr;

	end = devlist + maxlen;
	ptr = devlist;
	pthread_mutex_lock(&usbdevlock);
	for (bus = usbroot; bus != NULL; bus = bus->next) {
		for (dev = bus->devlist; dev != NULL; dev = dev->next) {
			if (ptr < end) {
				*ptr++ = (bus->busid << 16) | dev->devid;
			}
		}
	}
	pthread_mutex_unlock(&usbdevlock);
	return ptr - devlist;
}

int register_usbdev(void* cookie, struct usbdev_fn* fn, int busid, int devid) {
	struct usbbus** pbus;
	struct usbbus* bus;
	struct usbdev** pdev;
	struct usbdev* dev;

	if (busid == -1) {
		busid = BUSID;
	}
	if (busid < 0 || busid > 4095) {
		errno = EINVAL;
		return -1;
	}
	if (devid < -1 || devid > 65535) {
		errno = EINVAL;
		return -1;
	}
	pthread_mutex_lock(&usbdevlock);
	for (pbus = &usbroot; (bus = *pbus) != NULL && bus->busid < busid; pbus = &bus->next) {
	}
	if (!bus || bus->busid != busid) {
		struct usbbus* newbus;
		
		newbus = malloc(sizeof(*newbus));
		if (!newbus) {
			pthread_mutex_unlock(&usbdevlock);
			errno = ENOMEM;
			return -1;
		}
		newbus->busid = busid;
		newbus->devlist = NULL;
		newbus->next = bus;
		*pbus = newbus;
		pdev = &newbus->devlist;
	} else {
		pdev = &bus->devlist;
	}
	if (devid == -1) {
		devid = 0;
		for (; (dev = *pdev) != NULL && dev->devid == devid; pdev = &dev->next) {
			devid = dev->devid + 1;
		}
	} else {
		for (; (dev = *pdev) != NULL && dev->devid < devid; pdev = &dev->next) {
		}
	}
	if (!dev || dev->devid != devid) {
		struct usbdev* newdev;
		
		newdev = malloc(sizeof(*newdev));
		if (!newdev) {
			pthread_mutex_unlock(&usbdevlock);
			errno = ENOMEM;
			return -1;
		}
		newdev->devid = devid;
		newdev->cookie = cookie;
		newdev->next = dev;
		newdev->fns = fn;
		*pdev = newdev;
	} else {
		pthread_mutex_unlock(&usbdevlock);
		errno = EEXIST;
		return -1;
	}
	pthread_mutex_unlock(&usbdevlock);
	return 0;
}

static int usbbus_next(int busid) {
	struct usbbus* bus;
	
	pthread_mutex_lock(&usbdevlock);
	for (bus = usbroot; bus && bus->busid < busid; bus = bus->next) {
	}
	if (bus) {
		busid = bus->busid;
	} else {
		busid = -1;
	}
	pthread_mutex_unlock(&usbdevlock);
	return busid;
}

static int usbbus_exist(int busid) {
	struct usbbus* bus;
	int rval;
	
	pthread_mutex_lock(&usbdevlock);
	for (bus = usbroot; bus && bus->busid < busid; bus = bus->next) {
	}
	rval = bus && busid == bus->busid;
	pthread_mutex_unlock(&usbdevlock);
	return rval;
}

static int usbdev_getidx(int busid, int idx) {
	struct usbbus* bus;
	int rval = -1;
	
	pthread_mutex_lock(&usbdevlock);
	for (bus = usbroot; bus && bus->busid < busid; bus = bus->next) {
	}
	if (bus && busid == bus->busid) {
		struct usbdev* dev;
		
		for (dev = bus->devlist; dev && idx; idx--, dev = dev->next) {
		}
		if (dev) {
			rval = dev->devid;
		}
			
	}
	pthread_mutex_unlock(&usbdevlock);
	return rval;
}

static struct usbdev* usbdev_exist(int busid, int devid) {
	struct usbbus* bus;
	struct usbdev* rval = NULL;
	
	pthread_mutex_lock(&usbdevlock);
	for (bus = usbroot; bus && bus->busid < busid; bus = bus->next) {
	}
	if (bus && busid == bus->busid) {
		struct usbdev* dev;
		
		for (dev = bus->devlist; dev && dev->devid < devid; dev = dev->next) {
		}
		if (dev && devid == dev->devid) {
			rval = dev;
		}
	}
	pthread_mutex_unlock(&usbdevlock);
	return rval;
}

void* usbdev_find(int busid, int devid) {
	struct usbdev* usbdev;
	
	usbdev = usbdev_exist(busid, devid);
	if (usbdev) {
		return usbdev->cookie;
	}
	return NULL;
}

static char* usbdev_allstrings(size_t* len) {
	struct obstack obs;
	struct usbbus* bus;
	char* p;
	
	obstack_init(&obs);
	pthread_mutex_lock(&usbdevlock);
	for (bus = usbroot; bus; bus = bus->next) {
		struct usbdev* dev;
		
		for (dev = bus->devlist; dev; dev = dev->next) {
			/* Everything except Bus, Dev#, Vendor and ProdID is ignored */
			obstack_printf(&obs, "\n"
					     "T:  Bus=%3d Lev=00 Prnt=00 Port=00 Cnt=00 Dev#=%3d Spd=12  MxCh=0\n"
					     "P:  Vendor=0000 ProdID=0000 Rev= 0.00\n",
					     bus->busid, dev->devid);
		}
	}
	pthread_mutex_unlock(&usbdevlock);
	p = strdup(obstack_finish(&obs));
	obstack_free(&obs, NULL);
	*len = strlen(p);
	return p;
}

static void childCleanup(void) {
	struct usbbus* bus;
	
	bus = usbroot;
	usbroot = NULL;
	pthread_mutex_init(&usbdevlock, NULL);
	usbcontrolForkCleanup();
	while (bus) {
		struct usbbus* cbus;
		struct usbdev* dev;

		cbus = bus;
		bus = bus->next;
		
		usb_log("Clearing bus %p\n", cbus);
		dev = cbus->devlist;
		usb_log("1st Device: %p\n", dev);
		while (dev) {
			struct usbdev* cdev;
			
			cdev = dev;
			dev = dev->next;
			usb_log("Clearing device %p\n", cdev);
			if (cdev->fns->USBForkCleanup) {
				cdev->fns->USBForkCleanup(cdev->cookie);
			}
			free(cdev);
		}
		usb_log("Trying free %p\n", cbus);
		free(cbus);
		usb_log("Free done\n");
	}
}

static const char* getbid(const char* filename, int* busid, int* devid) {
	char* e;
	char* f;
	int bus, dev;
	
	*devid = -1;
	*busid = -1;
	if (!isdigit(filename[0])) {
		return filename;
	}
	bus = strtoul(filename, &e, 10);
	if (*e == 0) {
		*busid = bus;
		return "";
	}
	if (*e != '/') {
		return filename;
	}
	*busid = bus;
	if (!isdigit(e[1])) {
		return e + 1;
	}
	dev = strtoul(e + 1, &f, 10);
	if (*f == 0) {
		*devid = dev;
		return "";
	}
	if (*f != '/') {
		return e + 1;
	}
	*devid = dev;
	return f + 1;
}

static int check_open(const char* filename, int flags, mode_t mode) {
	if (filename && !strncmp(filename, pbud, pbud_len)) {
		const char* f2 = filename + pbud_len;

		usb_log("open(%s, 0x%08X, 0x%08X)\n", filename, flags, mode);

		if (!strcmp(f2, "devices")) {
			int r = __next_open(filename, flags, mode);
			if (r != -1) {
				struct mystr* p;
				struct usbfs_devices_info* info;
				
				info = usbfs_devices_init(r);
				if (!info) {
					__next_close(r);
					errno = ENOMEM;
					return -1;
				}
				p = newfd(r, info, &fd_devices);
				insertspecialfd(p);
			}
			return r;
		} else {
			const char* rest;
			int busid, devid;
			struct usbdev* dev;
			
			rest = getbid(f2, &busid, &devid);
			if (!*rest && (dev = usbdev_exist(busid, devid)) != NULL) {
				char xx[] = "/tmp/usbXXXXXX";
				int r;

				r = mkstemp(xx);
				if (r != -1) {
					struct opninfo* op = malloc(sizeof(*op));
					struct mystr* p;
					int sv[2];
					const int zero = 0;
					
					unlink(xx);
					op->file = r;
					socketpair(PF_UNIX, SOCK_STREAM, 0, sv);
					fcntl(sv[0], F_SETFL, O_NONBLOCK);
					fcntl(sv[1], F_SETFL, O_NONBLOCK);
					fcntl(sv[1], F_SETFD, FD_CLOEXEC);
					fcntl(r, F_SETFD, FD_CLOEXEC);
					setsockopt(sv[0], SOL_SOCKET, SO_SNDBUF, &zero, sizeof(zero));
					setsockopt(sv[1], SOL_SOCKET, SO_SNDBUF, &zero, sizeof(zero));
					while (write(sv[0], xx, sizeof(xx)) == sizeof(xx)) ;
					op->pipe0 = sv[0];
					op->pipe1 = sv[1];
					op->dev = dev;
					op->urb = NULL;
					init_usb_data(dev, op->file);
					p = newfd(op->pipe0, op, &fd_devnode);

					insertspecialfd(p);
					r = op->pipe0;
				}
				return r;
			}
		}
	}
	return USE_NEXT_FN;
}

static off64_t check_lseek(int fildes, off64_t offset, int whence) {
	struct mystr* ms;
	off64_t ret = USE_NEXT_FN;
	
	ms = isspecialfd(fildes);
	if (ms) {
		usb_log("lseek(%u, 0x%llX, %u)\n", fildes, offset, whence);
		if (ms->fns->lseek) {
			ret = ms->fns->lseek(ms->cookie, offset, whence);
		}
		releasefd(ms);
	}
	return ret;
}

static int check_read(int fildes, void* buffer, size_t size) {
	struct mystr* ms;
	int ret = USE_NEXT_FN;
	
	ms = isspecialfd(fildes);
	if (ms) {
		usb_log("read(%u, %p, %u)\n", fildes, buffer, size);
		if (ms->fns->read) {
			ret = ms->fns->read(ms->cookie, buffer, size);
		}
		releasefd(ms);
	}
	return ret;
}

static int check_write(int fildes, const void* buffer, size_t size) {
	struct mystr* ms;
	int ret = USE_NEXT_FN;
	
	ms = isspecialfd(fildes);
	if (ms) {
		usb_log("write(%u, %p, %u)\n", fildes, buffer, size);
		if (ms->fns->write) {
			ret = ms->fns->write(ms->cookie, buffer, size);
		}
		releasefd(ms);
	}
	return ret;
}

static int check_ioctl(int fildes, unsigned long int command, void* param) {
	struct mystr* ms;
	int ret = USE_NEXT_FN;
	
	ms = isspecialfd(fildes);
	if (ms) {
		usb_log("ioctl(%u, 0x%08lX, %p)\n", fildes, command, param);
		if (ms->fns->ioctl) {
			ret = ms->fns->ioctl(ms->cookie, command, param);
		}		
		releasefd(ms);
	}
	return ret;
}

static int check_close(int fildes) {
	struct mystr* ms;
	
	ms = removespecialfd(fildes);
	if (ms) {
		usb_log("close(%u)\n", fildes);
		releasefd(ms);
	}
	return USE_NEXT_FN;
}

struct devinfo {
	FILE* f;
	int realEof;
	int ttlpos;
	size_t posInFD;
	size_t maxInFD;
	char* fakeData;
};

static void deio_init(struct devinfo* di, FILE* f) {
	di->f = f;
	di->realEof = 0;
	di->posInFD = 0;
	di->ttlpos = 0;
	di->fakeData = usbdev_allstrings(&di->maxInFD);
}

static ssize_t deio_reader(void* cookie, char* buffer, size_t size) {
	struct devinfo* di = cookie;
	int rdb = 0;
	
	usb_log("deio_reader(%p, %p, %u)\n", di, buffer, size);
	if (!di->realEof) {
		rdb = fread(buffer, 1, size, di->f);
		if (rdb < 0) {
			return rdb;
		}
		if ((size_t)rdb >= size) {
			di->ttlpos += rdb;
			return rdb;
		}
		di->ttlpos += rdb;
		buffer += rdb;
		size -= rdb;
		di->realEof = 1;
		di->posInFD = 0;
	}
	if (size > di->maxInFD - di->posInFD) {
		size = di->maxInFD - di->posInFD;
	}
	memcpy(buffer, di->fakeData + di->posInFD, size);
	di->posInFD += size;
	return rdb + size;
}

static ssize_t deio_writer(void* cookie, const char* buffer, size_t size) {
	struct devinfo* di = cookie;
	
	usb_log("deio_writer(%p, %p, %u)\n", di, buffer, size);
	return 0;
}

static int deio_seeker(void* cookie, off64_t* position, int whence) {
	struct devinfo* di = cookie;
	
	usb_log("deio_seeker(%p, %lld, %u)\n", di, *position, whence);
	errno = ESPIPE;
	return -1;
}

static int deio_closer(void* cookie) {
	struct devinfo* di = cookie;

	usb_log("deio_closer(%p)\n", di);
	fclose(di->f);
	free(di->fakeData);
	free(di);
	return 0;
}

static const cookie_io_functions_t deiofns = {
	.read = deio_reader,
	.write = deio_writer,
	.seek = deio_seeker,
	.close = deio_closer,
};

static int check_fopen(FILE** pf, const char* filename, const char* mode) {
	if (filename && !strncmp(filename, pbud, pbud_len)) {
		const char* f2 = filename + pbud_len;

		usb_log("fopen(%s, %s)\n", filename, mode);

		if (!strcmp(f2, "devices")) {
			FILE* q;
			
			q = __next_fopen(filename, mode);
			if (q) {
				struct devinfo* di;
				
				di = malloc(sizeof(*di));
				deio_init(di, q);
				*pf = fopencookie(di, mode, deiofns);
				usb_log("... pf=%p, cookie=%p\n", *pf, di);
				if (!*pf) {
					deio_closer(di);
				}
				return 0;
			}
			*pf = NULL;
			return 0;
		} else {
			const char* rest;
			int busid, devid;
			
			rest = getbid(f2, &busid, &devid);
			if (!*rest && usbdev_exist(busid, devid)) {
				///
			}
		}
	}
	return USE_NEXT_FN;
}

int open(const char* filename, int flags, ...) {
	mode_t mode = 0;
	int r;
	
	if (flags & O_CREAT) {
		va_list va;

		va_start(va, flags);
		mode = va_arg(va, mode_t);
		va_end(va);
	}
	r = check_open(filename, flags, mode);
	if (r == USE_NEXT_FN) {
		return __next_open(filename, flags, mode);
	}
	return r;
}

int open64(const char* filename, int flags, ...) {
	mode_t mode = 0;
	int r;

	if (flags & O_CREAT) {
		va_list va;

		va_start(va, flags);
		mode = va_arg(va, mode_t);
		va_end(va);
	}
	r = check_open(filename, flags | O_LARGEFILE, mode);
	if (r == USE_NEXT_FN) {
		return __next_open64(filename, flags, mode);
	}
	return r;
}

off_t lseek(int fildes, off_t offset, int whence) {
	off64_t r;
	
	r = check_lseek(fildes, offset, whence);
	if (r == USE_NEXT_FN) {
		return __next_lseek(fildes, offset, whence);
	}
	return r;
}

off64_t lseek64(int fildes, off64_t offset, int whence) {
	off64_t r;
	
	r = check_lseek(fildes, offset, whence);
	if (r == USE_NEXT_FN) {
		return __next_lseek64(fildes, offset, whence);
	}
	return r;
}

int read(int fildes, void* buffer, size_t size) {
	int r;
	
	r = check_read(fildes, buffer, size);
	if (r == USE_NEXT_FN) {
		return __next_read(fildes, buffer, size);
	}
	return r;
}

int write(int fildes, const void* buffer, size_t size) {
	int r;
	
	r = check_write(fildes, buffer, size);
	if (r == USE_NEXT_FN) {
		return __next_write(fildes, buffer, size);
	}
	return r;
}

int ioctl(int fildes, unsigned long int command, ...) {
	void* arg;
	va_list va;
	int r;

	va_start(va, command);
	arg = va_arg(va, void*);
	va_end(va);
	r = check_ioctl(fildes, command, arg);
	if (r == USE_NEXT_FN) {
		return __next_ioctl(fildes, command, arg);
	}
	return r;
}

int close(int fildes) {
	int r;
	
	r = check_close(fildes);
	if (r == USE_NEXT_FN) {
		return __next_close(fildes);
	}
	return r;
}

FILE* fopen(const char* path, const char* mode) {
	FILE* f;
	
	if (check_fopen(&f, path, mode) == USE_NEXT_FN) {
		return __next_fopen(path, mode);
	}
	return f;
}

FILE* fopen64(const char* path, const char* mode) {
	FILE* f;
	
	if (check_fopen(&f, path, mode) == USE_NEXT_FN) {
		return __next_fopen64(path, mode);
	}
	return f;
}

static pthread_mutex_t dirop_lock = PTHREAD_MUTEX_INITIALIZER;

struct dirop_fn {
	int (*readdir64)(void* cookie, struct dirent64** de);
	void (*cleanup)(void* cookie);
};

struct dirop_usb {
	DIR* dirp;
	int dofake;
};

static int dirop_usb_readdir64(void* cookie, struct dirent64** de) {
	struct dirop_usb* du = cookie;
	struct dirent64* d;

	if (du->dirp) {
		d = __next_readdir64(du->dirp);
	} else {
		d = NULL;
	}
	if (!d && du->dofake >= 0) {
		static struct dirent64 dp;
		int bnum;
		
		bnum = usbbus_next(du->dofake);
		if (bnum < 0) {
			du->dofake = -1;
		} else {
			du->dofake = bnum + 1;
		}

		if (bnum >= 0) {
			dp.d_ino = 0xDEADBEAF;
			dp.d_off = 0xDEADBEAF;
			sprintf(dp.d_name, "%03u", bnum);
			dp.d_reclen = strlen(dp.d_name) + offsetof(struct dirent64, d_name) + 1;
			dp.d_type = DT_DIR;		
			d = &dp;
		} else {
			d = NULL;
		}
	}
	*de = d;
	return 0;
}

static void dirop_usb_cleanup(void* cookie) {
	struct dirop_usb* du = cookie;
	
	free(du);
}

static struct dirop_fn dirop_usb_ops = {
	.readdir64 = dirop_usb_readdir64,
	.cleanup = dirop_usb_cleanup,
};

struct dirop_devd {
	int pos;
	int busid;
	struct dirent64 de;
};

static int dirop_devd_readdir64(void* cookie, struct dirent64** de) {
	struct dirop_devd* dd = cookie;
	
	switch (dd->pos) {
		case 0:
			strcpy(dd->de.d_name, ".");
			dd->de.d_type = DT_DIR;
			break;
		case 1:
			strcpy(dd->de.d_name, "..");
			dd->de.d_type = DT_DIR;
			break;
		default:
			{
				int devid = usbdev_getidx(dd->busid, dd->pos - 2);
				
				if (devid < 0) {
					*de = NULL;
					return 0;
				}
				sprintf(dd->de.d_name, "%03u", devid);
			}
			dd->de.d_type = DT_REG;
			break;
	}
	dd->pos++;
	dd->de.d_ino = 0xDEADBEAF;
	dd->de.d_off = dd->pos;
	dd->de.d_reclen = strlen(dd->de.d_name) + offsetof(struct dirent64, d_name) + 1;
	*de = &dd->de;
	return 0;
}	
	
static void dirop_devd_cleanup(void* cookie) {
	struct dirop_devd* dd = cookie;
	
	free(dd);
}

static struct dirop_fn dirop_devd_ops = {
	.readdir64 = dirop_devd_readdir64,
	.cleanup = dirop_devd_cleanup,
};

struct dirop_info {
	struct dirop_info* next;
	DIR* dirp;
	int usecnt;
	
	void* cookie;
	const struct dirop_fn* dfn;
};

static struct dirop_info* dirop_list;

static void dirop_insert(DIR* dirp, void* cookie, const struct dirop_fn* dfn) {
	struct dirop_info* di;
	
	di = malloc(sizeof(*di));
	di->next = NULL;
	di->dirp = dirp;
	di->usecnt = 1;
	di->cookie = cookie;
	di->dfn = dfn;
	
	pthread_mutex_lock(&dirop_lock);
	di->next = dirop_list;
	dirop_list = di;
	pthread_mutex_unlock(&dirop_lock);
}

static struct dirop_info* dirop_isspecial(DIR* dirp) {
	struct dirop_info* cp;

	pthread_mutex_lock(&dirop_lock);
	for (cp = dirop_list; cp; cp = cp->next) {
		if (cp->dirp == dirp) {
			cp->usecnt++;
			pthread_mutex_unlock(&dirop_lock);
			return cp;
		}
	}
	pthread_mutex_unlock(&dirop_lock);
	return NULL;
}

static struct dirop_info* dirop_closing(DIR* dirp) {
	struct dirop_info** c2;
	struct dirop_info* cp;
	
	pthread_mutex_lock(&dirop_lock);
	for (c2 = &dirop_list; (cp = *c2) != NULL; c2 = &cp->next) {
		if (cp->dirp == dirp) {
			*c2 = cp->next;
			cp->next = NULL;
			pthread_mutex_unlock(&dirop_lock);
			return cp;
		}
	}
	pthread_mutex_unlock(&dirop_lock);
	return NULL;
}

static void dirop_release(struct dirop_info* di) {
	if (--di->usecnt == 0) {
		if (di->dfn->cleanup)
			di->dfn->cleanup(di->cookie);
		free(di);
	}
}
	
DIR* opendir(const char* filename) {
	if (filename && !strncmp(filename, pbud, pbud_len - 1)) {
		const char* f2 = filename + pbud_len - 1;

		usb_log("opendir(%s)\n", filename);

		if (!strcmp(f2, "") || !strcmp(f2, "/")) {
			DIR* dirp;
			
			dirp = __next_opendir(filename);
			if (dirp) {
				struct dirop_usb* du;
				
				du = malloc(sizeof(*du));
				du->dirp = dirp;
				du->dofake = 0;
				dirop_insert(dirp, du, &dirop_usb_ops);
			} else if (errno == ENOENT) {
				struct dirop_usb* du;

				dirp = __next_opendir("/");
				if (!dirp) {
					return dirp;
				}
				du = malloc(sizeof(*du));
				du->dirp = NULL;
				du->dofake = 0;
				dirop_insert(dirp, du, &dirop_usb_ops);
			}
			return dirp;
		} else if (f2[0] == '/') {
			const char* rest;
			int busid, devid;
			
			rest = getbid(f2 + 1, &busid, &devid);
			if (*rest == 0 && usbbus_exist(busid) && devid == -1) {
				DIR* dirp;

				dirp = __next_opendir(".");
				if (dirp) {
					struct dirop_devd* du;

					du = malloc(sizeof(*du));
					du->pos = 0;
					du->busid = busid;
					dirop_insert(dirp, du, &dirop_devd_ops);
				}
				return dirp;
			}
		} else {
		}
	}
	return __next_opendir(filename);
}

struct dirent* readdir(DIR* dirp) {
	struct dirop_info* di;
	
	di = dirop_isspecial(dirp);
	if (di) {
		int r = USE_NEXT_FN;
		struct dirent64* de64 = NULL;
		
		if (di->dfn->readdir64) {
			r = di->dfn->readdir64(di->cookie, &de64);
		}
		dirop_release(di);
		if (r != USE_NEXT_FN) {
			if (de64) {
				static struct dirent de;

				dirent64_to_dirent(&de, de64);
				return &de;
			}
			return NULL;
		}
	}
	return __next_readdir(dirp);	
}

struct dirent64* readdir64_22(DIR* dirp) {
	struct dirop_info* di;
	
	di = dirop_isspecial(dirp);
	if (di) {
		int r = USE_NEXT_FN;
		struct dirent64* de = NULL;
		
		if (di->dfn->readdir64) {
			r = di->dfn->readdir64(di->cookie, &de);
		}
		dirop_release(di);
		if (r != USE_NEXT_FN) {
			return de;
		}
	}
	return __next_readdir64(dirp);
}
asm(".symver readdir64_22, readdir64@GLIBC_2.2");

struct dirent64_21* readdir64_21(DIR* dirp) {
	struct dirop_info* di;
	
	di = dirop_isspecial(dirp);
	if (di) {
		int r = USE_NEXT_FN;
		struct dirent64* de = NULL;
		
		if (di->dfn->readdir64) {
			r = di->dfn->readdir64(di->cookie, &de);
		}
		dirop_release(di);
		if (r != USE_NEXT_FN) {
			if (de) {
				static struct dirent64_21 de21;
				
				dirent64_to_dirent64_21(&de21, de);
				return &de21;
			}
			return NULL;
		}
	}
	return __next_readdir64_21(dirp);
}

asm(".symver readdir64_21, readdir64@GLIBC_2.1");

void rewinddir(DIR* dirp) {
	struct dirop_info* di;
	
	di = dirop_isspecial(dirp);
	if (di) {
		///
		dirop_release(di);
	}
	__next_rewinddir(dirp);
}

void seekdir(DIR* dirp, long int offset) {
	struct dirop_info* di;
	
	di = dirop_isspecial(dirp);
	if (di) {
		///
		dirop_release(di);
	}
	__next_seekdir(dirp, offset);
}

long int telldir(DIR* dirp) {
	struct dirop_info* di;
	
	di = dirop_isspecial(dirp);
	if (di) {
		///
		dirop_release(di);
	}
	return __next_telldir(dirp);
}

int closedir(DIR* dirp) {
	struct dirop_info* di;
	
	di = dirop_closing(dirp);
	if (di) {
		///
		dirop_release(di);
	}
	return __next_closedir(dirp);
}


static void findsym(void* dstp, const void* disallowed, const char* sym) {
	void* ptr;
	void** dst = dstp;
	
	*dst = NULL;
	ptr = dlsym(RTLD_NEXT, sym);
	if (ptr && ptr != disallowed) {
		*dst = ptr;
	}
	if (*dst == NULL) {
		fprintf(stderr, "Unable to get %s...\n", sym);
		exit(0);
	}
}

static void findvsym(void* dstp, const void* disallowed, const char* sym, const char* version) {
	void* ptr;
	void** dst = dstp;
	
	*dst = NULL;
	ptr = dlvsym(RTLD_NEXT, sym, version);
	if (ptr && ptr != disallowed) {
		*dst = ptr;
	}
	if (*dst == NULL) {
		fprintf(stderr, "Unable to get %s@%s...\n", sym, version);
		exit(0);
	}
}

static int isVMX = 0;
static char buf[4096];

static void initVmxGlue(void) {
	FILE* f;
	size_t bl = 0;
	struct stat stb;
	
	f = __next_fopen64("/proc/self/cmdline", "rb");
	if (!f) {
		/* Failed to open cmdline? It cannot be vmx... */
		return;
	}
	while (fread(buf + bl, 1, 1, f) == 1) {
		if (buf[bl] == 0) {
			break;
		}
		if (buf[bl] == '/') {
			bl = 0;
			continue;
		}
		bl++;
		if (bl >= sizeof(buf) - 1) {
			break;
		}
	}
	buf[bl] = 0;
	if (memcmp(buf, "vmware-vmx", 10)) {
		/* It is not vmware-vmx... */
		fclose(f);
		return;
	}
	bl = 0;
	while (fread(buf + bl, 1, 1, f) == 1) {
		if (buf[bl] == 0) {
			bl = 0;
			continue;
		}
		if (bl < sizeof(buf) - 1) {
			bl++;
		}
	}
	fclose(f);
	if (bl != 0 || buf[0] != '/') {
		/* Last argument did not start with slash */
		return;
	}
	if (access(buf, R_OK)) {
		/* No access to file?! */
		return;
	}
	if (stat(buf, &stb)) {
		/* File not found */
		return;
	}
	if (!S_ISREG(stb.st_mode)) {
		/* Not regular file */
		return;
	}
	/* Fine, that's us... */
	isVMX = 1;
}

const char* vmxGetConfig(void) {
	return isVMX ? buf : NULL;
}

static void __attribute__((constructor)) init(void) {
	const char* dbg;
	uid_t uid;

	realUserId = getuid();

	uid = becomeNormalUser();
	dbg = getenv("PETRUSB_DEBUG");
	if (dbg && !strcmp(dbg, "yes")) {
		debug = 1;
	} else {
		debug = 0;
	}
	findsym(&__next_open,		open,		"open");
	findsym(&__next_open64,		open64,		"open64");
	findsym(&__next_lseek,		lseek,		"lseek");
	findsym(&__next_lseek64,	lseek64,	"lseek64");
	findsym(&__next_read,		read,		"read");
	findsym(&__next_write,		write,		"write");
	findsym(&__next_close,		close,		"close");
	findsym(&__next_ioctl,		ioctl,		"ioctl");
	findsym(&__next_fopen,		fopen,		"fopen");
	findsym(&__next_fopen64,	fopen64,	"fopen64");
	findsym(&__next_opendir,	opendir,	"opendir");
	findsym(&__next_readdir,	readdir,	"readdir");
	findvsym(&__next_readdir64_21,	readdir64_21,	"readdir64", "GLIBC_2.1");
	findvsym(&__next_readdir64,	readdir64_22,	"readdir64", "GLIBC_2.2");
	findsym(&__next_telldir,	telldir,	"telldir");
	findsym(&__next_seekdir,	seekdir,	"seekdir");
	findsym(&__next_rewinddir,	rewinddir,	"rewinddir");
	findsym(&__next_closedir,	closedir,	"closedir");

	initVmxGlue();

	if (debug) {
		fo = __next_fopen("/tmp/usb.log", "at");
	}
	pthread_atfork(NULL, NULL, childCleanup);

	{
		uid_t ruid, euid, suid;

		getresuid(&ruid, &euid, &suid);
		usb_log("ruid=%u, euid=%u, suid=%u [buid=%u]\n", ruid, euid, suid, uid);
	}

	usbcontrolInit();
	
	readFile(vmxGetConfig());
	restoreNormalUser(uid);
}

