#define _GNU_SOURCE
#define _LARGEFILE_SOURCE
#define _LARGEFILE64_SOURCE

#include "usbcontrol.h"

#include "dictionary.h"

#include "usbdev.h"

#include "vmnetDevice.h"

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <pthread.h>

#include <sys/un.h>
#include <sys/socket.h>
#include <sys/poll.h>
#include <sys/ioctl.h>

#include <net/if.h>

#include "vnet.h"

#define MAX_USBCDC_DEV		(7)

struct usbcdcdev;

struct myusbdev {
	struct usbdevice	usbdevice;
	struct usbconfig	usbconfig;
	struct usblogdev	logdev;
	unsigned int		id;
	int			motorola;
	unsigned int		devicesCount;
	struct usbcdcdev*	devices[MAX_USBCDC_DEV];
};

struct usbcdcxfer {
	struct queuedXfer	xfer;
	size_t			bufferLen;
	unsigned char		buffer[1536];
};

struct usbcdcdevCfg {
	char*			fileName;
	int			startConnected;
};

struct usbcdcdev {
	struct myusbdev*	parent;
	char*			configKey;
	struct usbinterface	controlUSBInterface;
	struct usbendpoint	controlUSBIntrEP;
	struct usbinterface	data0USBInterface;
	struct usbinterface	data1USBInterface;
	struct usbendpoint	data1USBBulkInEP;
	struct usbendpoint	data1USBBulkOutEP;
	struct usblogdev	logdev;
	struct vmnetDevice	vmnet;
	unsigned char		mac[6];
	int			macFromCfg;
	int			iControlInterface;
	int			iDataInterface;
	int			iEthernetMAC;
	int			ifnum;
	struct usbcdcdevCfg	cfg;
	enum { DOWN=0, RUNNING=1 } state;
	unsigned int		ifflags;
};

static int
controlUSBInterfaceCmd(void* cookie,
		       struct usbdevfs_ctrltransfer* ct)
{
	return -EPIPE;
}

static int 
controlUSBInterfaceGetDescriptor(void* cookie,
				 struct usb_interface_descriptor* uid)
{
	struct usbcdcdev* ubd = cookie;

        uid->bInterfaceClass = 0x02;
       	uid->bInterfaceSubClass = 0x06;
        uid->bInterfaceProtocol = 0x00;
        uid->iInterface = ubd->iControlInterface;
	return 0;
}

static int
controlUSBInterfaceGetFunctionalDescriptor(void* cookie,
					   unsigned char* buffer,
					   unsigned int idx)
{
	struct usbcdcdev* ubd = cookie;

	switch (idx) {
		case 0:
			buffer[1] = USB_DT_CS_INTERFACE;
			buffer[2] = 0x00;
			setLE16(buffer + 3, 0x0110);	/* Standard version */
			return 5;
		case 1:
			buffer[1] = USB_DT_CS_INTERFACE;
			buffer[2] = 0x06;
			buffer[3] = ubd->controlUSBInterface.bInterfaceNumber;
			buffer[4] = ubd->data0USBInterface.bInterfaceNumber;
			return 5;
		case 2:
			buffer[1] = USB_DT_CS_INTERFACE;
			buffer[2] = 0x0F;
			buffer[3] = ubd->iEthernetMAC;		/* MAC Address */
			setLE32(buffer + 4, 0x00000000);	/* Ethernet statistics: NONE */
			setLE16(buffer + 8, 1514);		/* Max. packet size without CRC */
			setLE16(buffer + 10, 0x0000);		/* Perfect filtering, zero entries */
			buffer[12] = 0;				/* Number of power wakeup events */
			return 13;
		default:
			return -ENOENT;
	}
}

static struct usbinterface_fns controlUSBInterface_fns = {
	.intfDeviceControl = controlUSBInterfaceCmd,
	.intfGetDescriptor = controlUSBInterfaceGetDescriptor,
	.intfGetFunctionalDescriptor = controlUSBInterfaceGetFunctionalDescriptor,
};

static int
data1USBInterfaceCopyData(struct queuedXfer* cookie,
			  size_t len,
			  void* buffer)
{
	struct usbcdcxfer* xfer = list_entry(cookie, struct usbcdcxfer, xfer);
	
	if (len > xfer->bufferLen) {
		len = xfer->bufferLen;
	}
	memcpy(buffer, xfer->buffer, len);
	free(xfer);
	usb_log("Returned %u buffer\n", len);
	return len;
}

static int
data1USBInterfaceDiscardData(struct queuedXfer* cookie)
{
	struct usbcdcxfer* xfer = list_entry(cookie, struct usbcdcxfer, xfer);
	
	usb_log("Discarding received buffer %p\n", xfer);
	free(xfer);
	return 0;
}

static int
data1USBInterfaceDataReady(void* cookie,
			   int fd,
			   int revents)
{
	int r;
	struct usbcdcdev* ubd = cookie;
	struct usbcdcxfer* xfer;	

	if (ubd->state != RUNNING) {
		char dummy[1536];
		
		usb_log("interface down, throwing packet away\n");
		read(fd, dummy, sizeof(dummy));
		return 0;
	}
	xfer = malloc(sizeof(*xfer));
	if (!xfer) {
		char dummy[1536];

		usb_log("malloc failed, throwing packet away\n");
		read(fd, dummy, sizeof(dummy));
		return 0;
	}
	r = read(fd, xfer->buffer, sizeof(xfer->buffer));
	usb_log("Received %d bytes from network\n", r);
	if (r > 12) {
		xfer->bufferLen = r;
		xfer->xfer.xfer = data1USBInterfaceCopyData;
		xfer->xfer.abort = data1USBInterfaceDiscardData;
		r = usbendpointQueueIACK(&ubd->data1USBBulkInEP, &xfer->xfer);
		if (r < 0) {
			usb_log("Queued IACK failed: %d\n", r);
			free(xfer);
		}
	} else {
		free(xfer);
	}
	return 0;
}

static void
usbcdcUpdateMAC(struct usbcdcdev* ubd)
{
	char c[40];
	int res;
	
	sprintf(c, "%02X%02X%02X%02X%02X%02X", ubd->mac[0], ubd->mac[1], ubd->mac[2],
					       ubd->mac[3], ubd->mac[4], ubd->mac[5]);
	res = usbdeviceModifyString(&ubd->parent->usbdevice, ubd->iEthernetMAC, c);
	usb_log("Updated MAC address #%u (to %s with result %d)\n", ubd->iEthernetMAC, c, res);
}

static void
usbcdcSetIFFlags(struct usbcdcdev* ubd,
		 unsigned int ifflags)
{
	ubd->ifflags = ifflags;
	if (vmnetIsConnected(&ubd->vmnet)) {
		vmnetSetIFFlags(&ubd->vmnet, ifflags);
	}
}

static void
usbcdcStartNetwork(struct usbcdcdev* ubd)
{
	if (!vmnetIsConnected(&ubd->vmnet)) {
		if (vmnetConnect(&ubd->vmnet, ubd->cfg.fileName) == 0) {
			if (ubd->macFromCfg) {
				if (vmnetSetMACAddress(&ubd->vmnet, ubd->mac)) {
					usb_log("Failed to set requested MAC address %02X%02X%02X:%02X%02X%02X\n",
							ubd->mac[0], ubd->mac[1], ubd->mac[2],
							ubd->mac[3], ubd->mac[4], ubd->mac[5]);
					ubd->macFromCfg = 0;
				}
			}
			vmnetSetIFFlags(&ubd->vmnet, ubd->ifflags);
			if (!ubd->macFromCfg) {
				if (vmnetGetMACAddress(&ubd->vmnet, ubd->mac) >= 0) {
					char c[40];
					
					usbcdcUpdateMAC(ubd);
					sprintf(c, "%02x:%02x:%02x:%02x:%02x:%02x", ubd->mac[0], ubd->mac[1], ubd->mac[2],
										    ubd->mac[3], ubd->mac[4], ubd->mac[5]);
					dictSetString(c, "%s.generatedAddress", ubd->configKey);
					dictSetString("generated", "%s.addressType", ubd->configKey);
					ubd->macFromCfg = 1;
				}
			}
			usbcontrolStartWatchingFileDescriptor(ubd->vmnet.fd, ubd, data1USBInterfaceDataReady, POLLIN);
		}
	}	
}

static void
usbcdcStopNetwork(struct usbcdcdev* ubd)
{
	if (vmnetIsConnected(&ubd->vmnet)) {
		usbcontrolStopWatchingFileDescriptor(ubd->vmnet.fd);
		vmnetDisconnect(&ubd->vmnet);
	}
}

static int
data0USBInterfaceCmd(void* cookie,
		     struct usbdevfs_ctrltransfer* ct)
{
	return -EPIPE;
}

static int 
data0USBInterfaceGetDescriptor(void* cookie,
			       struct usb_interface_descriptor* uid)
{
	struct usbcdcdev* ubd = cookie;

        uid->bInterfaceClass = 0x0A;
        uid->bInterfaceSubClass = 0x00;
        uid->bInterfaceProtocol = 0x00;
        uid->iInterface = ubd->iDataInterface;
	return 0;
}

static void
data0USBInterfaceSelected(void* cookie)
{
	struct usbcdcdev* ubd = cookie;
	
	ubd->state = DOWN;
	usbendpointOStall(&ubd->data1USBBulkOutEP);
	usbendpointIStall(&ubd->data1USBBulkInEP);
	usbcdcSetIFFlags(ubd, 0);
}

static struct usbinterface_fns data0USBInterface_fns = {
	.intfDeviceControl = data0USBInterfaceCmd,
	.intfGetDescriptor = data0USBInterfaceGetDescriptor,
	.intfSelected	   = data0USBInterfaceSelected,
};

static int
data1USBInterfaceCmd(void* cookie,
		     struct usbdevfs_ctrltransfer* ct)
{
	return -EPIPE;
}

static int 
data1USBInterfaceGetDescriptor(void* cookie,
			       struct usb_interface_descriptor* uid)
{
	struct usbcdcdev* ubd = cookie;

        uid->bInterfaceClass = 0x0A;
        uid->bInterfaceSubClass = 0x00;
        uid->bInterfaceProtocol = 0x00;
        uid->iInterface = ubd->iDataInterface;
	return 0;
}

static void
data1USBInterfaceSelected(void* cookie)
{
	struct usbcdcdev* ubd = cookie;

	usbendpointOACK(&ubd->data1USBBulkOutEP);
	usbendpointINAK(&ubd->data1USBBulkInEP);
	ubd->state = RUNNING;
	usbcdcSetIFFlags(ubd, IFF_UP | IFF_RUNNING | IFF_BROADCAST | IFF_ALLMULTI);
}

static struct usbinterface_fns data1USBInterface_fns = {
	.intfDeviceControl = data1USBInterfaceCmd,
	.intfGetDescriptor = data1USBInterfaceGetDescriptor,
	.intfSelected	   = data1USBInterfaceSelected,
};

static int
data1USBInterfaceSend(void* cookie,
                      size_t len,
                      const void* buffer)
{
	struct usbcdcdev* ubd = cookie;
	
	if (ubd->state == RUNNING) {
		vmnetSendPacket(&ubd->vmnet, buffer, len);
	}
	return len;
}

static void
getMAC(struct usbcdcdev* ubd,
       int offset)
{
	const char* addrType;
	const char* addr;

	memcpy(ubd->mac, "\x00\x50\x56\xFF\xFF\x00", 6);
	ubd->mac[5] += offset;
	ubd->macFromCfg = 0;

	addrType = dictGetString("generated", "%s.addressType", ubd->configKey);
	if (addrType && !strcmp(addrType, "static")) {
		addr = dictGetString(NULL, "%s.address", ubd->configKey);
	} else {
		addr = dictGetString(NULL, "%s.generatedAddress", ubd->configKey);
	}
	if (addr) {
		unsigned int c[6];
		
		if (sscanf(addr, "%02x:%02x:%02x:%02x:%02x:%02x",
			   &c[0], &c[1], &c[2], &c[3], &c[4], &c[5]) == 6) {
			if (c[0] <= 0xFF && (c[0] & 1) == 0 &&
			    c[1] <= 0xFF &&
			    c[2] <= 0xFF &&
			    c[3] <= 0xFF &&
			    c[4] <= 0xFF &&
			    c[5] <= 0xFF) {
				ubd->mac[0] = c[0];
				ubd->mac[1] = c[1];
				ubd->mac[2] = c[2];
				ubd->mac[3] = c[3];
				ubd->mac[4] = c[4];
				ubd->mac[5] = c[5];
				ubd->macFromCfg = 1;
				usbcdcUpdateMAC(ubd);
			}
		}
	}
}

static void
usbcdcForkCleanup(struct usblogdev* ul)
{
	struct usbcdcdev* ubd = list_entry(ul, struct usbcdcdev, logdev);

	/* Close opened fd... */
	usbcdcStopNetwork(ubd);
}

static int
usbcdcGetName(struct usblogdev* ul,
              unsigned char* buffer)
{
	struct usbcdcdev* ubd = list_entry(ul, struct usbcdcdev, logdev);

        setBE16(buffer, sprintf(buffer + 2, "USB CDC Ethernet #%u", ubd->ifnum));
        return getBE16(buffer) + 2;
}

static int
usbcdcGetSettings(struct usblogdev* ul,
                  unsigned char* buffer)
{
        struct usbcdcdev* ubd = list_entry(ul, struct usbcdcdev, logdev);
        unsigned char* ptr;

        ptr = stuffString(buffer, "CDCEthernet");
        ptr = stuffBoolean(ptr, 1, "startConnected", ubd->cfg.startConnected);
        ptr = stuffFileName(ptr, 1, "fileName", ubd->cfg.fileName);
	ptr = stuffBoolean(ptr, 1, "connected", vmnetIsConnected(&ubd->vmnet));
	ptr = stuffBoolean(ptr, 0, "running", ubd->state == RUNNING);
        return ptr - buffer;
}

static int
usbcdcSetSettings(struct usblogdev* ul,
                  const unsigned char* inp,
                  size_t inplen,
                  unsigned char* buffer)
{
        struct usbcdcdev* ubd = list_entry(ul, struct usbcdcdev, logdev);
        const unsigned char* ptr;
        const unsigned char* key;
        const unsigned char* val;
        unsigned char type;
	int oldConnected;
	int newConnected;
	int newStartConnected;
	char* newFileName;

        newStartConnected = ubd->cfg.startConnected;
        newFileName = strdup(ubd->cfg.fileName);
	newConnected = oldConnected = vmnetIsConnected(&ubd->vmnet);
        ptr = inp;
        while ((ptr = fetchChange(ptr, inp + inplen, &key, &type, &val)) != NULL) {
		if (!keycmp(key, "startConnected")) {
			retrieveBoolean(&newStartConnected, type, val);
		} else if (!keycmp(key, "connected")) {
			retrieveBoolean(&newConnected, type, val);
                } else if (!keycmp(key, "fileName")) {
			retrieveString(&newFileName, type, val);
                }
        }
	if (ubd->cfg.startConnected != newStartConnected) {
		dictSetBoolean(newStartConnected, "%s.startConnected", ubd->configKey);
		ubd->cfg.startConnected = newStartConnected;
		ul->seq++;
	}
        if (strcmp(newFileName, ubd->cfg.fileName) || oldConnected != newConnected) {
                if (strcmp(newFileName, ubd->cfg.fileName)) {
                        dictSetString(newFileName, "%s.fileName", ubd->configKey);
                }
		usbcdcStopNetwork(ubd);
                free(ubd->cfg.fileName);
		ubd->cfg.fileName = newFileName;
                if (newConnected) {
			usbcdcStartNetwork(ubd);
                }
		ul->seq++;
        }
        return 0;
}

static struct usbcdcdev*
buildUSBCDCDev(struct myusbdev* parent,
	       unsigned int fint)
{
	char key[40];
	struct usbinterface* ui;
	struct usbinterface* ui0;
	struct usbcdcdev* ubd;
	struct usbendpoint* ue;

	sprintf(key, "petrusbcdc%u:%u.eth", parent->id, fint);
	if (!dictGetBoolean(1, "%s.present", key)) {
		return NULL;
	}
	ubd = malloc(sizeof(*ubd));
	if (!ubd) {
		usb_log("Cannot allocate memory for USB CDC Ethernet device\n");
		return NULL;
	}
	memset(ubd, 0, sizeof(*ubd));
	ubd->ifnum = fint;
	ubd->configKey = strdup(key);
	ubd->parent = parent;
	ubd->iControlInterface  = usbdeviceAddString(&ubd->parent->usbdevice, "CDC Ethernet Interface", 1);
	ubd->iDataInterface     = usbdeviceAddString(&ubd->parent->usbdevice, "Data Interface", 1);
	ubd->iEthernetMAC       = usbdeviceAddString(&ubd->parent->usbdevice, "????????????", 0);
	ubd->cfg.fileName       = strdup(dictGetString("/dev/vmnet8", "%s.fileName", ubd->configKey));
	ubd->cfg.startConnected = dictGetBoolean(1, "%s.startConnected", ubd->configKey);
	vmnetInit(&ubd->vmnet);

	getMAC(ubd, parent->id * 16 + fint);
	ui = &ubd->controlUSBInterface;
	usbinterfaceInit(ui);
	ui->fns = &controlUSBInterface_fns;
	ui->intfCookie = ubd;

	ue = &ubd->controlUSBIntrEP;
	usbendpointInit(ue);
	ue->endpointCookie = ubd;
	ue->bEndpointAddress = USB_DIR_IN;
        ue->bmAttributes = USB_ENDPOINT_XFER_INT;
        ue->wMaxPacketSize = 32;
        ue->bInterval = 64;
	usbinterfaceAddEndpoint(ui, ue);
	
	usbconfigAddInterface(&ubd->parent->usbconfig, ui);

	ui0 = &ubd->data0USBInterface;
	usbinterfaceInit(ui0);
	ui0->fns = &data0USBInterface_fns;
	ui0->intfCookie = ubd;

	usbconfigAddInterface(&ubd->parent->usbconfig, ui0);
	
	ui = &ubd->data1USBInterface;
	usbinterfaceInit(ui);
	ui->fns = &data1USBInterface_fns;
	ui->intfCookie = ubd;
	
	ue = &ubd->data1USBBulkInEP;
	usbendpointInit(ue);
	ue->maxQueueLength = 64;
	ue->endpointCookie = ubd;
	ue->bEndpointAddress = USB_DIR_IN;
        ue->bmAttributes = USB_ENDPOINT_XFER_BULK;
        ue->wMaxPacketSize = 64;
        ue->bInterval =  0x00;
	
	usbinterfaceAddEndpoint(ui, ue);

	ue = &ubd->data1USBBulkOutEP;
	usbendpointInit(ue);
	ue->outputXfer = data1USBInterfaceSend;
	ue->endpointCookie = ubd;
	ue->bEndpointAddress = USB_DIR_OUT;
        ue->bmAttributes = USB_ENDPOINT_XFER_BULK;
        ue->wMaxPacketSize = 64;
        ue->bInterval =  0x00;
	
	usbinterfaceAddEndpoint(ui, ue);

	usbinterfaceAddAlternative(ui0, ui);
	
	usblogdevInit(&ubd->logdev);
        ubd->logdev.getName     = usbcdcGetName;
        ubd->logdev.getSettings = usbcdcGetSettings;
        ubd->logdev.setSettings = usbcdcSetSettings;
	ubd->logdev.forkCleanup = usbcdcForkCleanup;

        usbconfigAddLogicalDevice(&ubd->parent->usbconfig, &ubd->logdev);

	if (ubd->cfg.startConnected) {
		usbcdcStartNetwork(ubd);
	}

	return ubd;
}

static void
usbcdcdevDestroy(struct usbcdcdev* ubd)
{
	usbconfigRemoveLogicalDevice(&ubd->parent->usbconfig, &ubd->logdev);
	usbconfigRemoveInterface(&ubd->parent->usbconfig, &ubd->controlUSBInterface);
	usbconfigRemoveInterface(&ubd->parent->usbconfig, &ubd->data0USBInterface);
	usbconfigRemoveInterface(&ubd->parent->usbconfig, &ubd->data1USBInterface);
	usbinterfaceRemoveEndpoint(&ubd->controlUSBInterface, &ubd->controlUSBIntrEP);
	usbinterfaceRemoveEndpoint(&ubd->data1USBInterface, &ubd->data1USBBulkInEP);
	usbinterfaceRemoveEndpoint(&ubd->data1USBInterface, &ubd->data1USBBulkOutEP);
	usbcdcStopNetwork(ubd);
	usbdeviceRemoveString(&ubd->parent->usbdevice, ubd->iEthernetMAC);
	ubd->iEthernetMAC = 0;
	
	usblogdevDestroy(&ubd->logdev);
	usbinterfaceDestroy(&ubd->controlUSBInterface);
	usbinterfaceDestroy(&ubd->data0USBInterface);
	usbinterfaceDestroy(&ubd->data1USBInterface);
	usbendpointDestroy(&ubd->controlUSBIntrEP);
	usbendpointDestroy(&ubd->data1USBBulkInEP);
	usbendpointDestroy(&ubd->data1USBBulkOutEP);
	
	free(ubd->configKey);
	ubd->configKey = NULL;
	free(ubd->cfg.fileName);
	ubd->cfg.fileName = NULL;
	free(ubd);
}

static struct usbdevice_fns mydev_devfns = {
};

static void
usbcdcDevForkCleanup(struct usblogdev* ul)
{
	struct myusbdev* ubd = list_entry(ul, struct myusbdev, logdev);

}

static int
usbcdcDevGetName(struct usblogdev* ul,
                 unsigned char* buffer)
{
	struct myusbdev* ubd = list_entry(ul, struct myusbdev, logdev);

        setBE16(buffer, sprintf(buffer + 2, "USB CDC Ethernet"));
        return getBE16(buffer) + 2;
}

static int
usbcdcDevGetSettings(struct usblogdev* ul,
		     unsigned char* buffer)
{
        struct myusbdev* ubd = list_entry(ul, struct myusbdev, logdev);
        unsigned char* ptr;

        ptr = stuffString(buffer, "CDCDevice");
        ptr = stuffBoolean(ptr, 1, "motorola", ubd->motorola);
        ptr = stuffInteger(ptr, 1, "devs", ubd->devicesCount);
        return ptr - buffer;
}

static void
usbcdcdevSetMotorola(struct myusbdev* ud,
		     int motorola)
{
	ud->motorola = motorola != 0;
	if (motorola) {
		ud->usbdevice.desc.bDeviceClass = 0x02;
		ud->usbdevice.desc.idVendor     = 0x07B2;
		ud->usbdevice.desc.idProduct    = 0x5100;
	} else {
		ud->usbdevice.desc.bDeviceClass = 0x00;
                ud->usbdevice.desc.idVendor     = 0x0000;
       	        ud->usbdevice.desc.idProduct    = 0x0000;
	}
}

static void
usbcdcdevSyncDevicesCount(struct myusbdev* ud)
{
	unsigned int maxDev;
	unsigned int fint;
	
	maxDev = ud->motorola ? 1 : ud->devicesCount;
	for (fint = 0; fint < MAX_USBCDC_DEV; fint++) {
		struct usbcdcdev* dev;

		dev = ud->devices[fint];
		if (fint < maxDev) {
			if (!dev) {
				ud->devices[fint] = buildUSBCDCDev(ud, fint);
			}
		} else {
			if (dev) {
				usbcdcdevDestroy(dev);
				ud->devices[fint] = NULL;
			}
		}
	}
	usbconfigAssignEndpoints(&ud->usbconfig);
}

static int
usbcdcDevSetSettings(struct usblogdev* ul,
		     const unsigned char* inp,
		     size_t inplen,
		     unsigned char* buffer)
{
        struct myusbdev* ubd = list_entry(ul, struct myusbdev, logdev);
        const unsigned char* ptr;
        const unsigned char* key;
        const unsigned char* val;
        unsigned char type;
	int newMotorola;
	unsigned int newDevices;

	newMotorola = ubd->motorola;
	newDevices = ubd->devicesCount;	
        ptr = inp;
        while ((ptr = fetchChange(ptr, inp + inplen, &key, &type, &val)) != NULL) {
		if (!keycmp(key, "motorola")) {
			retrieveBoolean(&newMotorola, type, val);
		} else if (!keycmp(key, "devs")) {
			retrieveInteger(&newDevices, type, val);
			if (newDevices > MAX_USBCDC_DEV) {
				newDevices = MAX_USBCDC_DEV;
			}
                }
        }
	if (newMotorola != ubd->motorola || newDevices != ubd->devicesCount) {
		if (usbdeviceIsClaimed(&ubd->usbdevice)) {
			return -EBUSY;
		}
		if (newMotorola != ubd->motorola) {
			dictSetBoolean(newMotorola, "petrusbcdc%u.motorola", ubd->id);
			usbcdcdevSetMotorola(ubd, newMotorola);
			ul->seq++;
		}
		if (newDevices != ubd->devicesCount) {
			dictSetInteger(newDevices, "petrusbcdc%u.devs", ubd->id);
			ubd->devicesCount = newDevices;
			ul->seq++;
		}
		usbcdcdevSyncDevicesCount(ubd);
	}
        return 0;
}

#define VM_CHECKSIZE(structure, expected)       \
   if (sizeof(structure) != expected) {         \
      asm volatile(".print \"Unexpected ABI change: Structure " #structure " should take %c0 bytes, but took %c1!\"; .abort" : : "i"(expected), "i"(sizeof(structure))); \
   }


static void __attribute__((constructor)) init2(void) {
	unsigned int i;

	for (i = 0; i < 8; i++) {
		struct myusbdev* ud;

		if (!dictGetBoolean(i == 0, "petrusbcdc%u.present", i)) {
			continue;
		}
		ud = malloc(sizeof(*ud));
		if (!ud) {
			usb_log("Cannot allocate memory for USB device\n");
			continue;
		}
		memset(ud, 0, sizeof(*ud));
		usbdeviceInit(&ud->usbdevice);
		ud->usbdevice.devcookie = &ud->usbdevice;
		ud->usbdevice.fns = &mydev_devfns;
		ud->id = i;
		ud->devicesCount = dictGetInteger(1, "petrusbcdc%u.devs", i);
		usbcdcdevSetMotorola(ud, dictGetBoolean(0, "petrusbcdc%u.motorola", i));

                ud->usbdevice.desc.bcdUSB          = 0x0110;
                ud->usbdevice.desc.bMaxPacketSize0 = 8;
                ud->usbdevice.desc.bcdDevice       = 0x0125;
                ud->usbdevice.desc.iManufacturer   = usbdeviceAddString(&ud->usbdevice, "VMware, Inc.", 1);
                ud->usbdevice.desc.iProduct        = usbdeviceAddString(&ud->usbdevice, "CDC Ethernet", 1);
		{
			char sn[20] = "DEADBEEF0100";
			sprintf(sn + 10, "%2.2X", i);
	                ud->usbdevice.desc.iSerialNumber   = usbdeviceAddString(&ud->usbdevice, sn, 1);
		}
		usbconfigInit(&ud->usbconfig);
		usbdeviceAddConfig(&ud->usbdevice, &ud->usbconfig);

		usblogdevInit(&ud->logdev);
	        ud->logdev.getName     = usbcdcDevGetName;
	        ud->logdev.getSettings = usbcdcDevGetSettings;
	        ud->logdev.setSettings = usbcdcDevSetSettings;
		ud->logdev.forkCleanup = usbcdcDevForkCleanup;
	        usbconfigAddLogicalDevice(&ud->usbconfig, &ud->logdev);

		usbcdcdevSyncDevicesCount(ud);

		usbdeviceRegister(&ud->usbdevice, -1, -1);
	}
}
