#define _GNU_SOURCE
#define _LARGEFILE_SOURCE
#define _LARGEFILE64_SOURCE

#include "usbcontrol.h"

#include "dictionary.h"

#include "usbdev.h"

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <pthread.h>

#include <sys/un.h>
#include <sys/socket.h>
#include <sys/poll.h>

static pthread_t pollThread = 0;
static unsigned int pollCurrent = 0;

static int masterFD;

struct pthread_semaphore {
	pthread_cond_t cond;
	pthread_mutex_t mutex;
	int count;
};

static void
pthread_semaphore_init(struct pthread_semaphore* sema)
{
	pthread_cond_init(&sema->cond, NULL);
	pthread_mutex_init(&sema->mutex, NULL);
	sema->count = 0;
}

static void
pthread_semaphore_wait(struct pthread_semaphore* sema)
{
	pthread_mutex_lock(&sema->mutex);
	if (--sema->count >= 0) {
		pthread_mutex_unlock(&sema->mutex);
		return;
	}
	pthread_cond_wait(&sema->cond, &sema->mutex);
	pthread_mutex_unlock(&sema->mutex);
}

static void
pthread_semaphore_signal(struct pthread_semaphore* sema)
{
	int need_sig;

	pthread_mutex_lock(&sema->mutex);
	need_sig = ++sema->count <= 0;
	pthread_mutex_unlock(&sema->mutex);
	if (need_sig) {
		pthread_cond_signal(&sema->cond);
	}
}

struct pipeOps {
	int op;
	int retval;
	int fd;
	void* cookie;
	int (*fdready)(void* cookie, int fd, int revents);
	int events;
	struct pthread_semaphore sema;
};

#define MAX_POLL_ENTRIES	128

struct pollptr {
	void* cookie;
	int (*fdready)(void* cookie, int fd, int revents);
};

static unsigned int pollEntries = 0;
static struct pollfd fds[MAX_POLL_ENTRIES];
static struct pollptr ptrs[MAX_POLL_ENTRIES];

static int 
addPollEntry(int fd, 
	     void* cookie, 
	     int (*fdready)(void* cookie, int fd, int revents),
	     int events)
{
	if (pollEntries >= MAX_POLL_ENTRIES) {
		return -EOVERFLOW;
	}
	fds[pollEntries].fd = fd;
	fds[pollEntries].events = events;
	fds[pollEntries].revents = 0;
	ptrs[pollEntries].cookie = cookie;
	ptrs[pollEntries].fdready = fdready;
	pollEntries++;
	return 0;
}

static int
deletePollEntry(int fd)
{
	unsigned int i;
	
	for (i = 0; i < pollEntries; i++) {
		if (fds[i].fd == fd) {
			--pollEntries;
			fds[i] = fds[pollEntries];
			ptrs[i] = ptrs[pollEntries];
			if (i < pollCurrent) {
				i = pollCurrent;
			}
		}
	}
	return -ENOENT;
}

static void* 
pollLoop(void* dummy)
{
	pollThread = pthread_self();

	while (1) {
		int r;

		r = poll(fds, pollEntries, -1);
		if (r > 0) {
			unsigned int i;

			for (i = 0; i < pollEntries && r; ) {
				int revents = fds[i].revents;
				int del;
				
				fds[i].revents = 0;	/* Clear revents so removal from callbacks works correctly */
				if (revents & POLLNVAL) {
					del = 1;
					r--;
				} else if (revents) {
					pollCurrent = i;
					del = ptrs[i].fdready(ptrs[i].cookie, fds[i].fd, revents);
					i = pollCurrent;
					r--;
				} else {
					del = 0;
				}
				if (del) {
					--pollEntries;
					fds[i] = fds[pollEntries];
					ptrs[i] = ptrs[pollEntries];
				} else {
					i++;
				}
			}
		}
	}
	(void)dummy;
}

unsigned char* stuffString(unsigned char* buffer, const char* str) {
	size_t ln;
	
	if (!str) {
		str = "";
	}
	ln = strlen(str);
	setBE16(buffer, ln);
	memcpy(buffer + 2, str, ln);
	return buffer + 2 + ln;
}

unsigned char* stuffBoolean(unsigned char* buffer, unsigned int writeable, const char* name, int value) {
	buffer = stuffString(buffer, name);
	*buffer++ = writeable ? 'B' : 'b';
	setBE16(buffer, 1);
	buffer[2] = value != 0;
	return buffer + 3;
}

unsigned char* stuffInteger(unsigned char* buffer, unsigned int writeable, const char* name, unsigned int value) {
	buffer = stuffString(buffer, name);
	*buffer++ = writeable ? 'I' : 'i';
	setBE16(buffer, 4);
	setBE32(buffer + 2, value);
	return buffer + 6;
}

unsigned char* stuffFileName(unsigned char* buffer, unsigned int writeable, const char* name, const char* value) {
	buffer = stuffString(buffer, name);
	*buffer++ = writeable ? 'S' : 's';
	return stuffString(buffer, value);
}

int keycmp(const unsigned char* key, const char* val) {
	size_t ln;
	
	ln = strlen(val);
	if (getBE16(key) != ln) {
		return 1;
	}
	return memcmp(key + 2, val, ln);
}

const unsigned char* fetchChange(const unsigned char* ptr, const unsigned char* end, const unsigned char** key, char* type, const unsigned char** val) {
	unsigned int ln;
	
	*key = ptr;
	if (ptr + 5 > end) {
		return NULL;
	}
	ln = getBE16(ptr);
	ptr += 2 + ln;
	if (ptr + 3 > end) {
		return NULL;
	}
	*type = *ptr++;
	*val = ptr;
	ln = getBE16(ptr);
	ptr += 2 + ln;
	if (ptr > end) {	
		return NULL;
	}
	return ptr;
}

int
fetchBoolean(int type,
	     const unsigned char* val) {
	size_t ln;

	ln = getBE16(val);
	val += 2;
	if (type == 'B') {
		int w = 0;

		while (ln--) {
			w |= *val++;
		}
		return w != 0;
	}
	return -1;
}

char*
fetchString(int type,
	    const unsigned char* val) {
	size_t ln;

	ln = getBE16(val);
	val += 2;
	if (type == 'S') {
		char* s;

		s = malloc(ln + 1);
		if (!s) {
			return NULL;
		}
		memcpy(s, val, ln);
		s[ln] = 0;
		return s;
	}
	if (type == 'B') {
		const char* s;
		int w = 0;

		while (ln--) {
			w |= *val++;
		}
		s = w ? "TRUE" : "FALSE";
		return strdup(s);
	}
	if (type == 'I' || type == 'U') {
		unsigned int v = 0;
		char* s;

		if (type == 'I' && ln) {
			v = *(const char*)val;
			val++;
			ln--;
		}
		while (ln-- > 0) {
			v = (v << 8) | *val++;
		}
		s = malloc(20);
		if (!s)	{
			return NULL;
		}
		sprintf(s, (type == 'U') ? "%u" : "%d", v);
		return s;	
	}
	return NULL;
}

void
retrieveBoolean(int* v,
		int type,
		const unsigned char* val)
{
	int r;

	r = fetchBoolean(type, val);
	if (r >= 0) {
		*v = r;
	}
}

void
retrieveString(char** v,
	       int type,
	       const unsigned char* val)
{
	char* r;

	r = fetchString(type, val);
	if (r) {
		free(*v);
		*v = r;
	}
}

int
retrieveInteger(unsigned int* v,
		int type,
		const unsigned char* val)
{
	size_t ln;

	ln = getBE16(val);
	val += 2;
	if (type == 'B') {
		unsigned int w = 0;

		while (ln--) {
			w |= *val++;
		}
		*v = w ? 1 : 0;
		return 0;
	}
	if (type == 'I' || type == 'U') {
		unsigned int w = 0;

		if (type == 'I' && ln) {
			w = *(const char*)val;
			val++;
			ln--;
		}
		while (ln-- > 0) {
			w = (w << 8) | *val++;
		}
		*v = w;
		return 0;
	}
	return -1;
}

static unsigned char* ubdGetString(unsigned char* buffer, struct usbdevice* usbdev, unsigned int idx, unsigned int language) {
	if (idx != 0) {
		int len;
		unsigned char bstr[258];
	
		len = usbdeviceGetDescriptor(usbdev, USB_DT_STRING, idx, language, sizeof(bstr), bstr);
		if (len >= 2) {
			const unsigned short* ptr;
			const unsigned short* end;
			unsigned char* dptr;

			dptr = buffer + 2;
			ptr = (unsigned short*)(bstr + 2);
			end = (unsigned short*)(bstr + len);
			while (ptr < end) {
				unsigned int c = *ptr++;
			
				if (c < 1 << 7) {
					*dptr++ =   c;
				} else if (c < 1 << 11) {
					*dptr++ =  (c >>  6)         | 0xC0;
					*dptr++ = ( c        & 0x3F) | 0x80;
				} else if (c < 1 << 16) {
					*dptr++ =  (c >> 12)         | 0xE0;
					*dptr++ = ((c >>  6) & 0x3F) | 0x80;
					*dptr++ = ( c        & 0x3F) | 0x80;
				} else if (c < 1 << 21) {
					*dptr++ =  (c >> 18)         | 0xF0;
					*dptr++ = ((c >> 12) & 0x3F) | 0x80;
					*dptr++ = ((c >>  6) & 0x3F) | 0x80;
					*dptr++ = ( c        & 0x3F) | 0x80;
				} else if (c < 1 << 26) {
					*dptr++ =  (c >> 24)         | 0xF8;
					*dptr++ = ((c >> 18) & 0x3F) | 0x80;
					*dptr++ = ((c >> 12) & 0x3F) | 0x80;
					*dptr++ = ((c >>  6) & 0x3F) | 0x80;
					*dptr++ = ( c        & 0x3F) | 0x80;
				} else if (c < 1U << 31) {
					*dptr++ =  (c >> 30)         | 0xFC;
					*dptr++ = ((c >> 24) & 0x3F) | 0x80;
					*dptr++ = ((c >> 18) & 0x3F) | 0x80;
					*dptr++ = ((c >> 12) & 0x3F) | 0x80;
					*dptr++ = ((c >>  6) & 0x3F) | 0x80;
					*dptr++ = ( c        & 0x3F) | 0x80;
				} else {
					*dptr++ = '?';
				}
			}
			setBE16(buffer, dptr - buffer - 2);
			return dptr;
		}
		return stuffString(buffer, "???");
	}
	return stuffString(buffer, "");
}

static int webCallback(void* dummy, int fd, int revents) {
	struct sockaddr_un sun;
	int ln;
	unsigned char cmd[4096];
	socklen_t sl;
		
	sl = sizeof(sun);
	ln = recvfrom(fd, cmd, sizeof(cmd), MSG_DONTWAIT, &sun, &sl);
	if (ln >= 2) {
		size_t replen = 3;

		switch (cmd[1]) {
			case 0:	/* Get version */
				cmd[2] = 0;
				cmd[3] = 4;
				cmd[4] = '1';
				cmd[5] = '.';
				cmd[6] = '0';
				cmd[7] = '0';
				replen = 8;
				break;
			case 1:	/* Enumerate devices */
				{
					size_t restlen;
					
					cmd[2] = 0;
					cmd[3] = 0;
					restlen = get_registered_usbdevs((u_int32_t*)(cmd + 8), (sizeof(cmd) - 8) / 4);
					setBE32(cmd + 4, restlen);
					replen = 8 + restlen * 4;
				}
				break;
			case 2:	/* Retrieve device name */
				{
					unsigned int bus, device;
					struct usbdevice* usbdev;
					struct usb_device_descriptor udd;
					struct usb_string_descriptor usd;
					unsigned char* ptr;
					int erc;

					if (ln < 6) {
						cmd[2] = EPROTO;
						break;
					}
					bus = getBE16(cmd + 2);
					device = getBE16(cmd + 4);
					usbdev = usbdeviceFind(bus, device);
					if (!usbdev) {
						cmd[2] = ENOENT;
						break;
					}
					erc = usbdeviceGetDescriptor(usbdev, USB_DT_DEVICE, 0, 0, sizeof(udd), &udd);
					if (erc < 0) {
						cmd[2] = -erc;
						break;
					} else if ((size_t)erc < sizeof(udd)) {
						cmd[2] = EINVAL;
						break;
					}
					erc = usbdeviceGetDescriptor(usbdev, USB_DT_STRING, 0, 0, sizeof(usd), &usd);
					if (erc < 0) {
						cmd[2] = -erc;
						break;
					} else if ((size_t)erc < sizeof(usd)) {
						cmd[2] = EINVAL;
						break;
					}
					ptr = ubdGetString(cmd + 3, usbdev, udd.iManufacturer, usd.wData[0]);
					ptr = ubdGetString(ptr, usbdev, udd.iProduct, usd.wData[0]);
					ptr = ubdGetString(ptr, usbdev, udd.iSerialNumber, usd.wData[0]);
					cmd[2] = 0;
					replen = ptr - cmd;
				}
				break;
			case 3:	/* Retrieve logical devices */
				{
					unsigned int bus, device, intf;
					struct usbdevice* usbdev;
					struct usbconfig* usbconf;
					struct usblogdev* usbldev;
					
					if (ln < 8) {
						cmd[2] = EPROTO;
						break;
					}
					bus = getBE16(cmd + 2);
					device = getBE16(cmd + 4);
					intf = getBE16(cmd + 6);
					usbdev = usbdeviceFind(bus, device);
					if (!usbdev) {
						cmd[2] = ENOENT;
						break;
					}
					usbconf = usbdeviceCurrentConfig(usbdev);
					if (!usbconf) {
						cmd[2] = ENOENT;
						break;
					}
				rl31:;
					if (intf == 0xFFFF) {
						intf = 0;
					} else {
						intf++;
					}
					if (intf >= usbconf->logdevs.len) {
						cmd[2] = ESRCH;
						break;
					}
					usbldev = dynarray_get(&usbconf->logdevs, intf);
					if (usbldev == NULL) {
						goto rl31;
					}
					setBE16(cmd + 3, intf);
					replen = 5 + usbldev->getName(usbldev, cmd + 5);
					cmd[2] = 0;
				}
				break;
			case 4:	/* Retrieve config file name */
				{
					const char* cfg;
					size_t cfglen;
					
					cfg = vmxGetConfig();
					if (!cfg) {
						cmd[2] = ENODATA;
						break;
					}
					cfglen = strlen(cfg);
					setBE32(cmd + 3, cfglen);
					memcpy(cmd + 7, cfg, cfglen);
					replen = cfglen + 7;
					cmd[2] = 0;
				}
				break;
			case 5:	/* Retrieve claim info */
				{
					unsigned int bus, device, intf;
					struct usbdevice* usbdev;
					
					if (ln < 8) {
						cmd[2] = EPROTO;
						break;
					}
					bus = getBE16(cmd + 2);
					device = getBE16(cmd + 4);
					intf = getBE16(cmd + 6);
					usbdev = usbdeviceFind(bus, device);
					if (!usbdev) {
						cmd[2] = ENOENT;
						break;
					}
					if (intf == 0xFFFF) {
						cmd[3] = usbdeviceIsClaimed(usbdev);
					} else {
						struct usbinterface* ui;

						ui = usbdeviceGetInterface(usbdev, intf);
						if (!ui) {
							cmd[2] = ENOENT;
							break;
						}
						cmd[3] = usbinterfaceIsClaimed(ui);
					}
					cmd[2] = 0;
					replen = 4;
				}
				break;
			case 6:	/* Retrieve device settings */
				{
					unsigned int bus, device, intf;
					struct usbdevice* usbdev;
					struct usbconfig* usbconf;
					struct usblogdev* usbldev;
					
					if (ln < 8) {
						cmd[2] = EPROTO;
						break;
					}
					bus = getBE16(cmd + 2);
					device = getBE16(cmd + 4);
					intf = getBE16(cmd + 6);
					usbdev = usbdeviceFind(bus, device);
					if (!usbdev) {
						cmd[2] = ENOENT;
						break;
					}
					usbconf = usbdeviceCurrentConfig(usbdev);
					if (!usbconf) {
						cmd[2] = ENOENT;
						break;
					}
					usbldev = usbconfigGetLogicalDevice(usbconf, intf);
					if (!usbldev) {
						cmd[2] = ENOENT;
						break;
					}
					setBE32(cmd + 3, usbldev->seq);
					replen = 7 + usbldev->getSettings(usbldev, cmd + 7);
					cmd[2] = 0;
				}
				break;
			case 7:	/* Set device settings */
				{
					unsigned int bus, device, intf, seq;
					struct usbdevice* usbdev;
					struct usbconfig* usbconf;
					struct usblogdev* usbldev;
					int ret;
					
					if (ln < 12) {
						cmd[2] = EPROTO;
						break;
					}
					bus = getBE16(cmd + 2);
					device = getBE16(cmd + 4);
					intf = getBE16(cmd + 6);
					usbdev = usbdeviceFind(bus, device);
					if (!usbdev) {
						cmd[2] = ENOENT;
						break;
					}
					usbconf = usbdeviceCurrentConfig(usbdev);
					if (!usbconf) {
						cmd[2] = ENOENT;
						break;
					}
					usbldev = usbconfigGetLogicalDevice(usbconf, intf);
					if (!usbldev) {
						cmd[2] = ENOENT;
						break;
					}
					seq = getBE32(cmd + 8);
					if (usbldev->seq != seq) {
						cmd[2] = ESTALE;
						break;
					}
					ret = usbldev->setSettings(usbldev, cmd + 12, ln - 12, cmd + 3);
					if (ret < 0) {
						cmd[2] = -ret;
						break;
					}
					cmd[2] = 0;
					replen = ret + 3;
				}
				break;
			default:
				cmd[2] = EINVAL;
				break;
		}
		sendto(fd, cmd, replen, MSG_DONTWAIT, &sun, sl);
	}
	(void)dummy;
	(void)revents;
	return 0;
}

static int
webInit(void)
{
	int fd;
	struct sockaddr_un sun;
	size_t ln;
	static const int one = 1;
	
	fd = socket(PF_UNIX, SOCK_DGRAM, 0);
	if (fd < 0) {
		return -errno;
	}
	fcntl(fd, F_SETFD, fcntl(fd, F_GETFD) | FD_CLOEXEC);
	setsockopt(fd, SOL_SOCKET, SO_REUSEADDR, &one, sizeof(one));
	sun.sun_family = AF_UNIX;
	sun.sun_path[0] = 0;
	ln = offsetof(struct sockaddr_un, sun_path) + 1 + sprintf(sun.sun_path + 1, "%s-%u", "vmwusb", getpid());
	if (bind(fd, (struct sockaddr*)&sun, ln)) {
		close(fd);
		return -errno;
	}
	addPollEntry(fd, NULL, webCallback, POLLIN);
	return 0;
}

static int
pipeCallback(void* cookie,
	     int fd,
	     int revents)
{
	struct pipeOps* pop;

	if (read(fd, &pop, sizeof(pop)) == sizeof(pop)) {
		if (pop->op == 0) {
			pop->retval = addPollEntry(pop->fd, pop->cookie, pop->fdready, pop->events);
		} else if (pop->op == 1) {
			pop->retval = deletePollEntry(pop->fd);
		} else {
			pop->retval = -EINVAL;
		}
		pthread_semaphore_signal(&pop->sema);
	}
	return 0;
	(void)cookie;
	(void)revents;
}

int
usbcontrolStartWatchingFileDescriptor(int fd,
				      void* cookie,
				      int (*fdready)(void* cookie, int fd, int revents),
				      int events)
{
	if (pollThread == pthread_self()) {
		int ret;
		
		ret = addPollEntry(fd, cookie, fdready, events);
		return ret;
	} else {
		struct pipeOps op;
		struct pipeOps* pop;

		op.op = 0;
		op.fd = fd;
		op.cookie = cookie;
		op.fdready = fdready;
		op.events = events;
		op.retval = -EBADF;
		pthread_semaphore_init(&op.sema);
		pop = &op;
		write(masterFD, &pop, sizeof(pop));
		pthread_semaphore_wait(&op.sema);
		return op.retval;
	}
}

int
usbcontrolStopWatchingFileDescriptor(int fd)
{
	if (pollThread == pthread_self()) {
		int ret;

		ret = deletePollEntry(fd);
		return ret;
	} else {
		struct pipeOps op;
		struct pipeOps* pop;

		op.op = 1;
		op.fd = fd;
		op.retval = -EBADF;
		pthread_semaphore_init(&op.sema);
		pop = &op;
		write(masterFD, &pop, sizeof(pop));
		pthread_semaphore_wait(&op.sema);
		return op.retval;
	}
}

static int
cmdInit(void)
{
	int pipes[2];
	
	if (pipe(pipes) < 0) {
		return -errno;
	}
	fcntl(pipes[0], F_SETFL, fcntl(pipes[0], F_GETFL) | O_NONBLOCK);
	fcntl(pipes[1], F_SETFL, fcntl(pipes[1], F_GETFL) | O_NONBLOCK);
	masterFD = pipes[1];
	addPollEntry(pipes[0], NULL, pipeCallback, POLLIN);
	return 0;
}

void
usbcontrolInit(void) {
	pthread_t p;
	pthread_attr_t pa;

	cmdInit();
	webInit();
	pthread_attr_init(&pa);
	pthread_attr_setdetachstate(&pa, PTHREAD_CREATE_DETACHED);
	if (pthread_create(&p, NULL, pollLoop, NULL)) {
		usb_log("Cannot spawn web interface\n");
		return;
	}
	pthread_attr_destroy(&pa);
}

void
usbcontrolForkCleanup(void)
{
	close(masterFD);
	masterFD = -1;
	pollThread = pthread_self();
}
