#define _GNU_SOURCE
#define _LARGEFILE_SOURCE
#define _LARGEFILE64_SOURCE

#include "usbdev.h"

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>

#include <pthread.h>

struct usbstr {
	int constant;
	u_int16_t data[0];
};

typedef int (*usbendpointIProc)(void*, size_t, void*);
typedef int (*usbendpointOProc)(void*, size_t, const void*);

int usb_copy_desc(int dstlen, void* dst, int srclen, const void* src) {
	if (srclen > dstlen) {
		srclen = dstlen;
	}
	memcpy(dst, src, srclen);
	return srclen;
}

static void dynarray_init(struct dynarray* p) {
	p->len = 0;
	p->allocated = 0;
	p->ptr = NULL;
}

static void dynarray_done(struct dynarray* p) {
	free(p->ptr);
	dynarray_init(p);
}

static int
dynarray_putat(struct dynarray* p, void* newelem, unsigned int at) {
	if (at >= p->allocated) {
		unsigned int nl;
		void* q;
		
		nl = at + 4;
		q = realloc(p->ptr, nl * sizeof(void*));
		if (q == NULL) {
			return -ENOMEM;
		}
		p->ptr = q;
		p->allocated = nl;
	}
	while (at >= p->len) {
		p->ptr[p->len++] = NULL;
	}
	if (p->ptr[at]) {
		return -EBUSY;
	}
	p->ptr[at] = newelem;
	return at;	
}

static int dynarray_append(struct dynarray* p, void* newelem) {
	return dynarray_putat(p, newelem, p->len);
}

void* dynarray_get(const struct dynarray* p, unsigned int index) {
	if (index >= p->len) {
		return NULL;
	}
	return p->ptr[index];
}

static int
dynarray_set(const struct dynarray* p, unsigned int index, void* val)
{
	if (index >= p->len) {
		return -EINVAL;
	}
	p->ptr[index] = val;
	return 0;
}

static int
dynarray_put(struct dynarray* p, void* newelem, unsigned int first) {
	while (first < p->len) {
		if (p->ptr[first] == newelem) {
			return -EINVAL;
		}
		if (!p->ptr[first]) {
			p->ptr[first] = newelem;
			return first;
		}
		first++;
	}
	return dynarray_putat(p, newelem, first);
}

static int
dynarray_remove(struct dynarray* p, void* elem)
{
	unsigned int i;
	
	for (i = 0; i < p->len; i++) {
		if (p->ptr[i] == elem) {
			p->len--;
			memcpy(p->ptr + i, p->ptr + i + 1, sizeof(p->ptr[0]) * (p->len - i));
			return 0;
		}
	}
	return -ENOENT;
}

void usbendpointInit(struct usbendpoint* ue) {
	memset(ue, 0, sizeof(*ue));
	pthread_mutex_init(&ue->urbsLock, NULL);
	ue->nextQueuedInputXfer = &ue->queuedInputXfers;
}

void usbendpointDestroy(struct usbendpoint* ue) {
	if (ue->bEndpointAddress & USB_DIR_IN) {
		usbendpointIStall(ue);
	} else {
		usbendpointOStall(ue);
	}
	pthread_mutex_destroy(&ue->urbsLock);
}

static int usbendpointGetDescriptor(struct usbendpoint* ue, void* buffer) {
	struct usb_endpoint_descriptor* ued = buffer;

        memset(ued, 0, sizeof(*ued));
        ued->bLength = USB_DT_ENDPOINT_SIZE;
        ued->bDescriptorType = USB_DT_ENDPOINT;
        ued->bEndpointAddress = ue->bEndpointAddress;
	ued->bmAttributes = ue->bmAttributes;
	ued->wMaxPacketSize = ue->wMaxPacketSize;
	ued->bInterval = ue->bInterval;
	return ued->bLength;
}

static int
usbendpointCallProc(struct usbendpoint* ue,
		    struct urbs* urbs,
		    usbendpointIProc inputProc)
{
	int ret;
	struct usbdevfs_urb* urb = urbs->urb;
		
	ret = inputProc(ue->endpointCookie, urb->buffer_length, urb->buffer);
	if (ret < 0) {
		urb->actual_length = 0;
		urb->status = ret;
		urb->error_count = 1;
	} else {
		urb->actual_length = ret;
		urb->status = 0;
		urb->error_count = 0;
	}
	usbdev_puturb(urbs);
	return 0;
}

static int
usbendpointDispatchXfer(struct urbs* urbs,
			struct queuedXfer* xfer)
{
	int ret;
	struct usbdevfs_urb* urb = urbs->urb;

	ret = xfer->xfer(xfer, urb->buffer_length, urb->buffer);
	if (ret < 0) {
		urb->actual_length = 0;
		urb->status = ret;
		urb->error_count = 1;
	} else {
		urb->actual_length = ret;
		urb->status = 0;
		urb->error_count = 0;
	}
	usbdev_puturb(urbs);
	return 0;
}

static int
usbendpointCallOProc(struct usbendpoint* ue,
		     struct urbs* urbs,
		     usbendpointOProc outputProc)
{
	int ret;
	struct usbdevfs_urb* urb = urbs->urb;
		
	ret = outputProc(ue->endpointCookie, urb->buffer_length, urb->buffer);
	if (ret < 0) {
		urb->actual_length = 0;
		urb->status = ret;
		urb->error_count = 1;
	} else {
		urb->actual_length = ret;
		urb->status = 0;
		urb->error_count = 0;
	}
	usbdev_puturb(urbs);
	return 0;
}

static int usbendpointDeviceControl(struct usbendpoint* ue, const struct usbdevfs_ctrltransfer* ct) {
	switch (ct->bRequestType) {
		case USB_DIR_OUT | USB_TYPE_STANDARD | USB_RECIP_ENDPOINT:
			switch (ct->bRequest) {
				case USB_REQ_CLEAR_FEATURE:
					{
						struct urbs* urbs = NULL;
						usbendpointIProc inputProc = NULL;

						/* FIXME! Check that we are clearing HALT */
						pthread_mutex_lock(&ue->urbsLock);
						ue->stalled &= ~UE_STALL_ONCE;
						if (ue->stalled == 0 && (ue->bEndpointAddress & USB_DIR_IN) && ue->urbs && ue->inputXfer) {
							urbs = ue->urbs;
							ue->urbs = urbs->next;
							inputProc = ue->inputXfer;
							ue->inputXfer = NULL;
						}
						pthread_mutex_unlock(&ue->urbsLock);
						if (urbs) {
							urbs->next = NULL;
							usbendpointCallProc(ue, urbs, inputProc);
						}
					}
					return 0;
				default:
					break;
			}
			break;
		default:
			break;
	}
	usb_log("Unknown ENDPOINT command %02X/%02X\n", ct->bRequestType, ct->bRequest);
	return -EPIPE;
}

static void
usbendpointStallUrbs(struct urbs* urbs)
{
	while (urbs) {
		struct urbs* up;
		struct usbdevfs_urb* urb;

		up = urbs;
		urbs = urbs->next;
		urb = up->urb;
		urb->actual_length = 0;
		urb->status = -EPIPE;
		urb->error_count = 1;

		usb_log("... killing URB %p\n", urb);

		usbdev_puturb(up);
	}
}

static void
usbendpointStallXfers(struct queuedXfer* xfers)
{
	while (xfers) {
		struct queuedXfer* xf;

		xf = xfers;
		xfers = xfers->next;
		usb_log("... killing XFER %p\n", xf);
		xf->abort(xf);
	}
}

int
usbendpointQueueIACK(struct usbendpoint* ue,
		     struct queuedXfer* xfer)
{
	struct urbs* urbs;
	int err;

	usb_log("ACKing %p: xfer=%p\n", ue, xfer);
	if (!(ue->bEndpointAddress & USB_DIR_IN)) {
		usb_log("Not INPUT\n");
		return -EPIPE;
	}
	if (!ue->maxQueueLength) {
		usb_log("Not queue based\n");
		return -EPIPE;
	}
	if (!xfer || !xfer->xfer || !xfer->abort) {
		usb_log("Missing some mandatory xfer pointer\n");
		return -EPIPE;
	}
	xfer->next = NULL;
	pthread_mutex_lock(&ue->urbsLock);
	ue->stalled = 0;
	urbs = ue->urbs;
	if (urbs) {
		ue->urbs = urbs->next;
		err = 0;
	} else if (ue->queueLength >= ue->maxQueueLength) {
		err = -EPIPE;
	} else {
		ue->queueLength++;
		*ue->nextQueuedInputXfer = xfer;
		ue->nextQueuedInputXfer = &xfer->next;
		err = 0;
	}
	pthread_mutex_unlock(&ue->urbsLock);
	if (err) {
		usb_log("... Queue too deep\n");
		return err;
	}
	if (urbs) {
		usb_log("... Executing ACK\n");
		usbendpointDispatchXfer(urbs, xfer);
		return 0;
	}
	usb_log("... ACK queued\n");
	return 0;
}

int
usbendpointIACK(struct usbendpoint* ue,
		usbendpointIProc inputProc,
		int stall)
{
	struct urbs* urbs;

	usb_log("ACKing %p: proc=%p, stall=%u\n", ue, inputProc, stall);
	if (!(ue->bEndpointAddress & USB_DIR_IN)) {
		usb_log("... EPIPE\n");
		return -EPIPE;
	}
	if (ue->maxQueueLength) {
		struct queuedXfer* xfers;

		if (inputProc || stall != UE_STALL_FOREVER) {
			usb_log("... Queue based pipe!\n");
			return -EPIPE;
		}
		/* Want STALL on queue based pipe */
		pthread_mutex_lock(&ue->urbsLock);
		ue->stalled = stall;
		urbs = ue->urbs;
		ue->urbs = NULL;
		xfers = ue->queuedInputXfers;
		ue->queuedInputXfers = NULL;
		ue->nextQueuedInputXfer = &ue->queuedInputXfers;
		ue->queueLength = 0;
		pthread_mutex_unlock(&ue->urbsLock);
		usbendpointStallUrbs(urbs);
		usbendpointStallXfers(xfers);
		return 0;		
	}
	pthread_mutex_lock(&ue->urbsLock);
	ue->stalled = stall;
	if (stall) {
		/* STALL */
		urbs = ue->urbs;
		ue->urbs = NULL;
		ue->inputXfer = inputProc;
	} else if (inputProc) {
		/* ACK */
		urbs = ue->urbs;
		if (urbs) {
			ue->urbs = urbs->next;
		} else {
			/* if inputXfer was already set, we silently overwrite it... make sure no malloc is in functions */
			ue->inputXfer = inputProc;
		}
	} else {
		/* NACK */
		// ASSERT(ue->urbs == NULL || ue->inputXfer == NULL);
		urbs = NULL;
		ue->inputXfer = NULL;
	}
	pthread_mutex_unlock(&ue->urbsLock);
	if (stall) {
		usbendpointStallUrbs(urbs);
		return 0;
	}
	if (urbs) {
		urbs->next = NULL;
		usb_log("... URB was queued\n");
		/* Note! usbendpointCallProc must succeed, otherwise it is not clear what happened with urbs & with inputProc */
		return usbendpointCallProc(ue, urbs, inputProc);
	}
	usb_log("... queued\n");
	return 0;
}

int
usbendpointOACK(struct usbendpoint* ue)
{
	usb_log("O-ACKing %p\n", ue);

	pthread_mutex_lock(&ue->urbsLock);
	ue->stalled = 0;
	pthread_mutex_unlock(&ue->urbsLock);
	return 0;
}

int
usbendpointOStall(struct usbendpoint* ue)
{
	usb_log("O-Stalling %p\n", ue);

	pthread_mutex_lock(&ue->urbsLock);
	ue->stalled = UE_STALL_ONCE;
	pthread_mutex_unlock(&ue->urbsLock);
	return 0;
}

static int
usbendpointXfer(struct usbendpoint* ue,
		struct urbs* urbs,
		unsigned int type)
{
	if ((ue->bmAttributes & USB_ENDPOINT_XFERTYPE_MASK) != type) {
		usb_log("Warning: Transfer %X issued while %X was expected\n", type, ue->bmAttributes & USB_ENDPOINT_XFERTYPE_MASK);
//		return -EPIPE;
	}
	if (ue->bEndpointAddress & USB_DIR_IN) {
		if (ue->maxQueueLength) {
			struct queuedXfer* xfer;
			int err = 0;

			pthread_mutex_lock(&ue->urbsLock);
			xfer = ue->queuedInputXfers;
			if (ue->stalled) {
				err = -EPIPE;
			} else if (xfer) {
				ue->queueLength--;
				if ((ue->queuedInputXfers = xfer->next) == NULL) {
					ue->nextQueuedInputXfer = &ue->queuedInputXfers;
				}
			} else {
				urbs->next = ue->urbs;
				ue->urbs = urbs;
			}
			pthread_mutex_unlock(&ue->urbsLock);
			if (err) {
				struct usbdevfs_urb* urb = urbs->urb;

				urb->actual_length = 0;
				urb->status = err;
				urb->error_count = 1;
				usb_log("Failing EPIPE... %d\n", err);
				usbdev_puturb(urbs);
				return 0;
			}
			if (xfer) {
				xfer->next = NULL;
				usb_log("Dispatching on ep %u xfer %p\n", ue->bEndpointAddress, xfer);
				return usbendpointDispatchXfer(urbs, xfer);
			}
			usb_log("Queuing %p\n", urbs->urb);
			return 0;
		} else {
			usbendpointIProc proc = NULL;
			int err = 0;

			pthread_mutex_lock(&ue->urbsLock);
			proc = ue->inputXfer;
			if (ue->stalled) {
				err = -EPIPE;
			} else if (proc) {
				ue->inputXfer = NULL;
			} else {
				urbs->next = ue->urbs;
				ue->urbs = urbs;
			}
			pthread_mutex_unlock(&ue->urbsLock);
			if (err) {
				struct usbdevfs_urb* urb = urbs->urb;
	
				urb->actual_length = 0;
				urb->status = err;
				urb->error_count = 1;
				usb_log("Failing EPIPE... %d\n", err);
				usbdev_puturb(urbs);
				return 0;
			}
			if (proc) {
				usb_log("Dispatching on ep %u\n", ue->bEndpointAddress);
				return usbendpointCallProc(ue, urbs, proc);
			}
			usb_log("Queuing %p\n", urbs->urb);
			return 0;
		}
	} else {
		int ret = 0;
		usbendpointOProc proc = NULL;

		pthread_mutex_lock(&ue->urbsLock);
		proc = ue->outputXfer;
		if (ue->stalled) {
			ret = -EPIPE;
		} else if (proc) {
			;
		} else {
			urbs->next = ue->urbs;
			ue->urbs = urbs;
		}
		pthread_mutex_unlock(&ue->urbsLock);
		if (ret) {
			struct usbdevfs_urb* urb = urbs->urb;

			urb->actual_length = 0;
			urb->status = ret;
			urb->error_count = 1;

			usb_log("OXfer failed: %d\n", ret);

			usbdev_puturb(urbs);
			return 0;
		}
		if (proc) {
			usb_log("OXfer issued\n");
			usbendpointCallOProc(ue, urbs, proc);
			return 0;
		}
		usb_log("Queuing %p\n", urbs->urb);
		return 0;
	}
}

static int
usbendpointDiscardUrb(struct usbendpoint* ue, struct usbdevfs_urb* urb)
{
	struct urbs* urbs;
	struct urbs** purbs;

	pthread_mutex_lock(&ue->urbsLock);
	purbs = &ue->urbs;
	while (1) {
		urbs = *purbs;
		if (!urbs) {
			break;
		}
		usb_log("Checking %p ==? %p\n", urbs->urb, urb);
		if (urbs->urb == urb) {
			*purbs = urbs->next;
			break;
		}
		purbs = &urbs->next;
	}
	pthread_mutex_unlock(&ue->urbsLock);
	if (urbs) {
		urb->actual_length = 0;
		urb->status = -EINPROGRESS;
		urb->error_count = 1;
		usbdev_puturb(urbs);
		return 0;
	}
	return -ENOENT;
}

static void
usbendpointForkCleanup(struct usbendpoint* ue)
{
	if (ue->fns && ue->fns->forkCleanup) {
		ue->fns->forkCleanup(ue->endpointCookie);
	}
}

void usbinterfaceInit(struct usbinterface* ui) {
	memset(ui, 0, sizeof(*ui));
	dynarray_init(&ui->endpoints);
}

void usbinterfaceDestroy(struct usbinterface* ui) {
	dynarray_done(&ui->endpoints);
}

int usbinterfaceAddEndpoint(struct usbinterface* ui, struct usbendpoint* ue) {
	return dynarray_append(&ui->endpoints, ue);
}

int usbinterfaceRemoveEndpoint(struct usbinterface* ui, struct usbendpoint* ue) {
	return dynarray_remove(&ui->endpoints, ue);
}

static struct usbendpoint* usbinterfaceGetEndpoint(const struct usbinterface* ui, unsigned int epn) {
	return dynarray_get(&ui->endpoints, epn);
}

int usbinterfaceAddAlternative(struct usbinterface* ui, struct usbinterface* ni) {
	struct usbinterface* c;
	
	ni->parent = ui;
	ni->nextAlternative = NULL;
	ni->config = ui->config;
	ni->bInterfaceNumber = ui->bInterfaceNumber;
	for (c = ui; c->nextAlternative; c = c->nextAlternative) ;
	ni->bAlternateSetting = c->bAlternateSetting + 1;
	c->nextAlternative = ni;
	return 0;
}

static void
usbinterfaceSelect(struct usbinterface* ui)
{
	dynarray_set(&ui->config->interfaces, ui->bInterfaceNumber, ui);
	if (ui->fns && ui->fns->intfSelected) {
		ui->fns->intfSelected(ui->intfCookie);
	}
}

static int 
usbinterfaceSetInterface(struct usbinterface* ui,
			 int alts)
{
	usb_log("Selecting alternative %u for interface %u[%p] (now selected %u, parent %p)\n", alts, ui->bInterfaceNumber, ui, ui->bAlternateSetting, ui->parent);
	if (ui->bAlternateSetting == alts) {
		return 0;
	}
	for (ui = ui->parent; ui; ui = ui->nextAlternative) {
		usb_log("... checking %u.%u[%p]\n", ui->bInterfaceNumber, ui->bAlternateSetting, ui);
		if (ui->bAlternateSetting == alts) {
			usb_log("... match\n");
			usbinterfaceSelect(ui);
			return 0;
		}
	}
	usb_log("... no match\n");
	return -EPIPE;
}

static int usbinterfaceDeviceControl(struct usbinterface* ui, struct usbdevfs_ctrltransfer* ct) {
	switch (ct->bRequestType) {
		case USB_DIR_OUT | USB_TYPE_STANDARD | USB_RECIP_INTERFACE:
			switch (ct->bRequest) {
				case USB_REQ_SET_INTERFACE:
					return usbinterfaceSetInterface(ui, ct->wValue);
				default:
					break;
			}
			break;
		default:
			break;
	}
	/* Filter out generic calls... */
	/* ... then... */
	return ui->fns->intfDeviceControl(ui->intfCookie, ct);
}

static int usbinterfaceGetDescriptor(struct usbinterface* ui, unsigned char* buffer) {
	unsigned int ep;
	struct usb_interface_descriptor* uid = (struct usb_interface_descriptor*)buffer;
	int len;

	uid->bLength = USB_DT_INTERFACE_SIZE;
	uid->bDescriptorType = USB_DT_INTERFACE;
	uid->bInterfaceNumber = ui->bInterfaceNumber;
	uid->bAlternateSetting = ui->bAlternateSetting;
	uid->bNumEndpoints = ui->endpoints.len;
	uid->bInterfaceClass = 0;
	uid->bInterfaceSubClass = 0;
	uid->bInterfaceProtocol = 0;
	uid->iInterface = 0;
	len = ui->fns->intfGetDescriptor(ui->intfCookie, uid);
	if (len < 0) {
		return len;
	}
	len = uid->bLength;
	if (ui->fns->intfGetFunctionalDescriptor) {
		for (ep = 0; ep < 512; ep++) {
			int dln;

			dln = ui->fns->intfGetFunctionalDescriptor(ui->intfCookie, buffer + len, ep);
			if (dln < 2) {
				break;
			}
			buffer[len] = dln;
			len += dln;
		}
	}
	for (ep = 0; ep < ui->endpoints.len; ep++) {
		struct usbendpoint* ue;
		
		ue = usbinterfaceGetEndpoint(ui, ep);
		if (!ue) {
			return -1;
		}
		len += usbendpointGetDescriptor(ue, buffer + len);
	}
	return len;
}


static int usbinterfaceClaim(struct usbinterface* ui) {
	if (ui->claimed) {
		return -EBUSY;
	}
	ui->claimed = 1;
	return 0;
}

static int usbinterfaceRelease(struct usbinterface* ui) {
	ui->claimed = 0;
	return 0;
}

int usbinterfaceIsClaimed(const struct usbinterface* ui) {
	return ui->claimed;
}

static void
usbinterfaceForkCleanup(struct usbinterface* ui)
{
	unsigned int i;

	for (i = 0; i < ui->endpoints.len; i++) {
		struct usbendpoint* ue;

		ue = usbinterfaceGetEndpoint(ui, i);
		if (ue) {
			usbendpointForkCleanup(ue);
		}
	}
	if (ui->fns && ui->fns->forkCleanup) {
		ui->fns->forkCleanup(ui->intfCookie);
	}
	dynarray_done(&ui->endpoints);
}

void usbconfigInit(struct usbconfig* uc) {
	memset(uc, 0, sizeof(*uc));
	dynarray_init(&uc->interfaces);
	dynarray_init(&uc->iendpoints);
	dynarray_init(&uc->oendpoints);
	dynarray_init(&uc->logdevs);
};

static struct usbinterface* usbconfigGetInterface(const struct usbconfig* uc, int intf) {
	struct usbinterface* ui;
	
	if (intf >= 0) {
		ui = dynarray_get(&uc->interfaces, intf);
		if (ui) {
			return ui;
		}
	}
	return NULL;
}

static struct usbinterface* usbconfigGetMainInterface(const struct usbconfig* uc, int intf) {
	struct usbinterface* ui;
	
	ui = usbconfigGetInterface(uc, intf);
	if (ui) {
		return ui->parent;
	}
	return NULL;
}

int usbconfigAddInterface(struct usbconfig* uc, struct usbinterface* ui) {
	int i;

	ui->bAlternateSetting = 0;
	ui->parent = ui;
	ui->nextAlternative = NULL;
	i = dynarray_append(&uc->interfaces, ui);
	if (i >= 0) {
		ui->config = uc;
		ui->bInterfaceNumber = i;
	}
	return i;
}

int usbconfigRemoveInterface(struct usbconfig* uc, struct usbinterface* rui) {
	unsigned int i;

	for (i = 0; i < uc->interfaces.len; i++) {
		struct usbinterface* ui;

		ui = usbconfigGetInterface(uc, i);
		if (ui) {
			struct usbinterface* pui;

			pui = ui->parent;
			if (pui == rui) {
				pui = pui->nextAlternative;
				if (pui) {
					if (ui == rui) {
						pui->bAlternateSetting = 0;
						usbinterfaceSelect(pui);
					}
					for (ui = pui; ui; ui = ui->nextAlternative) {
						ui->parent = pui;
					}
					goto found;
				} else {
					dynarray_set(&uc->interfaces, i, NULL);
					while (uc->interfaces.len > 0 && dynarray_get(&uc->interfaces, uc->interfaces.len - 1) == NULL) {
						uc->interfaces.len--;
					}
					goto found;
				}
			} else {
				if (ui == rui) {
					dynarray_set(&uc->interfaces, i, pui);
				}
				for ( ; pui; pui = pui->nextAlternative) {
					if (pui->nextAlternative == rui) {
						pui->nextAlternative = rui->nextAlternative;
						goto found;
					}
				}
			}
		}
	}
	return -ENOENT;
found:;
	rui->nextAlternative = NULL;
	rui->parent = rui;
	rui->config = NULL;
	rui->bInterfaceNumber = -1;
	return 0;
}

struct usblogdev* usbconfigGetLogicalDevice(const struct usbconfig* uc, int logdev) {
	struct usblogdev* ul;
	
	if (logdev >= 0) {
		ul = dynarray_get(&uc->logdevs, logdev);
		if (ul) {
			return ul;
		}
	}
	return NULL;
}

int usbconfigAddLogicalDevice(struct usbconfig* uc, struct usblogdev* ul) {
	return dynarray_append(&uc->logdevs, ul);
}

int
usbconfigRemoveLogicalDevice(struct usbconfig* uc, struct usblogdev* ul)
{
	return dynarray_remove(&uc->logdevs, ul);
}

static struct usbendpoint* usbconfigGetEndpoint(const struct usbconfig* uc, int ep) {
	if (ep & USB_DIR_IN) {
		return dynarray_get(&uc->iendpoints, ep & ~USB_DIR_IN);
	} else {
		return dynarray_get(&uc->oendpoints, ep);
	}
}
	
static int usbconfigIsClaimed(const struct usbconfig* uc) {
	unsigned int intf;
	
	for (intf = 0; intf < uc->interfaces.len; intf++) {
		struct usbinterface* ui;
		
		ui = usbconfigGetMainInterface(uc, intf);
		if (ui) {
			int r;
			
			r = usbinterfaceIsClaimed(ui);
			if (r > 0) {
				return r;
			}
		}
	}
	return 0;
}

int
usbconfigAssignEndpoints(struct usbconfig* uc)
{
	unsigned int intf;
	unsigned int sep;
	unsigned int iFoundEndpoints = 0;
	unsigned int oFoundEndpoints = 0;
	int separateNumbering;
	struct dynarray iendpoints;
	struct dynarray oendpoints;
	struct dynarray unassigned;
	int lowest;
	
	dynarray_init(&unassigned);
	dynarray_init(&iendpoints);
	dynarray_append(&iendpoints, NULL);	/* ZERO endpoint is special - for now we do not support BULK/INTERRUPT on it */
	dynarray_init(&oendpoints);
	dynarray_append(&oendpoints, NULL);	/* ZERO endpoint is special - for now we do not support BULK/INTERRUPT on it */
	for (intf = 0; intf < uc->interfaces.len; intf++) {
		struct usbinterface* ui;
		
		for (ui = usbconfigGetMainInterface(uc, intf); ui; ui = ui->nextAlternative) {
			unsigned int epcnt;

			for (epcnt = 0; epcnt < ui->endpoints.len; epcnt++) {
				struct usbendpoint* ue;
		
				ue = usbinterfaceGetEndpoint(ui, epcnt);
				if (ue) {
					if (ue->bEndpointAddress & USB_DIR_IN) {
						iFoundEndpoints++;
					} else {
						oFoundEndpoints++;
					}
					if ((ue->bEndpointAddress & 0x0F) == 0) {
						dynarray_append(&unassigned, ue);
					} else {
						int err;

						if (ue->bEndpointAddress & USB_DIR_IN) {
							err = dynarray_putat(&iendpoints, ue, ue->bEndpointAddress & 0x0F);
						} else {
							err = dynarray_putat(&oendpoints, ue, ue->bEndpointAddress & 0x0F);
						}
						if (err < 0) {
							dynarray_done(&iendpoints);
							dynarray_done(&oendpoints);
							dynarray_done(&unassigned);
							return err;
						}
					}
				}
			}
		}
	}
	if (iFoundEndpoints > 15 || oFoundEndpoints > 15) {
		dynarray_done(&iendpoints);
		dynarray_done(&oendpoints);
		dynarray_done(&unassigned);
		return -EBUSY;
	}
	separateNumbering = (iFoundEndpoints + oFoundEndpoints > 15);
	lowest = 1;
	for (sep = 0; sep < unassigned.len; sep++) {
		struct usbendpoint* ue;
		
		ue = dynarray_get(&unassigned, sep);
		if (ue) {
			struct dynarray* da;
			int ep;

			if (!separateNumbering) {
				for (; lowest < 16; lowest++) {
					if (dynarray_get(&iendpoints, lowest) == NULL && dynarray_get(&oendpoints, lowest) == NULL) {
						break;
					}
				}
			}
			if (ue->bEndpointAddress & USB_DIR_IN) {
				da = &iendpoints;
			} else {
				da = &oendpoints;
			}
			ep = dynarray_put(da, ue, lowest);
			if (ep > 15) {
				dynarray_set(da, ep, NULL);
				ep = -EBUSY;
			}
			if (ep < 0) {
				ep = dynarray_put(da, ue, 1);
			}
			if (ep > 15) {
				ep = -EBUSY;
			}
			if (ep < 0) {
				dynarray_done(&iendpoints);
				dynarray_done(&oendpoints);
				dynarray_done(&unassigned);
				return ep;
			}
			ue->bEndpointAddress |= ep;
			lowest = ep;
		}
	}
	dynarray_done(&uc->oendpoints);
	dynarray_done(&uc->iendpoints);
	uc->oendpoints = oendpoints;
	uc->iendpoints = iendpoints;
	dynarray_done(&unassigned);
	return 0;
}

static int init_usb_config_descriptor(struct usbconfig* uc, unsigned int idx, void* buffer) {
        struct usb_config_descriptor* ucd = buffer;

        memset(ucd, 0, sizeof(*ucd));
        ucd->bLength = USB_DT_CONFIG_SIZE;
        ucd->bDescriptorType = USB_DT_CONFIG;
        ucd->wTotalLength = 0; //Set by caller
        ucd->bNumInterfaces = uc->interfaces.len;
        ucd->bConfigurationValue = idx;
        ucd->iConfiguration = 0; //FIXME
        ucd->bmAttributes = USB_CONFIG_ATT_ONE | USB_CONFIG_ATT_SELFPOWER;
        ucd->bMaxPower = 0;
	return ucd->bLength;
}

static int usbconfigGetConfigDescriptor(struct usbconfig* uc, unsigned int idx, size_t size, void* data) {
	unsigned char buffer[4096];
	int len;
	unsigned int j;
	struct usb_config_descriptor* ucd;
				
	len = init_usb_config_descriptor(uc, idx, buffer);
	for (j = 0; j < uc->interfaces.len; j++) {
		struct usbinterface* ui;
		
		for (ui = usbconfigGetMainInterface(uc, j); ui; ui = ui->nextAlternative) {
			len += usbinterfaceGetDescriptor(ui, buffer + len);
		}
	}
	ucd = (struct usb_config_descriptor*)buffer;
	ucd->wTotalLength = len;
	return usb_copy_desc(size, data, len, buffer);
}

static int
usbconfigDiscardUrb(struct usbconfig* uc, struct usbdevfs_urb* urb)
{
	unsigned int j;

	for (j = 0; j < uc->oendpoints.len; j++) {
		struct usbendpoint* ue = usbconfigGetEndpoint(uc, j);
		
		if (ue) {
			if (usbendpointDiscardUrb(ue, urb) == 0) {
				return 0;
			}
		}
	}
	for (j = 0; j < uc->iendpoints.len; j++) {
		struct usbendpoint* ue = usbconfigGetEndpoint(uc, j | USB_DIR_IN);
		
		if (ue) {
			if (usbendpointDiscardUrb(ue, urb) == 0) {
				return 0;
			}
		}
	}
	return -ENOENT;
}

static void
usbconfigForkCleanup(struct usbconfig* uc)
{
	unsigned int i;

	for (i = 0; i < uc->interfaces.len; i++) {
		struct usbinterface* ui;

		for (ui = usbconfigGetMainInterface(uc, i); ui; ui = ui->nextAlternative) {
			usbinterfaceForkCleanup(ui);
		}
	}
	for (i = 0; i < uc->logdevs.len; i++) {
		struct usblogdev* ul;
		
		ul = usbconfigGetLogicalDevice(uc, i);
		if (ul) {
			if (ul->forkCleanup) {
				ul->forkCleanup(ul);
			}
		}
	}
	if (uc->fns && uc->fns->forkCleanup) {
		uc->fns->forkCleanup(uc->configCookie);
	}
	dynarray_done(&uc->interfaces);
	dynarray_done(&uc->iendpoints);
	dynarray_done(&uc->oendpoints);
	dynarray_done(&uc->logdevs);
}

void usbdeviceInit(struct usbdevice* ud) {
	memset(ud, 0, sizeof(*ud));
	dynarray_init(&ud->configs);
	ud->activeConfig = 0;
	dynarray_init(&ud->strings);
	pthread_mutex_init(&ud->stringsLock, NULL);
	ud->desc.bLength = USB_DT_DEVICE_SIZE;
	ud->desc.bDescriptorType = USB_DT_DEVICE;
}

static struct usbconfig* usbdeviceGetConfig(const struct usbdevice* ud, int cfg) {
	return dynarray_get(&ud->configs, cfg);
}

struct usbconfig* usbdeviceCurrentConfig(const struct usbdevice* ud) {
	return usbdeviceGetConfig(ud, ud->activeConfig);
}

int usbdeviceAddConfig(struct usbdevice* ud, struct usbconfig* uc) {
	return dynarray_append(&ud->configs, uc);
}

struct usbinterface* usbdeviceGetInterface(struct usbdevice* ud, int intf) {
	struct usbconfig* uc;
	
	uc = usbdeviceCurrentConfig(ud);
	if (!uc) {
		return NULL;
	}
	return usbconfigGetInterface(uc, intf);
}

static struct usbendpoint* usbdeviceGetEndpoint(const struct usbdevice* ud, int ep) {
	struct usbconfig* uc;
	
	uc = usbdeviceCurrentConfig(ud);
	if (!uc) {
		return NULL;
	}
	return usbconfigGetEndpoint(uc, ep);
}

static int usbdeviceDeviceGetDescriptor(struct usbdevice* ud, unsigned int type, unsigned int idx, unsigned int language, size_t size, void* data) {
	switch (type) {
		case USB_DT_DEVICE:
			{
				if (idx == 0 && language == 0) {
					ud->desc.bNumConfigurations = ud->configs.len;
					return usb_copy_desc(size, data, ud->desc.bLength, &ud->desc);
				}
				return -EPIPE;
			}
		case USB_DT_CONFIG:
			{
				struct usbconfig* uc;
				
				if (language) {
					usb_log("Invalid language %04X for DT_CONFIG\n", language);
					return -EPIPE;
				}
				uc = usbdeviceGetConfig(ud, idx);
				if (!uc) {
					usb_log("Invalid configuration index %02X\n", idx);
					return -EPIPE;
				}
				return usbconfigGetConfigDescriptor(uc, idx + 1, size, data);
			}
		case USB_DT_STRING:
			{
				if (idx == 0) {
					if (language == 0) {
						unsigned char buffer[16];
						struct usb_string_descriptor* u = (struct usb_string_descriptor*)buffer;

						u->bLength         = 4;
						u->bDescriptorType = USB_DT_STRING;
						u->wData[0]        = USB_LANGID_en_US;
						return usb_copy_desc(size, data, u->bLength, u);
					}
				} else if (language == USB_LANGID_en_US) {
					struct usbstr* usbstr;
					int ret;

					pthread_mutex_lock(&ud->stringsLock);
					usbstr = dynarray_get(&ud->strings, idx - 1);
					if (usbstr) {
						ret = usb_copy_desc(size, data, usbstr->data[0] & 0xFF, usbstr->data);
					} else {
						ret = -EPIPE;
					}
					pthread_mutex_unlock(&ud->stringsLock);
					return ret;
				}
				return -EPIPE;
			}
	}
	if (ud->fns && ud->fns->getdesc) {
		return ud->fns->getdesc(ud->devcookie, type, idx, language, size, data);
	}
	return -EPIPE;
}

static int usbdeviceDeviceControl(struct usbdevice* ud, struct usbdevfs_ctrltransfer* ct) {
	switch (ct->bRequestType) {
		case USB_DIR_IN | USB_TYPE_STANDARD | USB_RECIP_DEVICE:
			switch (ct->bRequest) {
				case USB_REQ_GET_DESCRIPTOR:
					return usbdeviceDeviceGetDescriptor(ud, ct->wValue >> 8, ct->wValue & 0xFF, ct->wIndex, ct->wLength, ct->data);
				default:
					usb_log("Unexpected request %02X for IN/STD/DEVICE\n", ct->bRequest);
					break;
			}
			break;
		case USB_DIR_OUT | USB_TYPE_STANDARD | USB_RECIP_DEVICE:
			switch (ct->bRequest) {
				case USB_REQ_SET_CONFIGURATION:
					if (ct->wValue == ud->activeConfig + 1) {
						return 0;
					}
					/* FIXME: support alternative configurations */
					return -EPIPE;
				default:
					usb_log("Unexpected request %02X for OUT/STD/DEVICE\n", ct->bRequest);
					break;
			}
			break;
	}
	return -EPIPE;
}

int usbdeviceControl(void* cookie, struct usbdevfs_ctrltransfer* ct) {
	struct usbdevice* ud = cookie;

	switch (ct->bRequestType & USB_RECIP_MASK) {
		case USB_RECIP_DEVICE:
			return usbdeviceDeviceControl(ud, ct);
		case USB_RECIP_INTERFACE:
			{
				struct usbinterface* ui;
				
				ui = usbdeviceGetInterface(ud, ct->wIndex);
				if (!ui) {
					return -ENOENT;
				}
				return usbinterfaceDeviceControl(ui, ct);
			}
		case USB_RECIP_ENDPOINT:
			{
				struct usbendpoint* ue;
				
				ue = usbdeviceGetEndpoint(ud, ct->wIndex);
				if (!ue) {
					return -ENOENT;
				}
				return usbendpointDeviceControl(ue, ct);
			}
		case USB_RECIP_OTHER:
		default:
			usb_log("Unexpected recipient %X-%02X-%02X\n", ct->bRequestType >> 7, (ct->bRequestType >> 5) & 3, ct->bRequestType & 0x1F);
			break;
	}
#if 0
	switch (ct->bRequest) {
		case USB_REQ_GET_DESCRIPTOR:
			switch (ct->bRequestType) {
				case USB_DIR_IN | USB_TYPE_STANDARD | USB_RECIP_DEVICE:
					return usbfloppy_GetDescriptor(cookie, ct->wValue >> 8, ct->wValue & 0xFF, ct->wIndex, ct->wLength, ct->data);
				default:
					usb_log("Unexpected request type %02X for GET_DESCRIPTOR\n", ct->bRequestType);
					break;
			}
			break;
		case USB_REQ_GET_STATUS:
		case USB_REQ_CLEAR_FEATURE:
		case USB_REQ_SET_FEATURE:
		case USB_REQ_SET_ADDRESS:
		case USB_REQ_SET_DESCRIPTOR:
		case USB_REQ_GET_CONFIGURATION:
		case USB_REQ_SET_CONFIGURATION:
		case USB_REQ_GET_INTERFACE:
		case USB_REQ_SET_INTERFACE:
		case USB_REQ_SYNCH_FRAME:
			break;
		default:
			usb_log("Invalid request %02X\n", ct->bRequest);
			break;
	}
#endif
	return -EPIPE;
}

static int
usbdeviceXfer(void* cookie,
	      struct urbs* urbs,
	      unsigned int type)
{
	struct usbdevice* ud = cookie;
	struct usbendpoint* ue;
	
	ue = usbdeviceGetEndpoint(ud, urbs->urb->endpoint);
	if (!ue) {
		return -1;
	}
	return usbendpointXfer(ue, urbs, type);
}

static int usbdeviceClaimInterface(void* cookie, int num) {
	struct usbdevice* ud = cookie;
	struct usbconfig* uc;
	struct usbinterface* ui;
	
	uc = usbdeviceCurrentConfig(ud);
	if (!uc) {
		return -ENOENT;
	}
	ui = usbconfigGetMainInterface(uc, num);
	if (!ui) {
		return -ENOENT;
	}
	return usbinterfaceClaim(ui);
}

static int usbdeviceReleaseInterface(void* cookie, int num) {
	struct usbdevice* ud = cookie;
	struct usbconfig* uc;
	struct usbinterface* ui;
	
	uc = usbdeviceCurrentConfig(ud);
	if (!uc) {
		return -ENOENT;
	}
	ui = usbconfigGetMainInterface(uc, num);
	if (!ui) {
		return -ENOENT;
	}
	return usbinterfaceRelease(ui);
}

int usbdeviceIsClaimed(const struct usbdevice* ud) {
	struct usbconfig* uc;
	
	uc = usbdeviceCurrentConfig(ud);
	if (!uc) {
		return -ENOENT;
	}
	return usbconfigIsClaimed(uc);
}

static int usbdeviceAssignEndpoints(struct usbdevice* ud) {
	unsigned int i;
	int err = 0;
	
	for (i = 0; i < ud->configs.len; i++) {
		struct usbconfig* uc;
		int e;
		
		uc = usbdeviceGetConfig(ud, i);
		if (uc) {
			e = usbconfigAssignEndpoints(uc);
			if (err == 0 && e != 0) {
				err = e;
			}
		}
	}
	return err;
}

static int
usbdeviceDiscardUrb(void* cookie, struct usbdevfs_urb* urb)
{
	struct usbdevice* ud = cookie;
	struct usbconfig* uc;
	
	uc = usbdeviceCurrentConfig(ud);
	if (!uc) {
		return -EINVAL;
	}
	return usbconfigDiscardUrb(uc, urb);
}

static void
usbdeviceForkCleanup(void* cookie)
{
	struct usbdevice* ud = cookie;
	unsigned int i;

	for (i = 0; i < ud->configs.len; i++) {
		struct usbconfig* uc;

		uc = usbdeviceGetConfig(ud, i);
		if (uc) {
			usbconfigForkCleanup(uc);
		}
	}
	if (ud->fns && ud->fns->forkCleanup) {
		ud->fns->forkCleanup(ud->devcookie);
	}
	for (i = 0; i < ud->strings.len; i++) {
		struct usbstr* us;

		us = dynarray_get(&ud->strings, i);
		usb_log("Freeing string %p: %u\n", us, i);
		if (us) {
			free(us);
		}
	}
	ud->activeConfig = 0;
	dynarray_init(&ud->configs);
	dynarray_init(&ud->strings);
}

static struct usbdev_fn usbdev_fn = {
	.USBControl		= usbdeviceControl,
	.USBXfer		= usbdeviceXfer,
	.USBClaimInterface	= usbdeviceClaimInterface,
	.USBReleaseInterface	= usbdeviceReleaseInterface,
	.USBDiscardUrb		= usbdeviceDiscardUrb,
	.USBForkCleanup		= usbdeviceForkCleanup,
};

int usbdeviceRegister(struct usbdevice* ud, int busid, int devid) {
	int err;

	err = usbdeviceAssignEndpoints(ud);
	if (err) {
		return err;
	}
	return register_usbdev(ud, &usbdev_fn, busid, devid);
}

struct usbdevice* usbdeviceFind(int busid, int devid) {
	return usbdev_find(busid, devid);
}

int usb_set_string(void* buffer, const char* name) {
	int idx = 0;
	struct usb_string_descriptor* u = buffer;

	u->bDescriptorType = USB_DT_STRING;
	while (*name) {
		u->wData[idx++] = 0xFF & *name++;
	}
	u->bLength = idx * 2 + 2;
	return u->bLength;
}

static size_t
unilen(const u_int16_t* str)
{
	const u_int16_t* curr = str;
	
	while (*curr) {
		curr++;
	}
	return curr - str;
}

static struct usbstr*
usbstrAlloc(size_t len,
	    int constant)
{
	struct usbstr* str;

	if (len >= 127) {
		return NULL;
	}
	len *= sizeof(u_int16_t);
	len += 2;
	str = malloc(sizeof(*str) + len);
	if (str) {
		str->constant = constant;
		str->data[0] = len | (USB_DT_STRING << 8);
	}
	return str;
}

static struct usbstr*
usbstrInitW(const u_int16_t* msg,
	    int constant)
{
	size_t len;
	struct usbstr* str;

	len = unilen(msg);
	str = usbstrAlloc(len, constant);
	if (str) {
		memcpy(str->data + 1, msg, sizeof(u_int16_t) * len);
	}
	return str;
}

static struct usbstr* 
usbstrInit(const char* msg,
	   int constant)
{
	size_t len;
	struct usbstr* str;

	len = strlen(msg);
	str = usbstrAlloc(len, constant);
	if (str) {
		u_int16_t* ptr = str->data + 1;
		
		while (*msg) {
			*ptr++ = *msg++;
		}
	}
	return str;
}

static int
usbdeviceAddUsbStr(struct usbdevice* ud,
		   struct usbstr* usbstr)
{
	int ret;

	if (!usbstr) {
		return -ENOMEM;
	}
	pthread_mutex_lock(&ud->stringsLock);
	if (usbstr->constant) {
		unsigned int idx;
		
		for (idx = 0; idx < ud->strings.len; idx++) {
			struct usbstr* oldusbstr;
			
			oldusbstr = dynarray_get(&ud->strings, idx);
			if (oldusbstr && oldusbstr->constant && oldusbstr->data[0] == usbstr->data[0] && !memcmp(oldusbstr->data, usbstr->data, usbstr->data[0] & 0xFF)) {
				pthread_mutex_unlock(&ud->stringsLock);
				free(usbstr);
				return idx + 1;
			}
		}
	}
	ret = dynarray_append(&ud->strings, usbstr);
	usb_log("Appended string %p: %d\n", usbstr, ret);
	pthread_mutex_unlock(&ud->stringsLock);
	if (ret >= 0) {
		return ret + 1;
	}
	return ret;
}

static int
usbdeviceModifyUsbStr(struct usbdevice* ud,
		      int idx,
		      struct usbstr* usbstr)
{
	struct usbstr* oldusbstr;
	int ret;

	if (!usbstr) {
		return -ENOMEM;
	}
	if (usbstr->constant) {
		free(usbstr);
		return -EINVAL;
	}
	if (idx < 1) {
		free(usbstr);
		return -EINVAL;
	}
	idx--;
	pthread_mutex_lock(&ud->stringsLock);
	oldusbstr = dynarray_get(&ud->strings, idx);
	if (!oldusbstr) {
		ret = -EINVAL;
	} else if (oldusbstr->constant) {
		ret = -EROFS;
	} else {
		ret = dynarray_set(&ud->strings, idx, usbstr);
	}
	usb_log("Modified string %p: %d, old: %p\n", usbstr, ret, oldusbstr);
	pthread_mutex_unlock(&ud->stringsLock);
	if (ret < 0) {
		free(usbstr);
	} else {
		free(oldusbstr);
	}
	return ret;
}

int
usbdeviceRemoveString(struct usbdevice* ud,
		      int idx)
{
	int ret;
	struct usbstr* str;

	pthread_mutex_lock(&ud->stringsLock);
	str = dynarray_get(&ud->strings, idx);
	if (!str) {
		ret = -EINVAL;
	} else if (str->constant) {
		ret = -EROFS;
	} else {
		ret = dynarray_set(&ud->strings, idx, NULL);
	}
	pthread_mutex_unlock(&ud->stringsLock);
	if (ret >= 0) {
		free(str);
	}
	return ret;
}

int usbdeviceAddStringW(struct usbdevice* ud,
			const u_int16_t* msg,
			int constant)
{
	return usbdeviceAddUsbStr(ud, usbstrInitW(msg, constant));
}

int usbdeviceAddString(struct usbdevice* ud,
		       const char* msg,
		       int constant)
{
	return usbdeviceAddUsbStr(ud, usbstrInit(msg, constant));
}

int usbdeviceModifyStringW(struct usbdevice* ud,
			   int id,
			   const u_int16_t* msg)
{
	return usbdeviceModifyUsbStr(ud, id, usbstrInitW(msg, 0));
}

int usbdeviceModifyString(struct usbdevice* ud,
			  int id,
			  const char* msg)
{
	return usbdeviceModifyUsbStr(ud, id, usbstrInit(msg, 0));
}

void
usblogdevInit(struct usblogdev* ul)
{
	memset(ul, 0, sizeof(*ul));
}

void
usblogdevDestroy(struct usblogdev* ul)
{
}
