#define _GNU_SOURCE
#define _LARGEFILE_SOURCE
#define _LARGEFILE64_SOURCE

#include "usbcontrol.h"

#include "dictionary.h"

#include "usbdev.h"

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <pthread.h>

#include <sys/un.h>
#include <sys/socket.h>

#define MAX_USBMASS_DEV	(7)

struct ufifloppyDev;

struct myusbdev {
	struct usbdevice usbdevice;
	struct usbconfig usbconfig;
	struct usblogdev logdev;
	unsigned int id;
	unsigned int devicesCount;
	struct ufifloppyDev *devices[MAX_USBMASS_DEV];	
};

static void setString(unsigned char* dst, size_t len, const char* src) {
	size_t srcl;
	
	srcl = strlen(src);
	if (srcl >= len) {
		memcpy(dst, src, len);
		return;
	}
	memcpy(dst, src, srcl);
	memset(dst + srcl, ' ', len - srcl);
}

struct usbbulkCSW {
	u_int32_t sign;
	u_int32_t tag;
	u_int32_t residue;
	u_int8_t  status;
} __attribute__((packed));

struct usbbulkCBW {
	u_int32_t sign;
	u_int32_t tag;
	u_int32_t length;
	u_int8_t  flags;
	u_int8_t  lun;
	u_int8_t  cbwlen;
	u_int8_t  cbw[16] __attribute__((packed));
} __attribute__((packed));

struct ufifloppyCfg {
	int startConnected;
	char* fileName;
};

struct ufifloppyDevCfg {
	unsigned int transport;
	unsigned int maxLun;
};

enum ub_type { FLOPPY, CDROM };

struct ufifloppy {
	unsigned char senseInfo[18];
	size_t allocLen;
	unsigned char* bufBase;
	size_t pos;
	size_t len;
	int fildes;
	unsigned int unitAttention;
	unsigned int preventRemoval : 1;
	unsigned int writeProtected : 1;
	off64_t diskPos;
	u_int64_t diskLen;
	enum { UB_STATE_CMD = 0, UB_STATE_READ, UB_STATE_WRITE } state;
	struct ufifloppyCfg cfg;
	struct ufifloppyDev* parent;
	struct usblogdev logdev;
	unsigned int		lun;
	unsigned int		sectorSize;
	enum ub_type		deviceType;
};

struct ufifloppyDev {
	struct myusbdev* parent;
	struct ufifloppy *lun[17];
	unsigned int activeLun;
	struct usbendpoint* intrEndpoint;
	struct usbendpoint readEndpoint;
	struct usbendpoint writeEndpoint;
	struct ufifloppyDevCfg cfg;
	struct usblogdev logdev;
	struct usbinterface usbinterface;
	char* configKey;
	enum { UBD_STATE_CMD = 0, UBD_STATE_READ_BUFFER, UBD_STATE_READ_DATA, UBD_STATE_WRITE_DATA, UBD_STATE_STATUS } state;
	struct usbbulkCSW csw;
	struct {
		u_int32_t length;
		u_int32_t translen;
		int	  input;
	} cbw;
	unsigned int iInterface;
	unsigned int id;
};

static int ufifloppy_isErr(const struct ufifloppy* ub) {
	return (ub->senseInfo[2] & 0x0F) != 0;	/* Everything except NO_ERROR */
}

static int ufifloppyIntrProc(void* cookie, size_t len, void* buffer) {
	struct ufifloppyDev* ubd = cookie;
	struct ufifloppy* ub = ubd->lun[ubd->activeLun];
	unsigned char intr[2];
	
	ub->state = UB_STATE_CMD;
	if (0 /* !UFI */) {
		intr[0] = 0x00;
		intr[1] = ufifloppy_isErr(ub) ? 0x03 : 0x00;
	} else {
		intr[0] = ub->senseInfo[12];
		intr[1] = ub->senseInfo[13];
	}
	return usb_copy_desc(len, buffer, 2, intr);
}

static void ufifloppy_invalidCMDNoSig(struct ufifloppy* ub, int skascq, int pos) {
	memset(ub->senseInfo, 0, sizeof(ub->senseInfo));
	ub->senseInfo[0] = 0x00 | 0x70;
	ub->senseInfo[2] = skascq >> 16;	/* FILEMARK ; EOM ; ILI ; ??? ; 4x SENSE KEY */
	setBE32(ub->senseInfo + 3, 0);
	ub->senseInfo[7] = 18 - 8;
	setBE32(ub->senseInfo + 8, 0);
	ub->senseInfo[12] = skascq >> 8;	/* ASC */
	ub->senseInfo[13] = skascq;		/* ASCQ */
	ub->senseInfo[14] = 0x00;		/* FIELD REPLACEABLE UNIT */
	ub->senseInfo[15] = pos >> 16;		/* SKS VALID; COMMAND/DATA\ ; ??? ; ??? ; BPVALID ; 3x BP */
	setBE16(ub->senseInfo + 16, pos);
}

static int
usbbulkfloppyStatusProc(void* cookie,
			size_t len,
			void* buffer)
{
	struct ufifloppyDev* ubd = cookie;
	struct ufifloppy* ub = ubd->lun[ubd->activeLun];

	if (ubd->cfg.transport == 0x50) {
		usbendpointOACK(&ubd->writeEndpoint);
	}
	ubd->state = UB_STATE_CMD;
	ub->state = UB_STATE_CMD;
	return usb_copy_desc(len, buffer, sizeof(ubd->csw), &ubd->csw);
}

static void
ufifloppyStatusPhase(struct ufifloppy* ub) {
	struct ufifloppyDev* ubd = ub->parent;
	
	switch (ubd->cfg.transport) {
		case 0x00:
			usbendpointIACK(&ubd->readEndpoint, NULL, UE_STALL_FOREVER);
			usbendpointIACK(ubd->intrEndpoint, ufifloppyIntrProc, UE_STALL_NONE);
			break;
		case 0x50:
			ubd->csw.residue = ubd->cbw.length - ubd->cbw.translen;
			ubd->csw.status = (ub->senseInfo[2] & 0x0F) ? 0x01 : 0x00;
			if (ubd->csw.residue) {
				if (ubd->cbw.input) {
					usbendpointIACK(&ubd->readEndpoint, usbbulkfloppyStatusProc, UE_STALL_ONCE);
				} else {
					usbendpointIACK(&ubd->readEndpoint, usbbulkfloppyStatusProc, UE_STALL_NONE);
					usbendpointOStall(&ubd->writeEndpoint);
				}
			} else {
				usbendpointIACK(&ubd->readEndpoint, usbbulkfloppyStatusProc, UE_STALL_NONE);
			}
			break;
	}
}

static void
ufifloppyPhaseError(struct ufifloppyDev* ubd, int input) {
	switch (ubd->cfg.transport) {
		case 0x00:
			usbendpointIACK(&ubd->readEndpoint, NULL, UE_STALL_FOREVER);
			usbendpointIACK(ubd->intrEndpoint, ufifloppyIntrProc, UE_STALL_NONE);
			break;
		case 0x50:
			ubd->csw.residue = 0;
			ubd->csw.status = 0x02;
			if (input) {
				usbendpointIACK(&ubd->readEndpoint, usbbulkfloppyStatusProc, UE_STALL_ONCE);
			} else {
				usbendpointIACK(&ubd->readEndpoint, usbbulkfloppyStatusProc, UE_STALL_NONE);
				usbendpointOStall(&ubd->writeEndpoint);
			}
			break;
	}
}

static void ufifloppy_invalidCMD(struct ufifloppy* ub, int skascq, int pos) {
	memset(ub->senseInfo, 0, sizeof(ub->senseInfo));
	ub->senseInfo[0] = 0x00 | 0x70;
	ub->senseInfo[2] = skascq >> 16;	/* FILEMARK ; EOM ; ILI ; ??? ; 4x SENSE KEY */
	setBE32(ub->senseInfo + 3, 0);
	ub->senseInfo[7] = 18 - 8;
	setBE32(ub->senseInfo + 8, 0);
	ub->senseInfo[12] = skascq >> 8;	/* ASC */
	ub->senseInfo[13] = skascq;		/* ASCQ */
	ub->senseInfo[14] = 0x00;		/* FIELD REPLACEABLE UNIT */
	ub->senseInfo[15] = pos >> 16;		/* SKS VALID; COMMAND/DATA\ ; ??? ; ??? ; BPVALID ; 3x BP */
	setBE16(ub->senseInfo + 16, pos);
	ufifloppyStatusPhase(ub);
}

static void ufifloppy_invalidCMDinfo(struct ufifloppy* ub, unsigned int info, int skascq, int pos) {
	memset(ub->senseInfo, 0, sizeof(ub->senseInfo));
	ub->senseInfo[0] = 0x80 | 0x70;
	ub->senseInfo[2] = skascq >> 16;	/* FILEMARK ; EOM ; ILI ; ??? ; 4x SENSE KEY */
	setBE32(ub->senseInfo + 3, info);
	ub->senseInfo[7] = 18 - 8;
	setBE32(ub->senseInfo + 8, 0);
	ub->senseInfo[12] = skascq >> 8;	/* ASC */
	ub->senseInfo[13] = skascq;		/* ASCQ */
	ub->senseInfo[14] = 0x00;		/* FIELD REPLACEABLE UNIT */
	ub->senseInfo[15] = pos >> 16;		/* SKS VALID; COMMAND/DATA\ ; ??? ; ??? ; BPVALID ; 3x BP */
	setBE16(ub->senseInfo + 16, pos);
	ufifloppyStatusPhase(ub);
}

static void ufifloppy_clearError(struct ufifloppy* ub) {
	ufifloppy_invalidCMD(ub, 0, 0);
}

static int ufifloppyReadBuffer(void* cookie, size_t len, void* buffer) {
	struct ufifloppyDev* ubd = cookie;
	struct ufifloppy* ub = ubd->lun[ubd->activeLun];
	size_t mlen;

	usb_log("State = %u\n", ub->state);

	if (ub->state != UB_STATE_CMD) {
		return -EINVAL;
	}
	mlen = ub->len;
	if (ub->len < len) {
		len = ub->len;
	}
	memcpy(buffer, ub->bufBase + ub->pos, len);
	ubd->cbw.translen += len;
	if (len == ub->len) {
		ufifloppy_clearError(ub);
	} else {
		ub->len -= len;
		ub->pos += len;
		usbendpointIACK(&ubd->readEndpoint, ufifloppyReadBuffer, UE_STALL_NONE);
	}
	return len;
}

static void ufifloppy_minlen(struct ufifloppy* ub, size_t minLen) {
	if (ub->allocLen < minLen) {
		free(ub->bufBase);
		ub->bufBase = malloc(minLen);
		if (ub->bufBase) {
			ub->allocLen = minLen;
		}
	}
	ub->len = 0;
}

static unsigned char* ufifloppy_addBuffer(struct ufifloppy* ub, size_t len) {
	unsigned char* ptr;

	if (ub->len + len > ub->allocLen) {
		unsigned char* b;

		b = realloc(ub->bufBase, ub->len + len);
		if (!b) {
			return NULL;
		}
		ub->bufBase = b;
		ub->allocLen = ub->len + len;
	}
	ptr = ub->bufBase + ub->len;
	ub->len += len;
	return ptr;
}
			
static int ufifloppyReadFile(void* cookie, size_t len, void* buffer) {
	struct ufifloppyDev* ubd = cookie;
	struct ufifloppy* ub = ubd->lun[ubd->activeLun];
	size_t toread;

	if (len > ub->diskLen) {
		len = ub->diskLen;
	}
	toread = len;
	while (toread != 0) {
		ssize_t l;

		l = pread64(ub->fildes, buffer, toread, ub->diskPos);
		if (l == -1) {
			ufifloppy_invalidCMDinfo(ub, ub->diskPos / ub->sectorSize, 0x044400, 0x800001);
			return len - toread;
		}
		if (l == 0) {
			ufifloppy_invalidCMDinfo(ub, ub->diskPos / ub->sectorSize, 0x032100, 0x800001);
			return len - toread;
		}
		ubd->cbw.translen += l;
		toread -= l;
		buffer += l;
		ub->diskPos += l;
		ub->diskLen -= l;
	}
	if (ub->diskLen == 0) {
		ufifloppy_clearError(ub);
	} else {
		usbendpointIACK(&ubd->readEndpoint, ufifloppyReadFile, 0);
	}
	return len;
}

static void ufifloppy_dataReady(struct ufifloppy* ub, size_t len) {
	struct ufifloppyDev* ubd;

	ubd = ub->parent;
	if (ubd->cfg.transport == 0x50) {
		if (len && (len > ubd->cbw.length || !ubd->cbw.input)) {
			ufifloppyPhaseError(ubd, ubd->cbw.input);
			return;
		}
		ubd->state = UBD_STATE_READ_BUFFER;
	}
	ub->len = len;
	ub->pos = 0;
	usb_log("Expecting read, %u bytes from buffer\n", len);
	usbendpointIACK(&ubd->readEndpoint, ufifloppyReadBuffer, 0);
}

static void ufifloppy_Inquiry(struct ufifloppy* ub, const unsigned char* cmd, int valid) {
	unsigned char* d;
	int len;

	/* EVDP not supported */
	if (cmd[1] & 0x01) {
		ufifloppy_invalidCMD(ub, 0x052400, 0xC80001);
		return;
	}
	/* CmdDt not supported */
	if (cmd[1] & 0x02) {
		ufifloppy_invalidCMD(ub, 0x052400, 0xC90001);
		return;
	}
	if (cmd[1] & 0x1C) {
		ufifloppy_invalidCMD(ub, 0x052400, 0xCF0001);
		return;
	}
	ufifloppy_minlen(ub, 36);
	d = ub->bufBase;
	
	if (valid) {
		d[0] = (ub->deviceType == CDROM) ? 0x05 : 0x00;
	} else {
		d[0] = (ub->parent->cfg.transport == 0x50) ? 0x1F : 0x7F;	/* 0x7F for SCSI3+ */
	}
	d[1] = 0x80; /* RMB */
	d[2] = 0x00; /* Version */ /* UFI says 0x00... */
	if (ub->parent->cfg.transport == 0x50) {
		/* Bulk transport => should be SCSI-3, so we get 16 LUNs */
		d[2] = 0x02;	/* Use SCSI-2; this is what Linux forces upon us */
	} else {
		d[2] = 0x00;	/* Version: UFI says 0x00... */
	}
	d[3] = 0x01; /* Data format */ /* UFI says 0x01...; SCSI-2 is 0x02 */
	d[4] = 36 - 5; /* Additional Length */
	d[5] = 0x00;
	d[6] = 0x00;
	d[7] = 0x00;
	setString(d + 8, 8,   "VMware");
	setString(d + 16, 16, (ub->deviceType == CDROM) ? "USB CDROM Drive" : "USB Floppy Disk");
	setString(d + 32, 4,  "1.00");
	len = 36;
	if (len > cmd[4]) {
		len = cmd[4];
	}
	ufifloppy_dataReady(ub, len);
}

static void ufifloppy_TestUnitReady(struct ufifloppy* ub, const unsigned char* cmd) {
	ub->len = 0;
	if (ub->fildes != -1) {
		/* GOOD STATUS */
		ufifloppy_clearError(ub);
		return;
	}
	/* MEDIUM NOT PRESENT */
	ufifloppy_invalidCMD(ub, 0x023A00, 0x000000);
}

static void ufifloppy_PreventAllowRemoval(struct ufifloppy* ub, const unsigned char* cmd) {
	ub->len = 0;
	if (cmd[4] & 0xFC) {
		ufifloppy_invalidCMD(ub, 0x052400, 0xCF0004);
		return;
	}
	if (cmd[4] & 0x02) {
		ufifloppy_invalidCMD(ub, 0x052400, 0xC90004);
		return;
	}
	if (ub->fildes == -1) {
		ufifloppy_invalidCMD(ub, 0x023A00, 0x000000);
		return;
	}
	ub->preventRemoval = cmd[4] & 0x03;
	ufifloppy_clearError(ub);
}

static void ufifloppy_RequestSense(struct ufifloppy* ub, const unsigned char* cmd) {
	unsigned int len;

	ufifloppy_minlen(ub, 18);
	memcpy(ub->bufBase, ub->senseInfo, 18);
	len = 18;
	if (len > cmd[4]) {
		len = cmd[4];
	}
	ufifloppy_dataReady(ub, len);
}

static void ufifloppy_ReadCapacity(struct ufifloppy* ub, const unsigned char* cmd) {
	off64_t len;

	if (cmd[8] & 0x01) {
		ufifloppy_invalidCMD(ub, 0x052400, 0xC80008);
		return;
	}
	if (cmd[1] & 0x01) {
		ufifloppy_invalidCMD(ub, 0x052400, 0xC80001);
		return;
	}
	if (getBE32(cmd + 2)) {
		ufifloppy_invalidCMD(ub, 0x052400, 0xCF0002);
		return;
	}
	if (ub->fildes == -1) {
		ufifloppy_invalidCMD(ub, 0x023A00, 0x000000);
		return;
	}
	len = lseek64(ub->fildes, 0, SEEK_END);
	if (len == (off64_t)-1) {
		/* INTERNAL TARGET FAILURE */
		ufifloppy_invalidCMD(ub, 0x044400, 0x000000);
		return;
	}
	ufifloppy_minlen(ub, 8);
	setBE32(ub->bufBase, (len / ub->sectorSize) - 1);
	setBE32(ub->bufBase + 4, ub->sectorSize);
	
	ufifloppy_dataReady(ub, 8);
}

static void ufifloppy_ReadFormatCapacities(struct ufifloppy* ub, const unsigned char* cmd) {
	unsigned int len;
	unsigned char *b;

	ufifloppy_minlen(ub, 12);
	b = ub->bufBase;
	setBE32(b + 0, 8);
	if (ub->fildes == -1) {
		setBE32(b + 4, 2880);
		setBE32(b + 8, 0x03000000 | ub->sectorSize);	/* Max drive capacity */
	} else {
		off64_t size;
		
		size = lseek64(ub->fildes, 0, SEEK_END);
		if (size == (off64_t)-1) {
			setBE32(b + 4, 2880);
			setBE32(b + 8, 0x01000000 | ub->sectorSize);	/* Max size for current format */
		} else {
			setBE32(b + 4, size / ub->sectorSize);
			setBE32(b + 8, 0x02000000 | ub->sectorSize);	/* Current format */
		}
	}
	len = getBE16(cmd + 7);
	if (len > 12) {
		len = 12;
	}
	ufifloppy_dataReady(ub, len);
}

static void
ufifloppyRead(struct ufifloppy* ub,
	      u_int32_t block,
	      u_int32_t len)
{
	if (ub->fildes == -1) {
		ufifloppy_invalidCMD(ub, 0x023A00, 0x000000);
		return;
	}
	ub->diskPos = (u_int64_t)block * ub->sectorSize;
	ub->diskLen = (u_int64_t)len * ub->sectorSize;
	if (ub->parent->cfg.transport == 0x50 && len && (ub->diskLen > ub->parent->cbw.length || !ub->parent->cbw.input)) {
		ufifloppyPhaseError(ub->parent, ub->parent->cbw.input);
		return;
	}
	ub->state = UB_STATE_READ;
	ub->parent->state = UBD_STATE_READ_DATA;
	posix_fadvise(ub->fildes, ub->diskPos, ub->diskLen, POSIX_FADV_WILLNEED);
	usbendpointIACK(&ub->parent->readEndpoint, ufifloppyReadFile, 0);
}

static void ufifloppy_Read(struct ufifloppy* ub, const unsigned char* cmd) {
	if (cmd[1] & 0x01) {
		ufifloppy_invalidCMD(ub, 0x052400, 0xC80001);
		return;
	}
	if (cmd[1] & 0x08) {
		ufifloppy_invalidCMD(ub, 0x052400, 0xC80003);
		return;
	}
	if (cmd[1] & 0x10) {
		ufifloppy_invalidCMD(ub, 0x052400, 0xC80004);
		return;
	}
	if (cmd[1] & 0x06) {
		ufifloppy_invalidCMD(ub, 0x052400, 0xCA0001);
		return;
	}
	ufifloppyRead(ub, getBE32(cmd + 2), getBE16(cmd + 7));
}

static void ufifloppy_Read12(struct ufifloppy* ub, const unsigned char* cmd) {
	if (cmd[1] & 0x01) {
		ufifloppy_invalidCMD(ub, 0x052400, 0xC80001);
		return;
	}
	if (cmd[1] & 0x08) {
		ufifloppy_invalidCMD(ub, 0x052400, 0xC80003);
		return;
	}
	if (cmd[1] & 0x10) {
		ufifloppy_invalidCMD(ub, 0x052400, 0xC80004);
		return;
	}
	if (cmd[1] & 0x06) {
		ufifloppy_invalidCMD(ub, 0x052400, 0xCA0001);
		return;
	}
	ufifloppyRead(ub, getBE32(cmd + 2), getBE32(cmd + 6));
}

static void 
ufifloppyWrite(struct ufifloppy* ub, 
	       u_int32_t block, 
	       u_int32_t count)
{
	if (ub->fildes == -1) {
		ufifloppy_invalidCMD(ub, 0x023A00, 0x000000);
		return;
	}
	if (ub->writeProtected || ub->deviceType == CDROM) {
		ufifloppy_invalidCMD(ub, 0x072701, 0x000000);
		return;
	}
	ub->diskPos = (u_int64_t)block * ub->sectorSize;
	ub->diskLen = (u_int64_t)count * ub->sectorSize;
	if (ub->parent->cfg.transport == 0x50 && count && (ub->diskLen > ub->parent->cbw.length || ub->parent->cbw.input)) {
		ufifloppyPhaseError(ub->parent, ub->parent->cbw.input);
		return;
	}
	ub->state = UB_STATE_WRITE;
	ub->parent->state = UBD_STATE_WRITE_DATA;
}

static void ufifloppy_Write(struct ufifloppy* ub, const unsigned char* cmd) {
	if (cmd[1] & 0x01) {
		ufifloppy_invalidCMD(ub, 0x052400, 0xC80001);
		return;
	}
	if (cmd[1] & 0x08) {
		ufifloppy_invalidCMD(ub, 0x052400, 0xC80003);
		return;
	}
	if (cmd[1] & 0x10) {
		ufifloppy_invalidCMD(ub, 0x052400, 0xC80004);
		return;
	}
	if (cmd[1] & 0x06) {
		ufifloppy_invalidCMD(ub, 0x052400, 0xCA0001);
		return;
	}
	ufifloppyWrite(ub, getBE32(cmd + 2), getBE16(cmd + 7));
}

static void ufifloppy_Write12(struct ufifloppy* ub, const unsigned char* cmd) {
	if (cmd[1] & 0x01) {
		ufifloppy_invalidCMD(ub, 0x052400, 0xC80001);
		return;
	}
	if (cmd[1] & 0x08) {
		ufifloppy_invalidCMD(ub, 0x052400, 0xC80003);
		return;
	}
	if (cmd[1] & 0x10) {
		ufifloppy_invalidCMD(ub, 0x052400, 0xC80004);
		return;
	}
	if (cmd[1] & 0x06) {
		ufifloppy_invalidCMD(ub, 0x052400, 0xCA0001);
		return;
	}
	ufifloppyWrite(ub, getBE32(cmd + 2), getBE32(cmd + 6));
}

static void ufifloppy_StartStopUnit(struct ufifloppy* ub, const unsigned char* cmd) {
	if ((cmd[4] & 0xF0) != 0x00) {
		/* Change in power conditions... */
		ufifloppy_clearError(ub);
		return;
	}
	if ((cmd[4] & 0x02) == 0x00) {
		/* Only start/stop... */
		ufifloppy_clearError(ub);
		return;
	}
	if (cmd[4] & 1) {
		/* Load disc... */
	} else {
		/* Unload disc... */
	}
	ufifloppy_clearError(ub);
}

static unsigned int lba2msf(unsigned int lba) {
	unsigned int msf;

	lba += 150;

	msf = lba % 75;
	lba = lba / 75;
	msf |= (lba % 60) << 8;
	lba = lba / 60;
	msf |= lba << 16;
	return msf;
}

static void ufifloppy_ReadTOCPMAATIP(struct ufifloppy* ub, const unsigned char* cmd) {
	unsigned int msf = (cmd[1] & 0x02) != 0;
	unsigned int fmt = (cmd[2] & 0x0F);
	unsigned int trk = cmd[6];
	unsigned int ml  = getBE16(cmd + 7);
	unsigned char* b;
	off64_t len;

	if (ub->fildes == -1) {
		ufifloppy_invalidCMD(ub, 0x023A00, 0x000000);
		return;
	}
	usb_log("READTOC fmt=%X, msf=%u, trk=%02X, len=%u\n", fmt, msf, trk, ml);
	switch (fmt) {
		case 0x0:
			ufifloppy_minlen(ub, 20);
			b = ub->bufBase;
			setBE16(b, 16);
			b[2] = 0x01;
			b[3] = 0x01;

			len = 0;
			if (msf) {
				len = lba2msf(len);
			}
			b[4] = 0x00;
			b[5] = 0x14;
			b[6] = 0x01;
			b[7] = 0x00;
			setBE32(b + 8, len);

			len = lseek64(ub->fildes, 0, SEEK_END);
			if (len == (off64_t)-1) {
				len = 0;
			}
			if (msf) {
				len = lba2msf(len);
			}
			b[12] = 0x00;
			b[13] = 0x14;
			b[14] = 0xAA;
			b[15] = 0x00;
			setBE32(b + 16, len);

			if (ml > 20) {
				ml = 20;
			}
			ufifloppy_dataReady(ub, ml);
			return;
	}
	ufifloppy_invalidCMD(ub, 0x052000, 0xCF0000);
}

static int putModeSense(struct ufifloppy* ub, unsigned int page) {
	unsigned char* b;
	int val = page & 0xC0;

	switch (page & 0x3F) {
		case 0x2A:
			b = ufifloppy_addBuffer(ub, 22);
			b[0] = 0x2A;
			b[1] = 20;
			b[2] = 0x07;	/* Method 2, CDRW-read, CDR-read */
			b[3] = 0x00;	/* No write... */
			b[4] = 0x30;	/* No audio caps, no multisession */
			b[5] = 0x00;	/* No CDDA */
			b[6] = 0x00;	/* FIX! */
			b[7] = 0x00;	/* Changer, volume */
			setBE16(b + 8, 176 * 40);
			setBE16(b + 10, 1);
			setBE16(b + 12, 8192);
			setBE16(b + 14, 176 * 40);
			b[16] = 0x00;
			b[17] = 0x00;
			setBE16(b + 18, 0);
			setBE16(b + 20, 0);
			return 0;
		default:
			return -EINVAL;
	}
}

static void ufifloppy_ModeSense6(struct ufifloppy* ub, const unsigned char* cmd) {
	unsigned int pc = cmd[2];
	unsigned int ml = cmd[4];
	unsigned char* b;
	unsigned int ln;

	ufifloppy_minlen(ub, 20);
	b = ufifloppy_addBuffer(ub, 4);
	b[0] = 0x00;	/* replace with length */
	b[1] = 0x01;	/* medium type... */
	b[2] = 0x00;	/* device specific */
	b[3] = 0x00;	/* descriptors length */

	if ((pc & 0x3F) == 0x3F) {
		unsigned int i;

		pc &= ~0x3F;
		for (i = 0x01; i < 0x40; i++) {
			putModeSense(ub, pc | i);
		}
		putModeSense(ub, pc);
	} else {
		if (putModeSense(ub, pc)) {
			ufifloppy_invalidCMD(ub, 0x052000, 0xCF0000);	/* FIX! Invalid Mode page requested */
			return;
		}
	}
	ln = ub->len - 1;
	if (ln > 255) {
		ufifloppy_invalidCMD(ub, 0x052000, 0xCF0000);	/* FIX! */
		return;
	}
	ub->bufBase[0] = ln;
	ln++;
	if (ml > ln) {
		ml = ln;
	}
	ufifloppy_dataReady(ub, ml);
}

static void ufifloppy_ModeSense10(struct ufifloppy* ub, const unsigned char* cmd) {
	unsigned int pc = cmd[2];
	unsigned int ml = getBE16(cmd + 7);
	unsigned char* b;
	unsigned int ln;

	usb_log("Requested page %02X, max length %u\n", pc, ml);
	ufifloppy_minlen(ub, 20);
	b = ufifloppy_addBuffer(ub, 8);
	b[0] = 0x00;	/* replace with length */
	b[1] = 0x00;
	b[2] = 0x01;	/* medium type... */
	b[3] = 0x00;	/* device specific */
	b[4] = 0x00;	/* reserved */
	b[5] = 0x00;	/* reserved */
	setBE16(b + 6, 0);	/* descriptors length */

	if ((pc & 0x3F) == 0x3F) {
		unsigned int i;

		pc &= ~0x3F;
		for (i = 0x01; i < 0x40; i++) {
			putModeSense(ub, pc | i);
		}
		putModeSense(ub, pc);
	} else {
		if (putModeSense(ub, pc)) {
			ufifloppy_invalidCMD(ub, 0x052000, 0xCF0000);	/* FIX! Invalid Mode page requested */
			return;
		}
	}
	ln = ub->len;
	setBE16(ub->bufBase, ln - 2);
	if (ml > ln) {
		ml = ln;
	}
	ufifloppy_dataReady(ub, ml);
}

static void ufifloppy_SendDiagnostics(struct ufifloppy* ub, const unsigned char* cmd) {
	ufifloppy_clearError(ub);
}

static void ufifloppyCmd(struct ufifloppy* ub, const unsigned char* data) {
	usb_log("SCSI Command %02X issued on LUN %u\n", data[0], ub->lun);
	ub->state = UB_STATE_CMD;
	ub->len = 0;
	if (ub->unitAttention && data[0] != 0x12 && data[0] != 0x5A && data[0] != 0x1A) {
		ufifloppy_invalidCMDNoSig(ub, ub->unitAttention, 0x000000);
		ub->unitAttention = 0;
	}
	if (ufifloppy_isErr(ub)) {
		switch (data[0]) {
			case 0x03:
			case 0x12:
			case 0x1D:
				break;
			default:
				/* There is something in sense info already */
				ufifloppyStatusPhase(ub);
				return;
		}
	}
	switch (data[0]) {
		case 0x00:	/* TEST UNIT READY */
			ufifloppy_TestUnitReady(ub, data);
			break;
		case 0x03:	/* REQUEST SENSE */
			ufifloppy_RequestSense(ub, data);
			break;
		case 0x12:	/* INQUIRY */
			ufifloppy_Inquiry(ub, data, 1);
			break;
		case 0x1A:	/* MODE SENSE(6) */
			ufifloppy_ModeSense6(ub, data);
			break;
		case 0x1B:	/* START/STOP UNIT */
			ufifloppy_StartStopUnit(ub, data);
			break;
		case 0x1D:	/* SEND DIAGNOSTICS */
			ufifloppy_SendDiagnostics(ub, data);
			break;
		case 0x1E:	/* PREVENT/ALLOW MEDIA REMOVAL */
			ufifloppy_PreventAllowRemoval(ub, data);
			break;
		case 0x23:	/* READ FORMAT CAPACITIES */
			ufifloppy_ReadFormatCapacities(ub, data);
			break;
		case 0x25:	/* READ CAPACITY */
			ufifloppy_ReadCapacity(ub, data);
			break;
		case 0x28:	/* READ(10) */
			ufifloppy_Read(ub, data);
			break;
		case 0x2A:	/* WRITE(10) */
			ufifloppy_Write(ub, data);
			break;
		case 0x43:	/* READ TOC/PMA/ATIP */
			ufifloppy_ReadTOCPMAATIP(ub, data);
			break;
		case 0x5A:	/* MODE SENSE(10) */
			ufifloppy_ModeSense10(ub, data);
			break;
		case 0xA8:	/* READ(12) */
			ufifloppy_Read12(ub, data);
			break;
		case 0xAA:	/* WRITE(12) */
			ufifloppy_Write12(ub, data);
			break;
		default:
			usb_warn("Unknown SCSI command %02X issued\n", data[0]);
			ufifloppy_invalidCMD(ub, 0x052000, 0xCF0000);
			break;
	}
}

static int ufifloppy_SCSICmd(struct ufifloppyDev* ubd, int wValue, int size, unsigned char* data) {
	unsigned int lun;
	struct ufifloppy* ub;

	usbendpointINAK(ubd->intrEndpoint);
	usbendpointIACK(&ubd->readEndpoint, NULL, 3);
	if (size != 12) {
		return -EPIPE;
	}
	ubd->activeLun = 16;
	lun = data[1] >> 5;
	ub = ubd->lun[lun];
	if (!ub) {
		ub = ubd->lun[ubd->activeLun];
		ub->len = 0;
		switch (data[0]) {
			case 0x12:
				ufifloppy_Inquiry(ub, data, 0);
				break;
			case 0x03:
				ufifloppy_RequestSense(ub, data);
				break;
			default:
				ufifloppy_invalidCMD(ub, 0x052500, 0x000000);
				break;
		}
		return size;
	}
	ubd->activeLun = lun;
	ufifloppyCmd(ub, data);
	return size;
}

static int ufifloppy_SCSICommand(void* cookie, struct usbdevfs_ctrltransfer* ct) {
	switch (ct->bRequestType) {
		case USB_DIR_OUT | USB_TYPE_CLASS | USB_RECIP_INTERFACE:
			switch (ct->bRequest) {
				case 0x00:
					return ufifloppy_SCSICmd(cookie, ct->wValue, ct->wLength, ct->data);
				default:
					usb_log("Unexpected request %02X for OUT/CLASS/INTERFACE\n", ct->bRequest);
					break;
			}
			break;
		default:
			usb_log("Unexpected request type %02X\n", ct->bRequestType);
			break;
	}
	return -EPIPE;
}

static int ufifloppy_intfGetDescriptor(void* cookie, struct usb_interface_descriptor* uid) {
	struct ufifloppyDev* ubd = cookie;

        uid->bInterfaceClass = USB_CLASS_MASS_STORAGE;
        uid->bInterfaceSubClass = 0x04; /* UFI */
        uid->bInterfaceProtocol = ubd->cfg.transport;
        uid->iInterface = ubd->iInterface;
	return 0;
}

static int ufifloppy_writepipeproc(void* cookie, size_t len, const void* buffer) {
	struct ufifloppyDev* ubd = cookie;
	struct ufifloppy* ub = ubd->lun[ubd->activeLun];
	size_t towrite;

	if (ub->state != UB_STATE_WRITE) {
		return -EPIPE;
	}
	if (len > ub->diskLen) {
		len = ub->diskLen;
	}
	towrite = len;
	while (towrite != 0) {
		ssize_t l;
	
		usb_log("Writting %u bytes at %Lu\n", towrite, ub->diskPos);
		l = pwrite64(ub->fildes, buffer, towrite, ub->diskPos);
		if (l == -1) {
			ufifloppy_invalidCMDinfo(ub, ub->diskPos / ub->sectorSize, 0x044400, 0x800001);
			return len - towrite;
		}
		if (l == 0) {
			ufifloppy_invalidCMDinfo(ub, ub->diskPos / ub->sectorSize, 0x044400, 0x800001);
			return len - towrite;
		}
		ubd->cbw.translen += l;
		towrite -= l;
		buffer += l;
		ub->diskPos += l;
		ub->diskLen -= l;
	}
	if (ub->diskLen == 0) {
		ufifloppy_clearError(ub);
	} else {
//		usbendpointOACK(ubd->writeEndpoint, ufifloppy_writepipeproc);
	}
	usb_log("Returning length %u\n", len);
	return len;
}

static struct usbinterface_fns ufifloppy_intffns = {
	.intfDeviceControl = ufifloppy_SCSICommand,
	.intfGetDescriptor = ufifloppy_intfGetDescriptor,
};

static int
usbbulkExecCmd(struct ufifloppy* ub, 
	       size_t cbwlen,
	       const unsigned char* cbw)
{
	if (cbwlen != 12) {
		usb_log("Bad command length: %u != %u\n", cbwlen, 12);
		return -EPIPE;
	}
	ufifloppyCmd(ub, cbw);
	return 0;
}
	       
static int
usbbulkfloppy_writepipeproc(void* cookie, 
			    size_t len, 
			    const void* buffer)
{
	struct ufifloppyDev* ubd = cookie;
	
	if (ubd->state == UBD_STATE_CMD) {
		const struct usbbulkCBW* b = buffer;
		unsigned int lun;
		struct ufifloppy* ub;
		int r;

		if (len != sizeof(*b)) {
			usb_log("Bad length in bulk command write: %u != %u\n", len, sizeof(*b));
			ufifloppyPhaseError(ubd, 1);
			return -EPIPE;
		}
		if (b->sign != 0x43425355) {
			usb_log("Bad signature in command: %08X != %08X\n", b->sign, 0x43425355);
			ufifloppyPhaseError(ubd, 1);
			return -EPIPE;
		}
		lun = b->lun & 0x0F;
		ub = ubd->lun[lun];
		if (!ub) {
			usb_log("Bad lun: %u does not exist\n", b->lun & 0x0F);

			ufifloppyPhaseError(ubd, 1);
			return -EPIPE;
		}
		ubd->activeLun = lun;
		ubd->csw.tag = b->tag;
		ubd->cbw.length = b->length;
		ubd->cbw.input = (b->flags & 0x80) ? 1 : 0;
		ubd->cbw.translen = 0;
		r = usbbulkExecCmd(ub, b->cbwlen & 0x1F, b->cbw);
		if (r < 0) {
			usb_log("Command %02X failed: %d\n", b->cbw[0], r);
			ufifloppyPhaseError(ubd, 1);
			return r;
		}
		return 31;
	}
	if (ubd->state == UBD_STATE_WRITE_DATA) {
		struct ufifloppy* ub = ubd->lun[ubd->activeLun];
		size_t towrite;

		if (ub->state != UB_STATE_WRITE) {
			return -EPIPE;
		}
		if (len > ub->diskLen) {
			len = ub->diskLen;
		}
		towrite = len;
		while (towrite != 0) {
			ssize_t l;
	
			usb_log("Writting %u bytes at %Lu\n", towrite, ub->diskPos);
			l = pwrite64(ub->fildes, buffer, towrite, ub->diskPos);
			if (l == -1) {
				ufifloppy_invalidCMDinfo(ub, ub->diskPos / ub->sectorSize, 0x044400, 0x800001);
				return len - towrite;
			}
			if (l == 0) {
				ufifloppy_invalidCMDinfo(ub, ub->diskPos / ub->sectorSize, 0x044400, 0x800001);
				return len - towrite;
			}
			ubd->cbw.translen += l;
			towrite -= l;
			buffer += l;
			ub->diskPos += l;
			ub->diskLen -= l;
		}
		if (ub->diskLen == 0) {
			ufifloppy_clearError(ub);
		} else {
//			usbendpointOACK(ubd->writeEndpoint, ufifloppy_writepipeproc);
		}
		usb_log("Returning length %u\n", len);
		return len;
	}
	usb_log("Unexpected state %02X\n", ubd->state);
	return -EPIPE;
}

static int
usbbulkSCSIReset(struct ufifloppyDev* ubd)
{
	ubd->state = UBD_STATE_CMD;
	usbendpointINAK(&ubd->readEndpoint);
	usbendpointOACK(&ubd->writeEndpoint);
	return 0;
}

static int
usbbulkSCSIGetMaxLun(struct ufifloppyDev* ubd)
{
	unsigned int i;

	for (i = 15; i > 0; i--) {
		if (ubd->lun[i]) {
			return i;
		}
	}
	return 0;
}

static int
usbbulkSCSICommand(void* cookie,
		   struct usbdevfs_ctrltransfer* ct)
{
	unsigned char* buffer = ct->data;

	switch (ct->bRequestType) {
		case USB_DIR_OUT | USB_TYPE_CLASS | USB_RECIP_INTERFACE:
			switch (ct->bRequest) {
				case 0xFF:
					if (ct->wValue) {
						usb_log("Invalid wValue %04X for SCSI-BULK reset\n", ct->wValue);
						break;
					}
					if (ct->wLength) {
						usb_log("Invalid wLength %04X for SCSI-BULK reset\n", ct->wLength);
						break;
					}
					usbbulkSCSIReset(cookie);
					return 0;
				default:
					usb_log("Unexpected output command %02X for SCSI-BULK interface\n", ct->bRequest);
					break;
			}
			break;
		case USB_DIR_IN | USB_TYPE_CLASS | USB_RECIP_INTERFACE:
			switch (ct->bRequest) {
				case 0xFE:
					if (ct->wValue) {
						usb_log("Invalid wValue %04X for SCSI-BULK get LUNs\n", ct->wValue);
						break;
					}
					if (ct->wLength < 1) {
						usb_log("Invalid wLength %04X for SCSI-BULK get LUNs\n", ct->wLength);
						break;
					}
					buffer[0] = usbbulkSCSIGetMaxLun(cookie);
					return 1;
				default:
					usb_log("Unexpected input command %02X for SCSI-BULK interface\n", ct->bRequest);
					break;
			}
			break;
		default:
			usb_log("Unexpected request type %02X\n", ct->bRequestType);
			break;
	}
	return -EPIPE;
}

static int 
usbbulkIntfGetDescriptor(void* cookie, struct usb_interface_descriptor* uid) {
	struct ufifloppyDev* ubd = cookie;

        uid->bInterfaceClass = USB_CLASS_MASS_STORAGE;
        uid->bInterfaceSubClass = 0x05;	/* SFF-8070i */
        uid->bInterfaceProtocol = ubd->cfg.transport;
        uid->iInterface = 4;
	return 0;
}

static struct usbinterface_fns usbbulkfloppy_intffns = {
	.intfDeviceControl = usbbulkSCSICommand,
	.intfGetDescriptor = usbbulkIntfGetDescriptor,
};

static int
ufifloppyGetName(struct usblogdev* ul,
		 unsigned char* buffer)
{
	struct ufifloppy* ub = list_entry(ul, struct ufifloppy, logdev);

	setBE16(buffer, sprintf(buffer + 2, "%s #%u:%u", (ub->deviceType == CDROM) ? "CDROM" : "Floppy", ub->parent->id, ub->lun));
	return getBE16(buffer) + 2;
}

static int
ufifloppyGetSettings(struct usblogdev* ul,
		     unsigned char* buffer)
{
	struct ufifloppy* ub = list_entry(ul, struct ufifloppy, logdev);
	unsigned char* ptr;

	ptr = stuffString(buffer, "Floppy");
	ptr = stuffBoolean(ptr, 1, "startConnected", ub->cfg.startConnected);
	ptr = stuffBoolean(ptr, 1, "connected", ub->fildes != -1);
	ptr = stuffFileName(ptr, 1, "deviceType", (ub->deviceType == CDROM) ? "CDROM" : "Floppy");
	ptr = stuffFileName(ptr, 1, "fileName", ub->cfg.fileName);
	ptr = stuffBoolean(ptr, 0, "preventRemoval", ub->preventRemoval);
	return ptr - buffer;
}

static int 
ufifloppySetSettings(struct usblogdev* ul,
		     const unsigned char* inp, 
		     size_t inplen, 
		     unsigned char* buffer)
{
	struct ufifloppy* ub = list_entry(ul, struct ufifloppy, logdev);
	struct ufifloppyDev* ubd = ub->parent;
	struct ufifloppyCfg usbcfg;
	const unsigned char* ptr;
	const unsigned char* key;
	const unsigned char* val;
	unsigned char type;
	int oldConnected;
	int newConnected;
	enum ub_type newDeviceType;

	usbcfg.startConnected = ub->cfg.startConnected;
	usbcfg.fileName = strdup(ub->cfg.fileName);
	oldConnected = newConnected = ub->fildes != -1;
	newDeviceType = ub->deviceType;
	ptr = inp;
	while ((ptr = fetchChange(ptr, inp + inplen, &key, &type, &val)) != NULL) {
		if (!keycmp(key, "startConnected")) {
			retrieveBoolean(&usbcfg.startConnected, type, val);
		} else if (!keycmp(key, "fileName")) {
			retrieveString(&usbcfg.fileName, type, val);
		} else if (!keycmp(key, "connected")) {
			retrieveBoolean(&newConnected, type, val);
		} else if (!keycmp(key, "deviceType")) {
			char* dt = NULL;
			retrieveString(&dt, type, val);
			if (dt) {
				newDeviceType = strcasecmp(dt, "CDROM") ? FLOPPY : CDROM;
				free(dt);
			}
		}
	}
	if (ub->cfg.startConnected != usbcfg.startConnected) {
		dictSetBoolean(usbcfg.startConnected, "%s.lun%u.startConnected", ubd->configKey, ub->lun);
		ub->cfg.startConnected = usbcfg.startConnected;
		ul->seq++;
	}
	if (strcmp(usbcfg.fileName, ub->cfg.fileName) || oldConnected != newConnected || ub->deviceType != newDeviceType) {
		if (oldConnected) {
			if (ub->preventRemoval) {
				free(usbcfg.fileName);
				usbcfg.fileName = NULL;
				return -EBUSY;
			}
			close(ub->fildes);
			ub->fildes = -1;
		}
		if (strcmp(usbcfg.fileName, ub->cfg.fileName)) {
			dictSetString(usbcfg.fileName, "%s.lun%u.fileName", ubd->configKey, ub->lun);
			free(ub->cfg.fileName);
			ub->cfg.fileName = usbcfg.fileName;
		}
		if (newDeviceType != ub->deviceType) {
			ub->deviceType = newDeviceType;
			dictSetString((newDeviceType == CDROM) ? "CDROM" : "Floppy", "%s.lun%u.deviceType", ubd->configKey, ub->lun);
			if (ub->deviceType == CDROM) {
				ub->sectorSize = 2048;
			} else {
				ub->sectorSize = 512;
			}
		}
		if (newConnected) {
			ub->writeProtected = 0;
			ub->fildes = open(ub->cfg.fileName, O_RDWR);
			if (ub->fildes == -1) {
				ub->fildes = open(ub->cfg.fileName, O_RDONLY);
				if (ub->fildes != -1) {
					ub->writeProtected = 1;
				}
			}
			if (ub->fildes != -1) {
				ub->unitAttention = 0x062901;
			}
		}
		ul->seq++;
	}
	return 0;
}

static int
ufifloppyDevGetName(struct usblogdev* ul,
		    unsigned char* buffer)
{
	struct ufifloppyDev* ubd = list_entry(ul, struct ufifloppyDev, logdev);

	setBE16(buffer, sprintf(buffer + 2, "Storage #%u", ubd->id));
	return getBE16(buffer) + 2;
}

static int
ufifloppyDevGetSettings(struct usblogdev* ul,
		        unsigned char* buffer)
{
	struct ufifloppyDev* ubd = list_entry(ul, struct ufifloppyDev, logdev);
	unsigned char* ptr;

	ptr = stuffString(buffer, "FloppyDev");
	ptr = stuffBoolean(ptr, 1, "transport", ubd->cfg.transport ? 1 : 0);
	ptr = stuffInteger(ptr, 1, "luns", ubd->cfg.maxLun);
	return ptr - buffer;
}

static int
ufifloppyAdd(struct ufifloppyDev* ubd,
	     unsigned int lun)
{
	struct ufifloppy* ub;
	const char* d;

	ub = malloc(sizeof(*ub));
	if (!ub) {
		usb_log("Cannot allocate memory for USB floppy\n");
		return -ENOMEM;
	}
	memset(ub, 0, sizeof(*ub));
	ub->parent = ubd;
	ub->lun = lun;
	ub->cfg.fileName = strdup(dictGetString("/tmp/bootw98.flp", "%s.lun%u.fileName", ubd->configKey, lun));
	ub->cfg.startConnected = dictGetBoolean(1, "%s.lun%u.startConnected", ubd->configKey, lun);
	d = dictGetString("Floppy", "%s.lun%u.deviceType", ubd->configKey, ub->lun);
	if (d && !strcasecmp(d, "CDROM")) {
		ub->deviceType = CDROM;
	} else {
		ub->deviceType = FLOPPY;
	}
	ub->sectorSize = (ub->deviceType == CDROM) ? 2048 : 512;
	if (ub->cfg.startConnected) {
		ub->writeProtected = 0;
		ub->fildes = open(ub->cfg.fileName, O_RDWR);
		if (ub->fildes == -1) {
			ub->fildes = open(ub->cfg.fileName, O_RDONLY);
			if (ub->fildes != -1) {
				ub->writeProtected = 1;
			}
		}
	} else {
		ub->fildes = -1;
	}
	ub->unitAttention = 0x062901;
	usblogdevInit(&ub->logdev);
	ub->logdev.getName = ufifloppyGetName;
	ub->logdev.getSettings = ufifloppyGetSettings;
	ub->logdev.setSettings = ufifloppySetSettings;
	usbconfigAddLogicalDevice(&ubd->parent->usbconfig, &ub->logdev);
	ubd->lun[lun] = ub;
	return 0;
}

static int
ufifloppyDelete(struct ufifloppy* ub)
{
	struct ufifloppyDev* ubd;

	if (!ub) {
		return 0;
	}
	ubd = ub->parent;
	if (ubd->lun[ub->lun] != ub) {
		return -ESTALE;
	}
	ubd->lun[ub->lun] = NULL;
	usbconfigRemoveLogicalDevice(&ubd->parent->usbconfig, &ub->logdev);
	free(ub->cfg.fileName);
	close(ub->fildes);
	free(ub);
	return 0;
}

static int
ufifloppyDevTransport(struct ufifloppyDev* ubd,
		      unsigned int transport)
{
	if (transport != 0x00 && transport != 0x50) {
		return -EINVAL;
	}
	if (transport == 0x00) {
		struct usbendpoint* ue;

		ue = malloc(sizeof(*ue));
		if (!ue) {
			usb_log("Cannot allocate memory for USB endpoint\n");
			return -ENOMEM;
		}
		usbendpointInit(ue);
		ue->endpointCookie = ubd;
		ue->bEndpointAddress = USB_DIR_IN;
	        ue->bmAttributes = USB_ENDPOINT_XFER_INT;
        	ue->wMaxPacketSize = 2;
	        ue->bInterval = 0x20;
		ubd->intrEndpoint = ue;
		usbinterfaceAddEndpoint(&ubd->usbinterface, ue);
	} else if (ubd->intrEndpoint) {
		usbinterfaceRemoveEndpoint(&ubd->usbinterface, ubd->intrEndpoint);
		ubd->intrEndpoint = NULL;
	}
	ubd->cfg.transport = transport;
	ubd->usbinterface.fns = (transport == 0x00) ? &ufifloppy_intffns : &usbbulkfloppy_intffns;
	usbconfigAssignEndpoints(&ubd->parent->usbconfig);
	return 0;
}

static int 
ufifloppyDevSetSettings(struct usblogdev* ul,
		        const unsigned char* inp, 
		        size_t inplen, 
		        unsigned char* buffer)
{
	struct ufifloppyDev* ubd = list_entry(ul, struct ufifloppyDev, logdev);
	struct ufifloppyDevCfg usbcfg;
	const unsigned char* ptr;
	const unsigned char* key;
	const unsigned char* val;
	unsigned char type;

	usbcfg.transport = ubd->cfg.transport;
	usbcfg.maxLun = ubd->cfg.maxLun;
	ptr = inp;
	while ((ptr = fetchChange(ptr, inp + inplen, &key, &type, &val)) != NULL) {
		if (type == 'B' && !keycmp(key, "transport")) {
			if (val[0] == 0 && val[1] == 1) {
				usbcfg.transport = val[2] ? 0x50 : 0x00;
			}
		} else if (type == 'I' && !keycmp(key, "luns")) {
			size_t ln;
						
			ln = getBE16(val);
			if (ln != 4) {
				return -EINVAL;
			}
			usbcfg.maxLun = getBE32(val + 2);
		}
	}
	if (usbcfg.maxLun != ubd->cfg.maxLun || usbcfg.transport != ubd->cfg.transport) {
		if (usbinterfaceIsClaimed(&ubd->usbinterface)) {
			return -EBUSY;
		}
		if (usbcfg.maxLun < 1 || usbcfg.maxLun > 8) {
			return -EINVAL;
		}
		if (usbcfg.maxLun != ubd->cfg.maxLun) {
			char s[10];
			
			sprintf(s, "%u", usbcfg.maxLun);
			dictSetString(s, "%s.luns", ubd->configKey);
		}
		while (usbcfg.maxLun > ubd->cfg.maxLun) {
			int err;

			err = ufifloppyAdd(ubd, ubd->cfg.maxLun);
			if (err) {
				return err;
			}
			ubd->cfg.maxLun++;
		}
		while (usbcfg.maxLun < ubd->cfg.maxLun) {
			int err;
			
			err = ufifloppyDelete(ubd->lun[ubd->cfg.maxLun - 1]);
			if (err) {
				return err;
			}
			ubd->cfg.maxLun--;
		}
		if (usbcfg.transport != ubd->cfg.transport) {
			int err;
			
			dictSetString(usbcfg.transport ? "FALSE" : "TRUE", "%s.cbi", ubd->configKey);
			err = ufifloppyDevTransport(ubd, usbcfg.transport);
			if (err) {
				return err;
			}
		}
		ul->seq++;
	}
	return 0;
}

/* Connect/disconnect: 0x062800 */

static struct ufifloppyDev*
buildUFIFloppy(struct myusbdev* ud, unsigned int id)
{
	struct usbinterface* ui;
	struct ufifloppyDev* ubd;
	struct ufifloppy* ub;
	struct usbendpoint* ue;
	unsigned int transport;
	unsigned int i;
	char key[40];
			
	sprintf(key, "petrusb%u:%u.mass", ud->id, id);
	if (!dictGetBoolean(1, "%s.present", key)) {
		return NULL;
	}
	ubd = malloc(sizeof(*ubd));
	if (!ubd) {
		usb_log("Cannot allocate memory for USB floppy device\n");
		return NULL;
	}
	memset(ubd, 0, sizeof(*ubd));
	ubd->parent = ud;
	ubd->iInterface = usbdeviceAddString(&ud->usbdevice, "USB Mass Storage", 1);
	ubd->configKey = strdup(key);
	ubd->id = id;

	transport = dictGetBoolean(0, "%s.cbi", ubd->configKey) ? 0x00 : 0x50;
	ubd->csw.sign = 0x53425355;
	ubd->cfg.transport = transport;
	ubd->cfg.maxLun = dictGetInteger(1, "%s.luns", ubd->configKey);
	if (ubd->cfg.maxLun > 8) {
		ubd->cfg.maxLun = 8;
	}
	if (ubd->cfg.maxLun < 1) {
		ubd->cfg.maxLun = 1;
	}
	
	usblogdevInit(&ubd->logdev);
	ubd->logdev.getName = ufifloppyDevGetName;
	ubd->logdev.getSettings = ufifloppyDevGetSettings;
	ubd->logdev.setSettings = ufifloppyDevSetSettings;
	
	usbconfigAddLogicalDevice(&ud->usbconfig, &ubd->logdev);

	ub = malloc(sizeof(*ub));
	if (!ub) {
		usb_log("Cannot allocate memory for USB fake lun\n");
		return NULL;
	}
	memset(ub, 0, sizeof(*ub));
	ub->parent = ubd;
	ub->fildes = -1;
	ub->unitAttention = 0x000000;
	ub->lun = -1;
	ubd->lun[16] = ub;

	/* Crappy Linux says that we are SCSI-2, so LUNs above 7 are disallowed */
	for (i = 0; i < ubd->cfg.maxLun; i++) {
		int err;

		err = ufifloppyAdd(ubd, i);
		if (err) {
			return NULL;
		}
	}

	ui = &ubd->usbinterface;
	usbinterfaceInit(ui);
	ui->fns = (transport == 0x00) ? &ufifloppy_intffns : &usbbulkfloppy_intffns;
	ui->intfCookie = ubd;

	ue = &ubd->readEndpoint;
	usbendpointInit(ue);
	ue->endpointCookie = ubd;
	ue->bEndpointAddress = USB_DIR_IN;
        ue->bmAttributes = USB_ENDPOINT_XFER_BULK;
        ue->wMaxPacketSize = 64;
        ue->bInterval = 0x00;
	usbinterfaceAddEndpoint(ui, &ubd->readEndpoint);

	ue = &ubd->writeEndpoint;
	usbendpointInit(ue);
	ue->outputXfer = (transport == 0x00) ? ufifloppy_writepipeproc : usbbulkfloppy_writepipeproc;
	ue->endpointCookie = ubd;
	ue->bEndpointAddress = USB_DIR_OUT;
        ue->bmAttributes = USB_ENDPOINT_XFER_BULK;
        ue->wMaxPacketSize = 64;
        ue->bInterval =  0x00;
	usbinterfaceAddEndpoint(ui, &ubd->writeEndpoint);

	if (transport == 0x00) {
		ue = malloc(sizeof(*ue));
		if (!ue) {
			usb_log("Cannot allocate memory for USB endpoint\n");
			return NULL;
		}
		usbendpointInit(ue);
		ue->endpointCookie = ubd;
		ue->bEndpointAddress = USB_DIR_IN;
	        ue->bmAttributes = USB_ENDPOINT_XFER_INT;
        	ue->wMaxPacketSize = 2;
	        ue->bInterval = 0x20;
		ubd->intrEndpoint = ue;
		usbinterfaceAddEndpoint(ui, ue);
	}

	usbconfigAddInterface(&ud->usbconfig, ui);

	return ubd;
}

static struct usbdevice_fns mydev_devfns = {
};

static void
usbmassdevDestroy(struct ufifloppyDev* ubd)
{
	unsigned int i;

	for (i = 0; i < 16; i++) {
		ufifloppyDelete(ubd->lun[i]);
	}
	free(ubd->lun[16]);
	ubd->lun[16] = NULL;
	usbconfigRemoveLogicalDevice(&ubd->parent->usbconfig, &ubd->logdev);
	usbconfigRemoveInterface(&ubd->parent->usbconfig, &ubd->usbinterface);
	usbinterfaceRemoveEndpoint(&ubd->usbinterface, &ubd->writeEndpoint);
	usbinterfaceRemoveEndpoint(&ubd->usbinterface, &ubd->readEndpoint);
	if (ubd->intrEndpoint) {
		usbinterfaceRemoveEndpoint(&ubd->usbinterface, ubd->intrEndpoint);
	}
	
	usblogdevDestroy(&ubd->logdev);
	usbinterfaceDestroy(&ubd->usbinterface);
	usbendpointDestroy(&ubd->readEndpoint);
	usbendpointDestroy(&ubd->writeEndpoint);
	if (ubd->intrEndpoint) {
		usbendpointDestroy(ubd->intrEndpoint);
		free(ubd->intrEndpoint);
	}
	free(ubd->configKey);
	ubd->configKey = NULL;
	free(ubd);
}

static void
usbmassDevForkCleanup(struct usblogdev* ul)
{
	struct myusbdev* ubd = list_entry(ul, struct myusbdev, logdev);

}

static int
usbmassDevGetName(struct usblogdev* ul,
                 unsigned char* buffer)
{
	struct myusbdev* ubd = list_entry(ul, struct myusbdev, logdev);

        setBE16(buffer, sprintf(buffer + 2, "USB Storage"));
        return getBE16(buffer) + 2;
}

static int
usbmassDevGetSettings(struct usblogdev* ul,
		     unsigned char* buffer)
{
        struct myusbdev* ubd = list_entry(ul, struct myusbdev, logdev);
        unsigned char* ptr;

        ptr = stuffString(buffer, "MassDevice");
        ptr = stuffInteger(ptr, 1, "devs", ubd->devicesCount);
        return ptr - buffer;
}

static void
usbmassdevSyncDevicesCount(struct myusbdev* ud)
{
	unsigned int maxDev;
	unsigned int fint;
	
	maxDev = ud->devicesCount;
	for (fint = 0; fint < MAX_USBMASS_DEV; fint++) {
		struct ufifloppyDev* dev;

		dev = ud->devices[fint];
		if (fint < maxDev) {
			if (!dev) {
				ud->devices[fint] = buildUFIFloppy(ud, fint);
			}
		} else {
			if (dev) {
				usbmassdevDestroy(dev);
				ud->devices[fint] = NULL;
			}
		}
	}
	usbconfigAssignEndpoints(&ud->usbconfig);
}

static int
usbmassDevSetSettings(struct usblogdev* ul,
		     const unsigned char* inp,
		     size_t inplen,
		     unsigned char* buffer)
{
        struct myusbdev* ubd = list_entry(ul, struct myusbdev, logdev);
        const unsigned char* ptr;
        const unsigned char* key;
        const unsigned char* val;
        unsigned char type;
	unsigned int newDevices;

	newDevices = ubd->devicesCount;	
        ptr = inp;
        while ((ptr = fetchChange(ptr, inp + inplen, &key, &type, &val)) != NULL) {
		if (!keycmp(key, "devs")) {
			retrieveInteger(&newDevices, type, val);
			if (newDevices > MAX_USBMASS_DEV) {
				newDevices = MAX_USBMASS_DEV;
			}
                }
        }
	if (newDevices != ubd->devicesCount) {
		if (usbdeviceIsClaimed(&ubd->usbdevice)) {
			return -EBUSY;
		}
		if (newDevices != ubd->devicesCount) {
			dictSetInteger(newDevices, "petrusb%u.devs", ubd->id);
			ubd->devicesCount = newDevices;
			ul->seq++;
		}
		usbmassdevSyncDevicesCount(ubd);
	}
        return 0;
}

#define VM_CHECKSIZE(structure, expected)       \
   if (sizeof(structure) != expected) {         \
      asm volatile(".print \"Unexpected ABI change: Structure " #structure " should take %c0 bytes, but took %c1!\"; .abort" : : "i"(expected), "i"(sizeof(structure))); \
   }


static void __attribute__((constructor)) init2(void) {
	unsigned int i;

	for (i = 0; i < 8; i++) {
		struct myusbdev* ud;

		if (!dictGetBoolean(i == 0, "petrusb%u.present", i)) {
			continue;
		}
		ud = malloc(sizeof(*ud));
		if (!ud) {
			usb_log("Cannot allocate memory for USB device\n");
			return;
		}
		memset(ud, 0, sizeof(*ud));
		usbdeviceInit(&ud->usbdevice);
		ud->usbdevice.devcookie = &ud->usbdevice;
		ud->usbdevice.fns = &mydev_devfns;
		ud->id = i;
		ud->usbdevice.desc.bcdUSB        = 0x0110;
		ud->usbdevice.desc.bMaxPacketSize0 = 8;
		ud->usbdevice.desc.idVendor      = 0x0000;
		ud->usbdevice.desc.idProduct     = 0x0000;
		ud->usbdevice.desc.bcdDevice     = 0x0125;
		ud->usbdevice.desc.iManufacturer = usbdeviceAddString(&ud->usbdevice, "VMware, Inc.", 1);
		ud->usbdevice.desc.iProduct      = usbdeviceAddString(&ud->usbdevice, "Mass Storage", 1);
                {
                        char sn[20] = "DEADBEEF0000";
                        sprintf(sn + 10, "%2.2X", i);
                        ud->usbdevice.desc.iSerialNumber   = usbdeviceAddString(&ud->usbdevice, sn, 1);
                }
		ud->devicesCount = dictGetInteger(1, "petrusb%u.devs", i);
		usbconfigInit(&ud->usbconfig);
		usbdeviceAddConfig(&ud->usbdevice, &ud->usbconfig);

		usblogdevInit(&ud->logdev);
	        ud->logdev.getName     = usbmassDevGetName;
	        ud->logdev.getSettings = usbmassDevGetSettings;
	        ud->logdev.setSettings = usbmassDevSetSettings;
		ud->logdev.forkCleanup = usbmassDevForkCleanup;
	        usbconfigAddLogicalDevice(&ud->usbconfig, &ud->logdev);

		usbmassdevSyncDevicesCount(ud);

		usbdeviceRegister(&ud->usbdevice, -1, -1);
	}

	VM_CHECKSIZE(struct usbbulkCBW, 31);
	VM_CHECKSIZE(struct usbbulkCSW, 13);
}
