#define _GNU_SOURCE
#define _LARGEFILE_SOURCE
#define _LARGEFILE64_SOURCE

#include "vmnetDevice.h"
#include "usbcontrol.h"
#include "dictionary.h"
#include "usbdev.h"
#include "rndis.h"

#include <pthread.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <sys/poll.h>

struct usbrndisdev;

struct myusbdev {
	struct usbdevice usbdevice;
	struct usbconfig usbconfig;
	unsigned int id;
};

struct usbrndisxfer {
	struct queuedXfer   xfer;
	size_t		    bufferLen;
	unsigned char	    buffer[1536 + 44];
};

struct usbctlxfer {
	struct usbctlxfer*  next;
	size_t		    bufferLen;
	unsigned char	    buffer[1536 + 44];
};

struct usbrndisdevCfg {
        char*                   fileName;
        int                     startConnected;
};

struct usbrndisdev {
	struct myusbdev*	parent;
	pthread_mutex_t	    devLock;
	char*               configKey;
	struct usbinterface controlUSBInterface;
	struct usbendpoint  controlUSBIntrEP;
	struct usbinterface data0USBInterface;
	struct usbinterface data1USBInterface;
	struct usbendpoint  data1USBBulkInEP;
	struct usbendpoint  data1USBBulkOutEP;
	struct usblogdev	logdev;
	struct vmnetDevice  vmnet;
	unsigned char	    mac[6];
	int		    macFromCfg;
	struct usbctlxfer*  replies;
	struct usbctlxfer** nextReply;
	int		    iControlInterface;
	int		    iDataInterface;
	unsigned int		ifnum;
	struct usbrndisdevCfg	cfg;
        enum { DOWN=0, RUNNING=1 } state;
	unsigned int		ifflags;
};

static int
controlUSBIntrEPCallback(void* cookie,
			 size_t maxlen,
			 void* buffer)
{
	struct usbrndisdev* ubd = cookie;
	int addIACK;

	pthread_mutex_lock(&ubd->devLock);
	/* We should always find ubd->replies non-NULL here, otherwise we get there for no reason
	   and we should cancel this attempt to wakeup URB */
	addIACK = ubd->replies != NULL;
	pthread_mutex_unlock(&ubd->devLock);
	if (addIACK) {
		usbendpointIACK(&ubd->controlUSBIntrEP, controlUSBIntrEPCallback, 0);
	}
	/* Form notification: 1, reserved: 0 */
	if (maxlen > 8) {
		maxlen = 8;
	}
	memset(buffer, 0, maxlen);
	if (maxlen > 0) {
		*(unsigned char*)buffer = 1;
	}
	return maxlen;
}

static int
data1USBInterfaceCopyData(struct queuedXfer* cookie,
			  size_t len,
			  void* buffer)
{
	struct usbrndisxfer* xfer = list_entry(cookie, struct usbrndisxfer, xfer);
	
	if (len > xfer->bufferLen) {
		len = xfer->bufferLen;
	}
	memcpy(buffer, xfer->buffer, len);
	free(xfer);
	usb_log("Returned %u buffer\n", len);
	return len;
}

static int
data1USBInterfaceDiscardData(struct queuedXfer* cookie)
{
	struct usbrndisxfer* xfer = list_entry(cookie, struct usbrndisxfer, xfer);
	
	usb_log("Discarding received buffer %p\n", xfer);
	free(xfer);
	return 0;
}

static int
data1USBInterfaceDataReady(void* cookie,
			   int fd,
			   int revents)
{
	int r;
	struct usbrndisdev* ubd = cookie;
	struct usbrndisxfer* xfer;	

	if (ubd->state != RUNNING) {
		char dummy[1536];

		usb_log("interface down, throwing packet away\n");
		read(fd, dummy, sizeof(dummy));
		return 0;
	}
	xfer = malloc(sizeof(*xfer));
	if (!xfer) {
		char dummy[1536];

		usb_log("malloc failed, throwing packet away\n");
		read(fd, dummy, sizeof(dummy));
		return 0;
	}
	r = read(fd, xfer->buffer + 44, sizeof(xfer->buffer) - 44);
	usb_log("Received %d bytes from network\n", r);
	if (r > 12) {
		xfer->bufferLen = r + 44;
		setLE32(xfer->buffer,      REMOTE_NDIS_PACKET_MSG);
		setLE32(xfer->buffer +  4, xfer->bufferLen);
		setLE32(xfer->buffer +  8, 44 - 8);	/* DataOffset */
		setLE32(xfer->buffer + 12, r);		/* DataLength */
		memset(xfer->buffer + 16, 0, 28);
		xfer->xfer.xfer = data1USBInterfaceCopyData;
		xfer->xfer.abort = data1USBInterfaceDiscardData;
		r = usbendpointQueueIACK(&ubd->data1USBBulkInEP, &xfer->xfer);
		if (r < 0) {
			usb_log("Queued IACK failed: %d\n", r);
			free(xfer);
		}
	} else {
		free(xfer);
	}
	return 0;
}
			   
static int
data1USBInterfaceSend(void* cookie,
                      size_t len,
                      const void* bufferVoid)
{
	struct usbrndisdev* ubd = cookie;
	size_t xferlen;
	size_t speclen;
	unsigned int xferoffs;
	const unsigned char* buffer = bufferVoid;

	if (ubd->state != RUNNING) {
		return -EPIPE;
	}
	if (len < 44) {
		return -EPIPE;
	}
	if (getLE32(buffer) != REMOTE_NDIS_PACKET_MSG) {
		return -EPIPE;
	}
	speclen = getLE32(buffer + 4);
	if (speclen > len || speclen < 44) {
		return -EPIPE;
	}
	xferoffs = getLE32(buffer + 8) + 8;
	if (xferoffs > speclen || xferoffs < 44) {
		return -EPIPE;
	}
	xferlen = getLE32(buffer + 12);
	if (xferlen > speclen) {
		return -EPIPE;
	}
	if (xferoffs + xferlen > speclen) {
		return -EPIPE;
	}
	vmnetSendPacket(&ubd->vmnet, buffer + xferoffs, xferlen);
	return len;
}

static void
usbrndisUpdateMAC(struct usbrndisdev* ubd)
{
}

static void
usbrndisSetIFFlags(struct usbrndisdev* ubd,
                 unsigned int ifflags)
{
        ubd->ifflags = ifflags;
        if (vmnetIsConnected(&ubd->vmnet)) {
                vmnetSetIFFlags(&ubd->vmnet, ifflags);
        }
}

static void
usbrndisStartNetwork(struct usbrndisdev* ubd)
{
	if (!vmnetIsConnected(&ubd->vmnet)) {
		if (!vmnetConnect(&ubd->vmnet, "/dev/vmnet8")) {
			if (ubd->macFromCfg) {
				if (vmnetSetMACAddress(&ubd->vmnet, ubd->mac)) {
					ubd->macFromCfg = 0;
				}
			}
			vmnetSetIFFlags(&ubd->vmnet, ubd->ifflags);
			if (!ubd->macFromCfg) {
				if (vmnetGetMACAddress(&ubd->vmnet, ubd->mac) == 0) {
					char c[40];
					
					sprintf(c, "%02x:%02x:%02x:%02x:%02x:%02x", ubd->mac[0], ubd->mac[1], ubd->mac[2],
										    ubd->mac[3], ubd->mac[4], ubd->mac[5]);
					dictSetString(c, "%s.generatedAddress", ubd->configKey);
					dictSetString("generated", "%s.addressType", ubd->configKey);
					ubd->macFromCfg = 1;
					usbrndisUpdateMAC(ubd);
				}
			}
			
			usbcontrolStartWatchingFileDescriptor(ubd->vmnet.fd, ubd, data1USBInterfaceDataReady, POLLIN);
		}
	}	
}

static void
usbrndisStopNetwork(struct usbrndisdev* ubd)
{
        if (vmnetIsConnected(&ubd->vmnet)) {
                usbcontrolStopWatchingFileDescriptor(ubd->vmnet.fd);
                vmnetDisconnect(&ubd->vmnet);
        }
}

static int
addRNDISReply(struct usbrndisdev* ubd,
	      struct usbctlxfer* xfer)
{
	int addIACK = 0;

	xfer->next = NULL;
	pthread_mutex_lock(&ubd->devLock);
	if (ubd->nextReply) {
		*ubd->nextReply = xfer;
	} else {
		ubd->replies = xfer;
		ubd->nextReply = &xfer->next;
		addIACK = 1;
	}
	pthread_mutex_unlock(&ubd->devLock);
	if (addIACK) {
		usbendpointIACK(&ubd->controlUSBIntrEP, controlUSBIntrEPCallback, 0);
	}
	return 0;
}

static int
getRNDISReply(struct usbrndisdev* ubd,
	      size_t maxlen,
	      unsigned char* buffer)
{
	struct usbctlxfer* reply;

	pthread_mutex_lock(&ubd->devLock);
	reply = ubd->replies;
	if (reply) {
		if ((ubd->replies = reply->next) == NULL) {
			ubd->nextReply = NULL;
		}
	}
	if (ubd->replies) {
		usbendpointIACK(&ubd->controlUSBIntrEP, controlUSBIntrEPCallback, 0);
	} else {
		usbendpointINAK(&ubd->controlUSBIntrEP);
	}
	pthread_mutex_unlock(&ubd->devLock);
	if (reply) {
		if (maxlen > reply->bufferLen) {
			maxlen = reply->bufferLen;
		}
		memcpy(buffer, reply->buffer, maxlen);
		free(reply);
		return maxlen;
	} else {
		buffer[0] = 0;
		return 1;
	}
}

static struct usbctlxfer*
allocCtlXfer(void)
{
	struct usbctlxfer* ctl = malloc(sizeof(*ctl));
	return ctl;
}

static void
rndisInit(struct usbrndisdev* ubd,
	  struct usbctlxfer* ctl,
	  const unsigned char* request,
	  size_t rqlen)
{
	ubd->state = RUNNING;
	usbendpointOACK(&ubd->data1USBBulkOutEP);
	usbendpointINAK(&ubd->data1USBBulkInEP);
	setLE32(ctl->buffer,      REMOTE_NDIS_INITIALIZE_CMPLT);
	setLE32(ctl->buffer +  4, 52);
	setLE32(ctl->buffer +  8, getLE32(request + 8));
	setLE32(ctl->buffer + 12, 0);
	setLE32(ctl->buffer + 16, 1);	/* Major version */
	setLE32(ctl->buffer + 20, 1);	/* Minor version */
	setLE32(ctl->buffer + 24, 1);	/* Device Flags (1 = connectionless) */
	setLE32(ctl->buffer + 28, 0);	/* Medium (0 = ethernet) */
	setLE32(ctl->buffer + 32, 1);	/* Max Packets Per Transfer */
	setLE32(ctl->buffer + 36, 1518 + 44);	/* Max Transfer Size */
	setLE32(ctl->buffer + 40, 1);	/* Packet Alignment Factor */
	setLE32(ctl->buffer + 44, 0);	/* AF List Offset */
	setLE32(ctl->buffer + 48, 0);	/* AF List Length */
}

static u_int32_t oidlist[] = {
/* Required Object IDs (OIDs) */
OID_GEN_SUPPORTED_LIST,
OID_GEN_HARDWARE_STATUS,
OID_GEN_MEDIA_SUPPORTED,
OID_GEN_MEDIA_IN_USE,
OID_GEN_MAXIMUM_LOOKAHEAD,
OID_GEN_MAXIMUM_FRAME_SIZE,
OID_GEN_LINK_SPEED,
OID_GEN_TRANSMIT_BUFFER_SPACE,
OID_GEN_RECEIVE_BUFFER_SPACE,
OID_GEN_TRANSMIT_BLOCK_SIZE,
OID_GEN_RECEIVE_BLOCK_SIZE,
OID_GEN_VENDOR_ID,
OID_GEN_VENDOR_DESCRIPTION,
OID_GEN_CURRENT_PACKET_FILTER,
OID_GEN_CURRENT_LOOKAHEAD,
OID_GEN_DRIVER_VERSION,
OID_GEN_MAXIMUM_TOTAL_SIZE,
OID_GEN_PROTOCOL_OPTIONS,
OID_GEN_MAC_OPTIONS,
OID_GEN_MEDIA_CONNECT_STATUS,
OID_GEN_MAXIMUM_SEND_PACKETS,
OID_GEN_VENDOR_DRIVER_VERSION,
OID_GEN_SUPPORTED_GUIDS,
OID_GEN_NETWORK_LAYER_ADDRESSES,
OID_GEN_TRANSPORT_HEADER_OFFSET,
OID_GEN_MACHINE_NAME,
OID_GEN_RNDIS_CONFIG_PARAMETER,
OID_GEN_VLAN_ID,

/* Optional OIDs */
OID_GEN_MEDIA_CAPABILITIES,
OID_GEN_PHYSICAL_MEDIUM,

/* Required statistics OIDs */
OID_GEN_XMIT_OK,
OID_GEN_RCV_OK,
OID_GEN_XMIT_ERROR,
OID_GEN_RCV_ERROR,
OID_GEN_RCV_NO_BUFFER,

/* Optional statistics OIDs */
OID_GEN_DIRECTED_BYTES_XMIT,
OID_GEN_DIRECTED_FRAMES_XMIT,
OID_GEN_MULTICAST_BYTES_XMIT,
OID_GEN_MULTICAST_FRAMES_XMIT,
OID_GEN_BROADCAST_BYTES_XMIT,
OID_GEN_BROADCAST_FRAMES_XMIT,
OID_GEN_DIRECTED_BYTES_RCV,
OID_GEN_DIRECTED_FRAMES_RCV,
OID_GEN_MULTICAST_BYTES_RCV,
OID_GEN_MULTICAST_FRAMES_RCV,
OID_GEN_BROADCAST_BYTES_RCV,
OID_GEN_BROADCAST_FRAMES_RCV,
OID_GEN_RCV_CRC_ERROR,
OID_GEN_TRANSMIT_QUEUE_LENGTH,
OID_GEN_GET_TIME_CAPS,
OID_GEN_GET_NETCARD_TIME,
OID_GEN_NETCARD_LOAD,
OID_GEN_DEVICE_PROFILE,
OID_GEN_INIT_TIME_MS,
OID_GEN_RESET_COUNTS,
OID_GEN_MEDIA_SENSE_COUNTS,
OID_GEN_FRIENDLY_NAME,
OID_GEN_MINIPORT_INFO,
OID_GEN_RESET_VERIFY_PARAMETERS,

/* IEEE 802.3 (Ethernet) OIDs */
NDIS_802_3_MAC_OPTION_PRIORITY,

OID_802_3_PERMANENT_ADDRESS,
OID_802_3_CURRENT_ADDRESS,
OID_802_3_MULTICAST_LIST,
OID_802_3_MAXIMUM_LIST_SIZE,
OID_802_3_MAC_OPTIONS,
OID_802_3_RCV_ERROR_ALIGNMENT,
OID_802_3_XMIT_ONE_COLLISION,
OID_802_3_XMIT_MORE_COLLISIONS,
OID_802_3_XMIT_DEFERRED,
OID_802_3_XMIT_MAX_COLLISIONS,
OID_802_3_RCV_OVERRUN,
OID_802_3_XMIT_UNDERRUN,
OID_802_3_XMIT_HEARTBEAT_FAILURE,
OID_802_3_XMIT_TIMES_CRS_LOST,
OID_802_3_XMIT_LATE_COLLISIONS,
};

static u_int32_t
rndisQueryInt(struct usbrndisdev* ubd,
	      u_int32_t oid,
	      unsigned char* buffer)
{
	switch (oid) {
		case OID_GEN_SUPPORTED_LIST:
			memcpy(buffer, oidlist, sizeof(oidlist));
			return sizeof(oidlist);
		case OID_GEN_MAXIMUM_FRAME_SIZE:
			setLE32(buffer, 1514);
			return 4;
		case OID_GEN_MAC_OPTIONS:
			setLE32(buffer, NDIS_MAC_OPTION_RECEIVE_SERIALIZED | NDIS_MAC_OPTION_FULL_DUPLEX);
			return 4;
		case OID_GEN_SUPPORTED_GUIDS:
			return 0;
		case OID_GEN_MEDIA_CAPABILITIES:
			return 0;
		case OID_GEN_VENDOR_DESCRIPTION:
			strcpy(buffer, "VMware\'s USB to Ethernet Convertor");
			return strlen(buffer) + 1;
		case OID_GEN_LINK_SPEED:
			setLE32(buffer, 120000);	/* 12Mbps */
			return 4;
		case OID_GEN_MEDIA_CONNECT_STATUS:
			setLE32(buffer, vmnetIsConnected(&ubd->vmnet) ? NDIS_MEDIA_STATE_CONNECTED : NDIS_MEDIA_STATE_DISCONNECTED);
			return 4;
		case OID_802_3_CURRENT_ADDRESS:
			memcpy(buffer, ubd->mac, 6);
			return 6;
		case OID_802_3_MAXIMUM_LIST_SIZE:
			setLE32(buffer, 0);
			return 4;
		default:
			usb_warn("Unsupported OID %08X queried!\n", oid);
			return 0xC0010015;
	}
}

static void
rndisSetPacketFilter(struct usbrndisdev* ubd,
		     u_int32_t filter)
{
	unsigned int flags = 0;

	if (filter) {
		flags |= IFF_UP;
	}
	if (filter & (NDIS_PACKET_TYPE_BROADCAST)) {
		flags |= IFF_BROADCAST;
	}
	if (filter & (NDIS_PACKET_TYPE_ALL_MULTICAST)) {
		flags |= IFF_ALLMULTI;
	}
	if (filter & (NDIS_PACKET_TYPE_MULTICAST)) {
		flags |= IFF_MULTICAST;
	}
	if (filter & (NDIS_PACKET_TYPE_PROMISCUOUS)) {
		flags |= IFF_PROMISC;
	}
	usbrndisSetIFFlags(ubd, flags);
}

static u_int32_t
rndisSetInt(struct usbrndisdev* ubd,
	    u_int32_t oid,
	    const unsigned char* buffer,
	    size_t len)
{
	switch (oid) {
		case OID_GEN_CURRENT_PACKET_FILTER:
			if (len != 4) {
				usb_log("Invalid packet filter length %u\n", len);
				return 0xC0010015;
			}
			rndisSetPacketFilter(ubd, getLE32(buffer));
			return 0;
		case OID_GEN_CURRENT_LOOKAHEAD:
			if (len != 4) {
				usb_log("Invalid lookahead buffer length %u\n", len);
				return 0xC0010015;
			}
			usb_log("Lookahead length: %u\n", getLE32(buffer));
			return 0;
		case OID_GEN_TRANSPORT_HEADER_OFFSET:
			return 0;		/* No TCP/UDP/IPX/... offload... */
		case OID_GEN_NETWORK_LAYER_ADDRESSES:
			return 0;		/* No TCP/UDP/IPX/... offload... */
		default:
			usb_warn("Unsupported OID %08X set!\n", oid);
			return 0xC0010015;
	}
}

static void
rndisQuery(struct usbrndisdev* ubd,
	   struct usbctlxfer* ctl,
	   const unsigned char* request,
	   size_t rqlen)
{
	u_int32_t ln;
	
	setLE32(ctl->buffer,      REMOTE_NDIS_QUERY_CMPLT);
	setLE32(ctl->buffer +  4, 24);
	setLE32(ctl->buffer +  8, getLE32(request + 8));
	setLE32(ctl->buffer + 12, 0);
	setLE32(ctl->buffer + 16, 0);
	setLE32(ctl->buffer + 20, 0);
	ln = rndisQueryInt(ubd, getLE32(request + 12), ctl->buffer + 24);
	if (ln < 1024) {
		setLE32(ctl->buffer +  4, 24 + ln);
		setLE32(ctl->buffer + 16, ln);
		setLE32(ctl->buffer + 20, 24 - 8);
	} else {
		setLE32(ctl->buffer + 12, ln);
	}
}

static void
rndisSet(struct usbrndisdev* ubd,
	 struct usbctlxfer* ctl,
	 const unsigned char* request,
	 size_t rqlen)
{
	u_int32_t err;

	setLE32(ctl->buffer,      REMOTE_NDIS_SET_CMPLT);
	setLE32(ctl->buffer +  4, 16);
	setLE32(ctl->buffer +  8, getLE32(request + 8));
	err = rndisSetInt(ubd, getLE32(request + 12), request + 8 + getLE32(request + 20), getLE32(request + 16));
	setLE32(ctl->buffer + 12, err);
}

static void
rndisReset(struct usbrndisdev* ubd,
	   struct usbctlxfer* ctl,
	   const unsigned char* request,
	   size_t rqlen)
{
	setLE32(ctl->buffer,      REMOTE_NDIS_RESET_CMPLT);
	setLE32(ctl->buffer +  4, 16);
	setLE32(ctl->buffer +  8, 0);
	setLE32(ctl->buffer + 12, 1);
}

static void
rndisKeepalive(struct usbrndisdev* ubd,
	       struct usbctlxfer* ctl,
	       const unsigned char* request,
	       size_t rqlen)
{
	setLE32(ctl->buffer,      REMOTE_NDIS_KEEPALIVE_CMPLT);
	setLE32(ctl->buffer +  4, 16);
	setLE32(ctl->buffer +  8, getLE32(request + 8));
	setLE32(ctl->buffer + 12, 0);
}

static int
processRNDISRequest(struct usbrndisdev* ubd,
		    size_t len,
		    const unsigned char* buffer)
{
	struct usbctlxfer* ctl;
	size_t nlen;

	if (len < 12) {
		return -EPIPE;
	}
	nlen = getLE32(buffer + 4);
	if (nlen > len) {
		return -EPIPE;
	}
	ctl = allocCtlXfer();
	switch (getLE32(buffer)) {
		case REMOTE_NDIS_INITIALIZE_MSG:
			rndisInit(ubd, ctl, buffer, nlen);
			break;
		case REMOTE_NDIS_HALT_MSG:
			free(ctl);
		        ubd->state = DOWN;
		        usbendpointOStall(&ubd->data1USBBulkOutEP);
		        usbendpointIStall(&ubd->data1USBBulkInEP);
		        usbrndisSetIFFlags(ubd, 0);
			return len;
		case REMOTE_NDIS_QUERY_MSG:
			rndisQuery(ubd, ctl, buffer, nlen);
			break;
		case REMOTE_NDIS_SET_MSG:
			rndisSet(ubd, ctl, buffer, nlen);
			break;
		case REMOTE_NDIS_RESET_MSG:
			rndisReset(ubd, ctl, buffer, nlen);
			break;
		case REMOTE_NDIS_KEEPALIVE_MSG:
			rndisKeepalive(ubd, ctl, buffer, nlen);
			break;
		default:
			setLE32(ctl->buffer,      getLE32(buffer) | 0x80000000);
			setLE32(ctl->buffer +  4, 16);
			setLE32(ctl->buffer +  8, getLE32(buffer + 8));
			setLE32(ctl->buffer + 12, 0xC00000BB);
			break;
	}
	ctl->bufferLen = getLE32(ctl->buffer + 4);
	addRNDISReply(ubd, ctl);
	return len;
}

static int
controlUSBInterfaceCmd(void* cookie,
		       struct usbdevfs_ctrltransfer* ct)
{
	int mask = 0;
	unsigned char* dt;
	struct usbrndisdev* ubd = cookie;
	
	if (ct->wValue) {
		return -EPIPE;
	}
	dt = ct->data;
	if (ct->bRequestType & USB_DIR_IN) {
		mask = 0x100;
	}
	switch (mask | ct->bRequest) {
		case 0x000:	/* SEND_ENCAPSULATED_COMMAND */
			return processRNDISRequest(ubd, ct->wLength, ct->data);
		case 0x101:	/* GET_ENCAPSULATED_RESPONSE */
			return getRNDISReply(ubd, ct->wLength, ct->data);
	}
	return -EPIPE;
}

static int 
controlUSBInterfaceGetDescriptor(void* cookie, struct usb_interface_descriptor* uid) {
	struct usbrndisdev* ubd = cookie;

	uid->bInterfaceClass = 0x02;
	uid->bInterfaceSubClass = 0x02;
	uid->bInterfaceProtocol = 0xFF;
        uid->iInterface = ubd->iControlInterface;
	return 0;
}

static int
controlUSBInterfaceGetFunctionalDescriptor(void* cookie,
					   unsigned char* buffer,
					   unsigned int idx)
{
	struct usbrndisdev* ubd = cookie;

	switch (idx) {
		case 0:
			buffer[1] = USB_DT_CS_INTERFACE;
			buffer[2] = 0x00;
			setLE16(buffer + 3, 0x0110);	/* Standard version */
			return 5;
		case 1:
			buffer[1] = USB_DT_CS_INTERFACE;
			buffer[2] = 0x06;
			buffer[3] = ubd->controlUSBInterface.bInterfaceNumber;
			buffer[4] = ubd->data0USBInterface.bInterfaceNumber;
			return 5;
		default:
			return -ENOENT;
	}
}

static struct usbinterface_fns controlUSBInterface_fns = {
	.intfDeviceControl = controlUSBInterfaceCmd,
	.intfGetDescriptor = controlUSBInterfaceGetDescriptor,
	.intfGetFunctionalDescriptor = controlUSBInterfaceGetFunctionalDescriptor,
};

static int
data1USBInterfaceCmd(void* cookie,
		       struct usbdevfs_ctrltransfer* ct)
{
	return -EPIPE;
}

static int 
data1USBInterfaceGetDescriptor(void* cookie, struct usb_interface_descriptor* uid) {
	struct usbrndisdev* ubd = cookie;

        uid->bInterfaceClass = 0x0A;
        uid->bInterfaceSubClass = 0x00;
        uid->bInterfaceProtocol = 0x00;
        uid->iInterface = ubd->iDataInterface;
	return 0;
}

static struct usbinterface_fns data1USBInterface_fns = {
	.intfDeviceControl = data1USBInterfaceCmd,
	.intfGetDescriptor = data1USBInterfaceGetDescriptor,
};

static void
getMAC(struct usbrndisdev* ubd, int offset)
{
	const char* addrType;
	const char* addr;

	memcpy(ubd->mac, "\x00\x50\x56\xFF\xFF\x00", 6);
	memcpy(ubd->mac, "\x00\x50\x56\xFF\xFF\x00", 6);
	ubd->mac[5] += offset;
	ubd->macFromCfg = 0;

	addrType = dictGetString("generated", "%s.addressType", ubd->configKey);
	if (addrType && !strcmp(addrType, "static")) {
		addr = dictGetString(NULL, "%s.address", ubd->configKey);
	} else {
		addr = dictGetString(NULL, "%s.generatedAddress", ubd->configKey);
	}
	if (addr) {
		unsigned int c[6];
		
		if (sscanf(addr, "%02x:%02x:%02x:%02x:%02x:%02x",
			   &c[0], &c[1], &c[2], &c[3], &c[4], &c[5]) == 6) {
			if (c[0] <= 0xFF && (c[0] & 1) == 0 &&
			    c[1] <= 0xFF &&
			    c[2] <= 0xFF &&
			    c[3] <= 0xFF &&
			    c[4] <= 0xFF &&
			    c[5] <= 0xFF) {
				ubd->mac[0] = c[0];
				ubd->mac[1] = c[1];
				ubd->mac[2] = c[2];
				ubd->mac[3] = c[3];
				ubd->mac[4] = c[4];
				ubd->mac[5] = c[5];
				ubd->macFromCfg = 1;
			}
		}
	}
}

static void
usbrndisForkCleanup(struct usblogdev* ul)
{
        struct usbrndisdev* ubd = list_entry(ul, struct usbrndisdev, logdev);

        /* Close opened fd... */
        usbrndisStopNetwork(ubd);
}

static int
usbrndisGetName(struct usblogdev* ul,
              unsigned char* buffer)
{
        struct usbrndisdev* ubd = list_entry(ul, struct usbrndisdev, logdev);

        setBE16(buffer, sprintf(buffer + 2, "USB RNDIS Ethernet #%u", ubd->ifnum));
        return getBE16(buffer) + 2;
}

static int
usbrndisGetSettings(struct usblogdev* ul,
                  unsigned char* buffer)
{
        struct usbrndisdev* ubd = list_entry(ul, struct usbrndisdev, logdev);
        unsigned char* ptr;

        ptr = stuffString(buffer, "RNDISEthernet");
        ptr = stuffBoolean(ptr, 1, "startConnected", ubd->cfg.startConnected);
        ptr = stuffFileName(ptr, 1, "fileName", ubd->cfg.fileName);
        ptr = stuffBoolean(ptr, 1, "connected", vmnetIsConnected(&ubd->vmnet));
        ptr = stuffBoolean(ptr, 0, "running", ubd->state == RUNNING);
        return ptr - buffer;
}

static int
usbrndisSetSettings(struct usblogdev* ul,
                  const unsigned char* inp,
                  size_t inplen,
                  unsigned char* buffer)
{
        struct usbrndisdev* ubd = list_entry(ul, struct usbrndisdev, logdev);
        const unsigned char* ptr;
        const unsigned char* key;
        const unsigned char* val;
        unsigned char type;
        int oldConnected;
        int newConnected;
        int newStartConnected;
        char* newFileName;

        newStartConnected = ubd->cfg.startConnected;
        newFileName = strdup(ubd->cfg.fileName);
        newConnected = oldConnected = vmnetIsConnected(&ubd->vmnet);
        ptr = inp;
        while ((ptr = fetchChange(ptr, inp + inplen, &key, &type, &val)) != NULL) {
                if (!keycmp(key, "startConnected")) {
                        retrieveBoolean(&newStartConnected, type, val);
                } else if (!keycmp(key, "connected")) {
                        retrieveBoolean(&newConnected, type, val);
                } else if (!keycmp(key, "fileName")) {
                        retrieveString(&newFileName, type, val);
                }
        }
        if (ubd->cfg.startConnected != newStartConnected) {
                dictSetBoolean(newStartConnected, "%s.startConnected", ubd->configKey);
                ubd->cfg.startConnected = newStartConnected;
                ul->seq++;
        }
        if (strcmp(newFileName, ubd->cfg.fileName) || oldConnected != newConnected) {
                if (strcmp(newFileName, ubd->cfg.fileName)) {
                        dictSetString(newFileName, "%s.fileName", ubd->configKey);
                }
                usbrndisStopNetwork(ubd);
                free(ubd->cfg.fileName);
                ubd->cfg.fileName = newFileName;
                if (newConnected) {
                        usbrndisStartNetwork(ubd);
                }
                ul->seq++;
        }
        return 0;
}

static struct usbrndisdev*
usbrndisCreate(struct myusbdev* parent, unsigned int fint)
{
	struct usbinterface* ui;
	struct usbrndisdev* ubd;
	struct usbendpoint* ue;
	char key[40];
			
	sprintf(key, "petrusbrndis%u:%u.eth", parent->id, fint);
	if (!dictGetBoolean(fint == 0, "%s.present", key)) {
		return NULL;
	}
	ubd = malloc(sizeof(*ubd));
	if (!ubd) {
		usb_log("Cannot allocate memory for USB CDC Ethernet device\n");
		return NULL;
	}
	memset(ubd, 0, sizeof(*ubd));

	pthread_mutex_init(&ubd->devLock, NULL);

	ubd->parent = parent;
	ubd->configKey = strdup(key);
	ubd->ifnum = fint;
	ubd->cfg.startConnected = dictGetBoolean(1, "%s.startConnected", key);
	ubd->cfg.fileName = strdup(dictGetString("/dev/vmnet8", "%s.fileName", key));
	ubd->state = DOWN;
	vmnetInit(&ubd->vmnet);
	getMAC(ubd, parent->id * 16 + fint);
	
	ubd->iControlInterface = usbdeviceAddString(&parent->usbdevice, "Ethernet Interface", 1);
	ubd->iDataInterface = usbdeviceAddString(&parent->usbdevice, "Data Interface", 1);
	
	ui = &ubd->controlUSBInterface;
	usbinterfaceInit(ui);
	ui->fns = &controlUSBInterface_fns;
	ui->intfCookie = ubd;

	ue = &ubd->controlUSBIntrEP;
	usbendpointInit(ue);
	ue->endpointCookie = ubd;
	ue->bEndpointAddress = USB_DIR_IN;
        ue->bmAttributes = USB_ENDPOINT_XFER_INT;
        ue->wMaxPacketSize = 32;
        ue->bInterval = 64;
	usbinterfaceAddEndpoint(ui, ue);
	
	usbconfigAddInterface(&parent->usbconfig, ui);

	ui = &ubd->data1USBInterface;
	usbinterfaceInit(ui);
	ui->fns = &data1USBInterface_fns;
	ui->intfCookie = ubd;
	
	ue = &ubd->data1USBBulkInEP;
	usbendpointInit(ue);
	ue->maxQueueLength = 64;
	ue->endpointCookie = ubd;
	ue->bEndpointAddress = USB_DIR_IN;
        ue->bmAttributes = USB_ENDPOINT_XFER_BULK;
        ue->wMaxPacketSize = 64;
        ue->bInterval =  0x00;
	
	usbinterfaceAddEndpoint(ui, ue);

	ue = &ubd->data1USBBulkOutEP;
	usbendpointInit(ue);
	ue->outputXfer = data1USBInterfaceSend;
	ue->endpointCookie = ubd;
	ue->bEndpointAddress = USB_DIR_OUT;
        ue->bmAttributes = USB_ENDPOINT_XFER_BULK;
        ue->wMaxPacketSize = 64;
        ue->bInterval =  0x00;
	
	usbinterfaceAddEndpoint(ui, ue);

	usbconfigAddInterface(&parent->usbconfig, ui);

        usblogdevInit(&ubd->logdev);
        ubd->logdev.getName     = usbrndisGetName;
        ubd->logdev.getSettings = usbrndisGetSettings;
        ubd->logdev.setSettings = usbrndisSetSettings;
        ubd->logdev.forkCleanup = usbrndisForkCleanup;
        usbconfigAddLogicalDevice(&ubd->parent->usbconfig, &ubd->logdev);

        if (ubd->cfg.startConnected) {
                usbrndisStartNetwork(ubd);
        }

	return ubd;
}

static struct usbdevice_fns mydev_devfns = {
};

#define VM_CHECKSIZE(structure, expected)       \
   if (sizeof(structure) != expected) {         \
      asm volatile(".print \"Unexpected ABI change: Structure " #structure " should take %c0 bytes, but took %c1!\"; .abort" : : "i"(expected), "i"(sizeof(structure))); \
   }


static void __attribute__((constructor)) init2(void) {
	unsigned int i;

	for (i = 0; i < 8; i++) {
		struct myusbdev* ud;
		unsigned int fint;

		if (!dictGetBoolean(i == 0, "petrusbrndis%u.present", i)) {
			continue;
		}
		ud = malloc(sizeof(*ud));
		if (!ud) {
			usb_log("Cannot allocate memory for USB device\n");
			return;
		}
		memset(ud, 0, sizeof(*ud));
		usbdeviceInit(&ud->usbdevice);
		ud->usbdevice.devcookie = &ud->usbdevice;
		ud->usbdevice.fns = &mydev_devfns;
		ud->id = i;
		ud->usbdevice.desc.bDeviceClass  = 0x02;
                ud->usbdevice.desc.bcdUSB        = 0x0110;
                ud->usbdevice.desc.bMaxPacketSize0 = 8;
                ud->usbdevice.desc.idVendor      = 0x07B2;
                ud->usbdevice.desc.idProduct     = 0x4220;
                ud->usbdevice.desc.bcdDevice     = 0x0125;
                ud->usbdevice.desc.iManufacturer = usbdeviceAddString(&ud->usbdevice, "VMware, Inc.", 1);
                ud->usbdevice.desc.iProduct      = usbdeviceAddString(&ud->usbdevice, "RNDIS Ethernet", 1);
                {
                        char sn[20] = "DEADBEEF0200";
                        sprintf(sn + 10, "%2.2X", i);
                        ud->usbdevice.desc.iSerialNumber   = usbdeviceAddString(&ud->usbdevice, sn, 1);
                }


		usbconfigInit(&ud->usbconfig);
		usbdeviceAddConfig(&ud->usbdevice, &ud->usbconfig);

		for (fint = 0; fint < 1; fint++) {
			usbrndisCreate(ud, fint);
		}

		usbdeviceRegister(&ud->usbdevice, -1, -1);
	}
}
