#define _GNU_SOURCE
#define _LARGEFILE_SOURCE
#define _LARGEFILE64_SOURCE

#include "vmnetDevice.h"
#include "usbdev.h"

#include <errno.h>
#include <fcntl.h>
#include <string.h>

#include <sys/ioctl.h>

#include "vnet.h"

void
vmnetInit(struct vmnetDevice* dev)
{
	dev->fd = -1;
	dev->ifver = -1;
}

static void
vmnetGetAPIVersion(struct vmnetDevice* dev)
{
	int version;
	int err;

	/* Only new/old ioctl matters for us... That is, success => new interface, failure => old or nothing */	
	err = ioctl(dev->fd, SIOCGETAPIVERSION, &version);
	dev->ifver = (err == -1) ? 0 : 1;
}

int
vmnetIsConnected(const struct vmnetDevice* dev)
{
	return dev->fd != -1;
}

int
vmnetConnect(struct vmnetDevice* dev,
	     const char* fileName)
{
	uid_t su;
	int fd;

	if (dev->fd != -1) {
		errno = EBUSY;
		return -1;
	}
	su = becomeSuperUser();
	fd = open(fileName, O_RDWR | O_NONBLOCK);
	restoreSuperUser(su);
	usb_log("Opened network device: %d, %d\n", fd, errno);
	if (fd != -1) {
		dev->fd = fd;
		fcntl(fd, F_SETFD, fcntl(fd, F_GETFD) | FD_CLOEXEC);
		vmnetGetAPIVersion(dev);
		return 0;
	}
	return -1;
}

void
vmnetDisconnect(struct vmnetDevice* dev)
{
	if (dev->fd != -1) {
		close(dev->fd);
		dev->fd = -1;
	}
}

int
vmnetSetMACAddress(const struct vmnetDevice* dev,
		   const unsigned char mac[6])
{
	struct VNet_SetMacAddrIOCTL io;
	int err;
	
	io.version = 1;
	io.flags = VNET_SETMACADDRF_UNIQUE;
	memcpy(io.addr, mac, 6);
	switch (dev->ifver) {
		case 0:
			err = ioctl(dev->fd, OLD_SIOCSETMACADDR, &io);
			break;
		case 1:
			err = ioctl(dev->fd, SIOCSETMACADDR, &io);
			break;
		default:
			errno = EINVAL;
			err = -1;
			break;
	}
	return err;
}

int
vmnetSetIFFlags(const struct vmnetDevice* dev,
		u_int32_t flags)
{
	int err;

	flags &= IFF_UP | IFF_BROADCAST | IFF_PROMISC | IFF_MULTICAST | IFF_ALLMULTI;
	switch (dev->ifver) {
		case 0:
			err = ioctl(dev->fd, SIOCSIFFLAGS, flags);
			break;
		case 1:
			err = ioctl(dev->fd, SIOCSIFFLAGS, &flags);
			break;
		default:
			errno = EINVAL;
			err = -1;
			break;
	}
	return err;	
}

int
vmnetGetMACAddress(const struct vmnetDevice* dev,
		   unsigned char mac[6])
{
	int err;

	err = ioctl(dev->fd, SIOCGIFADDR, mac);
	return err;
}

void
vmnetSendPacket(const struct vmnetDevice* dev,
		const void* data,
		size_t len)
{
	if (vmnetIsConnected(dev)) {
		ssize_t ln;

		ln = write(dev->fd, data, len);
		usb_log("Attempt to send %u bytes to network: %d bytes sent\n", len, ln);
	}
}
