#! /usr/bin/perl -w

use strict;
use warnings;
use Socket;
use POSIX ":sys_wait_h";
	

my $standalone = 0;

sub escape($) {
	my ($chr) = @_;
	
	if ($chr eq '&') {
		return '&amp;';
	} elsif ($chr eq '<') {
		return '&lt;';
	} elsif ($chr eq '>') {
		return '&gt;';
	} elsif ($chr eq '"') {
		return '&quot;';
	} else {
		return '&#'.ord($chr).';';
	}
}

sub htmlspecialchars($) {
	my ($val) = @_;
	
	$val =~ s/([\000-\037<>&"\177-\377])/escape($1)/ge;  #"
	return $val;
}

sub urlencode($) {
	my ($val) = @_;
	
	$val =~ s/([\000-\057\177-\377])/sprintf('%%%02X', ord($1))/ge;
	return $val;
}

sub mysockaddr_un($) {
	my ($sun) = @_;
	
	return chr(1).chr(0).$sun;
}

sub httpStatus($) {
	my ($status) = @_;

	if ($standalone) {
		print "HTTP/1.0 $status\r\n";
	} else {
		print "Status: $status\n";
	}
}

sub listConns() {
	my @conns;
	my $line;

	open UNIXNET, "</proc/net/unix" or die "Unable to open /proc/net/unix: $!";
	$line = <UNIXNET>;	# Skip first line
	while (defined($line = <UNIXNET>)) {
		chomp $line;

		if ($line =~ /^\S+:\s+\S+\s+\S+\s+\S+\s+\S+\s+\S+\s+\S+\s+\@(vmwusb-.*)$/) {
			my $p = $1;

			push @conns, $1;
		}

	}
	close UNIXNET;
	return \@conns;
}

sub connectVM($) {
	my ($name) = @_;
	
	socket(SOCKET, PF_UNIX, SOCK_DGRAM, 0) or die "Cannot create socket: $!";
	bind(SOCKET, mysockaddr_un('')) or die "Cannot bind: $!";
	connect(SOCKET, mysockaddr_un(chr(0).$name)) or die "Cannot connect: $!";
	return *SOCKET;
}

my $seq = rand(256);

sub askVM($$$) {
	my ($sock, $op, $query) = @_;
	my $cnt;
	my $sq = $seq++ & 255;
	my $key = chr($sq).chr($op);
	
	for ($cnt = 0; $cnt < 5; $cnt++) {
		send($sock, $key.$query, MSG_DONTWAIT);
		my $timeout = 1;
		my $rin = '';
		vec($rin, fileno($sock), 1) = 1;
		do {
			my $nfound;
			my $rout;

			($nfound, $timeout) = select($rout = $rin, undef, undef, $timeout);
			if ($nfound == 1) {
				recv($sock, $query, 4096, MSG_DONTWAIT);
				if (length($query) >= 2) {
					if (substr($query, 0, 2) eq $key) {
						return substr($query, 2);
					}
				}
			}
		} while ($timeout != 0);
	}
	die "No answer";
}

sub vmGetVersion($) {
	my ($sock) = @_;
	my $v;
	
	$v = askVM($sock, 0, '');
	my ($err, $str) = unpack('CC/a', $v);
	if ($err != 0) {
		return '???';
	}
	return $str;
}

sub vmGetDeviceList($) {
	my ($sock) = @_;
	my $v;
	my @d2;

	$v = askVM($sock, 1, '');
	my ($err, $dummy, @devs) = unpack('CCN/L', $v);
	if ($err == 0) {
		foreach my $v (@devs) {
			push @d2, [ $v / 65536, $v & 65535 ];
		}
	}
	return \@d2;
}

sub vmGetDeviceName($$$) {
	my ($sock, $busid, $devid) = @_;
	my $v;
	my $err;
	
	$v = askVM($sock, 2, pack('nn', $busid, $devid));
	($err, $v) = unpack('Ca*', $v);
	if ($err == 0) {
		my ($vendor, $product, $serial);
		
		eval {
			($vendor, $product, $serial) = unpack('n/an/an/a', $v);
		};
		if ($@) {
			return undef;
		}
		return [ $vendor, $product, $serial ];
		
	}
	return undef;
}

sub vmGetNextInterface($$$$) {
	my ($sock, $busid, $devid, $intf) = @_;
	my $v;
	my $err;
	
	$v = askVM($sock, 3, pack('nnn', $busid, $devid, $intf));
	($err, $v) = unpack('Ca*', $v);
	if ($err == 0) {
		my ($id, $name);
		
		eval {
			($id, $name) = unpack('nn/a', $v);
		};
		if ($@) {
			return undef;
		}
		return [ $id, $name ];
	}
	return undef;
}

sub vmGetVmxConfig($) {
	my ($sock) = @_;
	my $v;
	my $err;
	
	$v = askVM($sock, 4, '');
	($err, $v) = unpack('Ca*', $v);
	if ($err == 0) {
		my ($name);
		
		eval {
			($name) = unpack('N/a', $v);
		};
		if ($@) {
			return undef;
		}
		return $name;
	}
	return undef;
}

sub vmGetDeviceInfo($$$;$) {
	my ($sock, $busid, $devid, $intf) = @_;
	my $v;
	my $err;
	
	# Note that INTERFACE here is USB interface, not logical device ID
	if (not defined($intf)) {
		$intf = -1;
	}
	$v = askVM($sock, 5, pack('nnn', $busid, $devid, $intf));
	($err, $v) = unpack('Ca*', $v);
	if ($err == 0) {
		my ($rslt);
		
		eval {
			($rslt) = unpack('C', $v);
		};
		if ($@) {
			return undef;
		}
		return $rslt;
	}
	return undef;
}	

sub vmGetSettings($$$$) {
	my ($sock, $busid, $devid, $intf) = @_;
	my $v;
	my $err;
	
	$v = askVM($sock, 6, pack('nnn', $busid, $devid, $intf));
	($err, $v) = unpack('Ca*', $v);
	if ($err == 0) {
		my @arr;
		my ($cfgseq, $devname);
		
		eval {
			($cfgseq, $devname, $v) = unpack('Nn/aa*', $v);
		};
		return undef if ($@);
		print STDERR ">>>".unpack('H*', $v)."<<<\n";
		while (length($v) >= 5) {
			my ($str, $flag, $val);
			
			eval {
				($str, $flag, $val, $v) = unpack('n/aa1n/aa*', $v);
			};
			last if ($@);
			push @arr, [ $str, $flag, $val ];
		}
		return [ $cfgseq, $devname, \@arr ];
	}
	return undef;
}

sub vmSetSettings($$$$$$) {
	my ($sock, $busid, $devid, $intf, $seq, $ops) = @_;
	my $v;
	my $err;
	
	$v = pack('nnnN', $busid, $devid, $intf, $seq);
	foreach my $k (@$ops) {
		$v .= pack('n/a*a1n/a*', $k->[0], $k->[1], $k->[2]);
	}
	$v = askVM($sock, 7, $v);
	($err, $v) = unpack('Ca*', $v);
	return $err;
}
	
sub listVM($) {
	my ($active) = @_;
	my $ok = undef;
	
	print '<form method="GET" id="fvm" name="fvm">'."\n";
	print '<input type="hidden" name="op" value="main" />'."\n";
	print '<tr><td><b>Virtual machine:</b></td><td>';
	print '<select name="vm" size="1" onChange="javascript:fvm.submit()">';
	print '<option value="">-- please select VM --</option>'."\n";
	foreach my $v (sort @{listConns()}) {
		my $sock;
		my $d2;
		my $nm;
		
		$sock = connectVM($v);
		$nm = vmGetVmxConfig($sock);
		close $sock;
		if (defined($nm)) {
			$nm = $nm.' ('.$v.')';
		} else {
			$nm = $v;
			next;
		}
		print '<option value="'.htmlspecialchars($v).'"';
		if ($v eq $active) {
			print ' selected';
			$ok = $v;
		}
		print '>'.htmlspecialchars($nm).'</option>'."\n";
	}
	print '</select>'."\n";
	print '<input type="submit" name="dummy" value=" &gt;&gt; " /></td></tr>';
	print '</form>'."\n";
	return $ok;
}

sub showDevList($$$$) {
	my ($sock, $vm, $dev, $devs) = @_;
	my $ret = undef;
	
	print '<form method="GET" id="sdl" name="sdl">'."\n";
	printf '<input type="hidden" name="op" value="main" />'."\n";
	printf '<input type="hidden" name="vm" value="%s" />'."\n", htmlspecialchars($vm);
	print '<tr><td><b>Devices available:</b></td><td>';
	print '<select name="dev" onChange="javascript:sdl.submit()">'."\n";
	print '<option value="">-- select device --</option>'."\n";
	foreach my $v (@$devs) {
		my $vd;
		my $nm;
		
		$nm = vmGetDeviceName($sock, $v->[0], $v->[1]);
		$vd = sprintf '%d-%d', $v->[0], $v->[1];
		if (defined($nm)) {
			$nm = join(', ', @$nm);
		} else {
			$nm = $vd;
		}
		printf '<option value="%s"', htmlspecialchars($vd);
		if ($dev eq $vd) {
			print ' selected';
			$ret = $v;
		}
		printf '>%s</option>'."\n", htmlspecialchars($nm);
	}
	print '</select>'."\n";
	print '<input type="submit" name="dummy" value=" &gt;&gt; " /></td></tr>'."\n";
	print '</form>'."\n";
	return $ret;
}

sub showIntfList($$$$$) {
	my ($sock, $vm, $dev, $intf, $devid) = @_;
	my $ret = undef;
	my $last = -1;
	
	print '<form method="GET" id="sil" name="sil">'."\n";
	printf '<input type="hidden" name="op" value="main" />'."\n";
	printf '<input type="hidden" name="vm" value="%s" />'."\n", htmlspecialchars($vm);
	printf '<input type="hidden" name="dev" value="%s" />'."\n", htmlspecialchars($dev);
	print '<tr>';
	print '<td><b>Interfaces available:</b></td><td>';
	print '<select name="intf" onChange="javascript:sil.submit()">'."\n";
	print '<option value="">-- select interface --</option>'."\n";
	while (1) {
		my $i;
		
		$i = vmGetNextInterface($sock, $devid->[0], $devid->[1], $last);
		last if not defined($i);
		$last = $i->[0];
		my $nm = $i->[1];
		
		if ($nm eq '') {
			$nm = '#'.$last;
		}
		printf '<option value="%s"', htmlspecialchars($last);
		if ($intf eq $last) {
			print ' selected';
			$ret = $last;
		}
		printf '>%s</option>'."\n", htmlspecialchars($nm);
	}
	print '</select>'."\n";
	print '<input type="submit" name="dummy" value=" &gt;&gt; " /></td></tr>';
	print '</form>'."\n";
	return $ret;
}

sub getString($$) {
	my ($ops, $key) = @_;
	
	foreach my $dv (@$ops) {
		if ($dv->[0] eq $key) {
			if ($dv->[1] eq 'B' or $dv->[1] eq 'b') {
				if ($dv->[2] eq chr(1)) {
					return 'TRUE';
				}
				return 'FALSE';
			}
			if ($dv->[1] eq 'S' or $dv->[1] eq 's') {
				return $dv->[2];
			}
			if ($dv->[1] eq 'I' or $dv->[1] eq 'i') {
				print STDERR ">>>".unpack('H*', $dv->[2])."<<<\n";
				return unpack('N', $dv->[2]);
			}
			return '';
		}
	}
	return '';
}

sub getBool($$) {
	my ($ops, $key) = @_;
	
	my $v;
	
	$v = getString($ops, $key);
	if ($v =~ /^[1TYPtyp]/) {	# 1, TRUE, YES, POSITIVE, true, yes, positive
		return 1;
	}
	return 0;
}

sub getInteger($$) {
	my ($ops, $key) = @_;
	my $v;

	$v = getString($ops, $key);
	if ($v =~ /^\d+$/) {
		return $v;
	}
	return 0;
}

sub doSelection($$$) {
	my ($val, $name, $choices) = @_;
	my $ret;
	my $i;
	
	$ret = sprintf '<select name="%s">', htmlspecialchars($name);
	for ($i = 0; $i < scalar @$choices; $i += 2) {
		my ($k, $v) = ($choices->[$i], $choices->[$i + 1]);
		$ret .= sprintf '<option value="%s"%s>%s</option>', htmlspecialchars($k), $k eq $val ? ' selected' : '', htmlspecialchars($v);
	}
	$ret .= '</select>';
	return $ret;
}
	
sub boolSelection($$) {
	my ($val, $name) = @_;
	
	return doSelection($val, $name, [ '0' => 'No', '1' => 'Yes' ]);
}

sub booleanValue($$$$) {
	my ($keys, $name, $w0, $w1) = @_;
	my $v = getBool($keys, $name);
	return doSelection($v, 'fmt_new_'.$name, [ '0' => $w0, '1' => $w1 ]);
}

sub choiceValue($$$$) {
	my ($keys, $name, $w0, $w1) = @_;
	my $v = getString($keys, $name);
	return doSelection($v, 'fmt_new_'.$name, [ $w0 => $w0, $w1 => $w1 ]);
}

sub integerValue($$) {
	my ($keys, $name) = @_;

	my $v = getInteger($keys, $name);
	return sprintf '<input type="text" name="fmt_new_%s" value="%s" />', htmlspecialchars($name), htmlspecialchars($v);
}

sub showDeviceSettings($$$$$) {
	my ($sock, $vm, $dev, $intf, $devid) = @_;
	my $ret = undef;
	my $settings;
	
	print '<form method="GET" id="sds" name="sds">'."\n";
	printf '<input type="hidden" name="op" value="setval" />'."\n";
	printf '<input type="hidden" name="vm" value="%s" />'."\n", htmlspecialchars($vm);
	printf '<input type="hidden" name="dev" value="%s" />'."\n", htmlspecialchars($dev);
	printf '<input type="hidden" name="intf" value="%s" />'."\n", htmlspecialchars($intf);
	$settings = vmGetSettings($sock, $devid->[0], $devid->[1], $intf);
	if (not defined($settings)) {
		print '<tr><td colspan="2"><font color="#FF0000">Unable to retrieve settings</font></td></tr>'."\n";
	} else {
		my ($seq, $devname, $ops) = @$settings;
		
		if (scalar @$ops == 0) {
			print '<tr><td colspan="2"><font color="#0000FF">Device has no properties</font></td></tr>'."\n";
		} else {
			printf '<input type="hidden" name="seq" value="%s" />'."\n", htmlspecialchars($seq);
			foreach my $dv (@$ops) {
				if ($dv->[1] eq 'S') {
					printf '<input type="hidden" name="fmt_type_%s" value="%s" />'."\n", htmlspecialchars($dv->[0]), htmlspecialchars($dv->[1]);
					printf '<input type="hidden" name="fmt_old_%s" value="%s" />'."\n", htmlspecialchars($dv->[0]), htmlspecialchars(unpack('H*', $dv->[2]));
				} elsif ($dv->[1] eq 'B' or $dv->[1] eq 'I') {
					printf '<input type="hidden" name="fmt_type_%s" value="%s" />'."\n", htmlspecialchars($dv->[0]), htmlspecialchars($dv->[1]);
					printf '<input type="hidden" name="fmt_old_%s" value="%s" />'."\n", htmlspecialchars($dv->[0]), htmlspecialchars(unpack('H*', $dv->[2]));
				}
			}
			printf '<tr><td colspan="2">%s</td></tr>'."\n", htmlspecialchars($devname);
			if ($devname eq 'Floppy') {
				printf '<tr><td>Device Status:</td><td>%s</td></tr>'."\n", getBool($ops, 'preventRemoval') ? '<font color="#FF0000">In use</font>' : '<font color="#00FF00">Unused</font>';
				printf '<tr><td>Device Type:</td><td>%s</td></tr>'."\n", choiceValue($ops, 'deviceType', 'CDROM', 'Floppy');
				printf '<tr><td>Connected:</td><td>%s</td></tr>'."\n", booleanValue($ops, 'connected', 'No', 'Yes');
				printf '<tr><td>Start Connected:</td><td>%s</td></tr>'."\n", booleanValue($ops, 'startConnected', 'No', 'Yes');
				printf '<tr><td>Floppy Image File Name:</td><td><input type="text" name="fmt_new_fileName" value="%s" size="80" /></td></tr>'."\n",
						htmlspecialchars(getString($ops, 'fileName'));
			} elsif ($devname eq 'FloppyDev') {
				printf '<tr><td>Transport Type:</td><td>%s</td></tr>'."\n",      booleanValue($ops, 'transport', 'CBI', 'Bulk Only');
				printf '<tr><td>Logical Units (1-8):</td><td>%s (Windows support only 1 LUN when using CBI transport)</td></tr>'."\n", integerValue($ops, 'luns');
			} elsif ($devname eq 'CDCEthernet') {
				printf '<tr><td>Connected:</td><td>%s</td></tr>'."\n", booleanValue($ops, 'connected', 'No', 'Yes');
				printf '<tr><td>Start Connected:</td><td>%s</td></tr>'."\n", booleanValue($ops, 'startConnected', 'No', 'Yes');
				printf '<tr><td>VMnet Device Name:</td><td><input type="text" name="fmt_new_fileName" value="%s" size="80" /></td></tr>'."\n",
						htmlspecialchars(getString($ops, 'fileName'));
				printf '<tr><td>Status:</td><td>%s</td></tr>'."\n", getBool($ops, 'running') ? '<font color="#FF0000">Running</font>' : '<font color="#00FF00">Down</font>';
			} elsif ($devname eq 'RNDISEthernet') {
				printf '<tr><td>Connected:</td><td>%s</td></tr>'."\n", booleanValue($ops, 'connected', 'No', 'Yes');
				printf '<tr><td>Start Connected:</td><td>%s</td></tr>'."\n", booleanValue($ops, 'startConnected', 'No', 'Yes');
				printf '<tr><td>VMnet Device Name:</td><td><input type="text" name="fmt_new_fileName" value="%s" size="80" /></td></tr>'."\n",
						htmlspecialchars(getString($ops, 'fileName'));
				printf '<tr><td>Status:</td><td>%s</td></tr>'."\n", getBool($ops, 'running') ? '<font color="#FF0000">Running</font>' : '<font color="#00FF00">Down</font>';
			} elsif ($devname eq 'CDCDevice') {
				printf '<tr><td>Device Type:</td><td>%s</td></tr>'."\n", booleanValue($ops, 'motorola', 'Linux', 'Windows (Motorola SURFboard SB5100 drivers)');
				printf '<tr><td>Logical Devices (1-7):</td><td>%s (only active for Linux device type)</td></tr>'."\n", integerValue($ops, 'devs');
			} elsif ($devname eq 'MassDevice') {
				printf '<tr><td>Logical Devices (1-7):</td><td>%s</td></tr>'."\n", integerValue($ops, 'devs');
			} else {
				foreach my $dv (@$ops) {
					printf '<tr><td>%s (%s)</td><td>%s</td></tr>'."\n", htmlspecialchars($dv->[0]), htmlspecialchars($dv->[1]), htmlspecialchars(getString($ops, $dv->[0]));
				}
			}
			print '<tr><td colspan="2"><input type="submit" name="dummy" value=" Save new values " /></td></tr>';
		}
	}
	print '</form>'."\n";
	return $ret;
}	

sub httprequest($) {
	my ($query) = @_;

	if (not defined($query)) {
		$query = '';
	}
	my %cmds = map { $_ =~ s/\+/ /g; $_ =~ s/%([0-9a-fA-F][0-9a-fA-F])/chr(hex($1))/ge; $_; } map { split(/=/, $_, 2) } split(/&/, $query);

	my $op = $cmds{'op'};
	if (not defined($op)) {
		$op = 'main';
	}
	my $vm = $cmds{'vm'};
	if (not defined($vm)) {
		$vm = '';
	}
	my $dev = $cmds{'dev'};
	if (not defined($dev)) {
		$dev = '';
	}
	my $intf = $cmds{'intf'};
	if (not defined($intf)) {
		$intf = '';
	}

	if ($op eq 'main') {
		httpStatus('200 OK');
		print 'Content-type: text/html'."\n";
		print "\n";
		print '<html><head><title>VMware USB Device Control</title></head><body bgcolor="#FFFFC0">';
		print '<h1>VMware USB Device Control</h1>';
		print '<table>';
		if (defined($cmds{'msg'})) {
			printf '<tr><td colspan="2"><font size="+2" color="#FF0000">%s</font></td></tr>', htmlspecialchars($cmds{'msg'});
		}
		if (defined(listVM($vm))) {
			my $sock;
			my $d2;
		
			$sock = connectVM($vm);
			printf "<tr><td>Version:</td><td>%s</td></tr>\n", htmlspecialchars(vmGetVersion($sock));
			$d2 = showDevList($sock, $vm, $dev, vmGetDeviceList($sock));
			if ($d2) {
				my $busy;
				
				$busy = vmGetDeviceInfo($sock, $d2->[0], $d2->[1]);
				if (not defined($busy)) {
					$busy = '<font color="#FF0000">Unavailable</font>';
				} elsif ($busy) {
					$busy = '<font color="#0000FF">Connected</font>';
				} else {
					$busy = '<font color="#00FF00">Available</font>';
				}
				printf "<tr><td>Status:</td><td>%s</td></tr>\n", $busy;
				
				if (defined(showIntfList($sock, $vm, $dev, $intf, $d2))) {
					showDeviceSettings($sock, $vm, $dev, $intf, $d2);
				}
			}
			close $sock;
		}
		print '</table>';
		print '</body></html>';
	} elsif ($op eq 'setval') {
		my $sock;
		my ($bus, $devn);
		
		$sock = connectVM($vm);
		if ($dev =~ /^(\d+)-(\d+)$/) {
			($bus, $devn) = ($1, $2);
			if ($intf =~ /^(\d+)$/) {
				my @ops = ();
				my $err;

				foreach my $v (keys %cmds) {
					if ($v =~ /^fmt_new_(.+)$/) {
						my $nk = $1;
						my $nv = $cmds{$v};
						my $ov = pack('H*', $cmds{'fmt_old_'.$nk});
						my $t = $cmds{'fmt_type_'.$nk};
						
						if ($t eq 'S') {
							# No translation needed
						} elsif ($t eq 'B') {
							if ($nv =~ /^[1TYPtyp]/) {
								$nv = chr(1);
							} else {
								$nv = chr(0);
							}
						} elsif ($t eq 'I') {
							$nv = pack('N', $nv);
						} else {
							next;
						}
						if ($nv ne $ov) {
							push @ops, [ $nk, $t, $nv ];
						}
					}
				}
				$err = 0;
				if (scalar @ops > 0) {
					$err = vmSetSettings($sock, $bus, $devn, $intf, $cmds{'seq'}, \@ops);
				}
				httpStatus('302 Moved');
				$! = $err;
				$err = "$!";
				printf 'Location: /cgi-bin/usbcfg.pl?op=main&vm=%s&dev=%s&intf=%s&msg=%s'."\n", urlencode($vm), urlencode($dev), urlencode($intf), urlencode($err);
				print "\n";
				print 'Page moved...';
				return;
			}
		}
		httpStatus('302 Moved');
		printf 'Location: /cgi-bin/usbcfg.pl?op=main&vm=%s&dev=%s&intf=%s&msg=%s'."\n", urlencode($vm), urlencode($dev), urlencode($intf), urlencode('Bad URL... Attack attempted?');
		print "\n";
		print 'Page moved...';
		return;
	} else {
		httpStatus('200 OK');
		print 'Content-type: text/html'."\n";
		print "\n";
		print '<html><head><title>VMware USB Device Control</title></head><body bgcolor="#FFFFC0">';
		print '<h1><font color="#FF0000">Bad command</font></h1>';
		print '<p>Unknown operation <b><font color="#0000FF">'.$op.'</font></b> was requested.</p>';
		print '</body></html>';
	}
}


if (defined($ENV{'GATEWAY_INTERFACE'})) {
	httprequest($ENV{'QUERY_STRING'});
} else {
	$standalone = 1;

	$SIG{CHLD} = sub {
	                        while (waitpid(-1, WNOHANG) > 0) {
				}
		     };

	socket ACPT, AF_INET, SOCK_STREAM, getprotobyname('tcp') or die "socket: $!";
	setsockopt(ACPT, SOL_SOCKET, SO_REUSEADDR, pack("l", 1)) or die "setsockopt: $!";
	bind ACPT, sockaddr_in(4444, INADDR_ANY) or die "bind: $!";
	listen ACPT, 5 or die "listen: $!";

	printf "Listening on http://localhost:%s\n", 4444;
	
	while (1) {
		my $clnt;
		
		$clnt = accept NSOCK, ACPT;
		if (defined($clnt)) {
			my ($port, $addr) = sockaddr_in($clnt);
			my $name = gethostbyaddr($addr,AF_INET);
			if (not defined($name)) {
				$name = join('.', unpack('C4', $addr));
			}
			print "Accepted connection from ".$name.":$port\n";
			my $pid;
			if (!defined($pid = fork)) {
				print "cannot fork: $!\n";
			} elsif ($pid) {
				print "begat $pid\n";
			} else {
				my $r;
				my ($method, $url, $http);

				open STDIN, '<&NSOCK' or die "open STDIN: $!";
				open STDOUT, '>&NSOCK' or die "open STDOUT: $!";

				if (!defined($r = <NSOCK>)) {
					exit 1;
				}
				$r =~ s/\r?\n$//;

				print STDERR "==> $r\n";
				if ($r =~ m!^(\S+)\s+(.*)\s+HTTP/(\d+\.\d+)$!) {
					$method = $1;
					$url = $2;
					$http = $3;
				} elsif ($r =~ m!^(\S+)\s+(\S+)$!) {
					$method = $1;
					$url = $2;
					$http = '0.9';
				} else {
					die "Invalid request ".unpack('H*', $r);
				}
				if ($http >= 1.0) {
					while (defined($r = <NSOCK>)) {
						$r =~ s/\r?\n$//;

						print STDERR "==> $r\n";
						last if $r eq '';
					}
				}
				my $query;
				if ($url =~ /^[^?]*\?(.*)$/) {
					$query = $1;
				} else {
					$query = '';
				}
				httprequest($query);
				exit 0;
			}
			close NSOCK;
		} else {
			print "accept: $!\n";
		}
	}
}


